/* -*- Mode: C++ ; Coding: euc-japan -*- */
/* Time-stamp: <2014-07-30 03:53:13 cyamauch> */

/**
 * @file   fits_table_col.cc
 * @brief  ASCII or Binary Table1ɽ륯饹fits_table_colΥ
 */

#define CLASS_NAME "fits_table_col"

#include "config.h"

#include "fits_table_col.h"
#include "fits_table.h"

#include <math.h>
#include <stdlib.h>
#include <sli/numeric_minmax.h>

#include "private/err_report.h"

#include "private/c_memset.h"


namespace sli
{

/* this->tany array ֹ */
static const int TTYPE_IDX         = 0;
static const int TTYPE_COMMENT_IDX = 1;
static const int TALAS_IDX         = 2;
static const int TALAS_COMMENT_IDX = 3;
static const int TELEM_IDX         = 4;
static const int TELEM_COMMENT_IDX = 5;
static const int TUNIT_IDX         = 6;
static const int TUNIT_COMMENT_IDX = 7;
static const int TDISP_IDX         = 8;
static const int TDISP_COMMENT_IDX = 9;
static const int TFORM_IDX         = 10;
static const int TFORM_COMMENT_IDX = 11;
static const int TDIM_IDX          = 12;
static const int TDIM_COMMENT_IDX  = 13;
static const int TNULL_IDX         = 14;
static const int TNULL_COMMENT_IDX = 15;
static const int TZERO_IDX         = 16;
static const int TZERO_COMMENT_IDX = 17;
static const int TSCAL_IDX         = 18;
static const int TSCAL_COMMENT_IDX = 19;

static const int TANY_ARR_SIZE     = 20;

/* */
inline static size_t abs_sz2z( ssize_t val )
{
    ssize_t ret;
    if ( val < 0 ) ret = -val;
    else ret = val;
    return (size_t)ret;
}

/* */
inline static int32_t round_f2i32( float v )
{
    if ( v < 0 ) return (int32_t)(v-0.5);
    else return (int32_t)(v+0.5);
}
inline static int64_t round_f2i64( float v )
{
    if ( v < 0 ) return (int64_t)(v-0.5);
    else return (int64_t)(v+0.5);
}
inline static int32_t round_d2i32( double v )
{
    if ( v < 0 ) return (int32_t)(v-0.5);
    else return (int32_t)(v+0.5);
}
inline static int64_t round_d2i64( double v )
{
    if ( v < 0 ) return (int64_t)(v-0.5);
    else return (int64_t)(v+0.5);
}

inline static long round_f2l( float v )
{
    if ( v < 0 ) return (long)(v-0.5);
    else return (long)(v+0.5);
}
inline static long long round_f2ll( float v )
{
    if ( v < 0 ) return (long long)(v-0.5);
    else return (long long)(v+0.5);
}
inline static long round_d2l( double v )
{
    if ( v < 0 ) return (long)(v-0.5);
    else return (long)(v+0.5);
}
inline static long long round_d2ll( double v )
{
    if ( v < 0 ) return (long long)(v-0.5);
    else return (long long)(v+0.5);
}

/* */

inline static double get_min_for_type( int type )
{
    if ( type == FITS::BYTE_T ) return MIN_DOUBLE_ROUND_UCHAR;
    else if ( type == FITS::SHORT_T ) return MIN_DOUBLE_ROUND_INT16;
    else if ( type == FITS::LONG_T ) return MIN_DOUBLE_ROUND_INT32;
    else if ( type == FITS::LONGLONG_T ) return MIN_DOUBLE_ROUND_INT64;
    else if ( type == FITS::LOGICAL_T ) return MIN_DOUBLE_ROUND_UCHAR;
    else return MIN_DOUBLE_ROUND_LLONG;
}

inline static double get_max_for_type( int type )
{
    if ( type == FITS::BYTE_T ) return MAX_DOUBLE_ROUND_UCHAR;
    else if ( type == FITS::SHORT_T ) return MAX_DOUBLE_ROUND_INT16;
    else if ( type == FITS::LONG_T ) return MAX_DOUBLE_ROUND_INT32;
    else if ( type == FITS::LONGLONG_T ) return MAX_DOUBLE_ROUND_INT64;
    else if ( type == FITS::LOGICAL_T ) return MAX_DOUBLE_ROUND_UCHAR;
    else return MAX_DOUBLE_ROUND_LLONG;
}

/**
 * @brief  ̤֤ͤ ()
 *
 * @note    private ʴؿǤ
 */
inline static long long get_indef_for_type( int type )
{
    if ( type == FITS::BYTE_T ) return INDEF_UCHAR;
    else if ( type == FITS::SHORT_T ) return INDEF_INT16;
    else if ( type == FITS::LONG_T ) return INDEF_INT32;
    else if ( type == FITS::LONGLONG_T ) return INDEF_INT64;
    else if ( type == FITS::BIT_T ) return 0;
    else if ( type == FITS::LOGICAL_T ) return '\0';
    else return INDEF_LLONG;
}

/**
 * @brief  񤭽Ф tnull ֤ͤ ()
 *
 * @note    private ʴؿǤ
 */
inline static long long get_tnull_for_type( int type, long long tnull )
{
    if ( type == FITS::BYTE_T ) {
	if ( MIN_UCHAR <= tnull && tnull <= MAX_UCHAR ) return tnull;
	else return INDEF_UCHAR;
    }
    else if ( type == FITS::SHORT_T ) {
	if ( MIN_INT16 <= tnull && tnull <= MAX_INT16 ) return tnull;
	else return INDEF_INT16;
    }
    else if ( type == FITS::LONG_T ) {
	if ( MIN_INT32 <= tnull && tnull <= MAX_INT32 ) return tnull;
	else return INDEF_INT32;
    }
    else if ( type == FITS::LONGLONG_T ) {
	if ( MIN_INT64 <= tnull && tnull <= MAX_INT64 ) return tnull;
	else return INDEF_INT64;
    }
    else if ( type == FITS::BIT_T ) {
	return 0;
    }
    else if ( type == FITS::LOGICAL_T ) {
	if ( MIN_UCHAR <= tnull && tnull <= MAX_UCHAR ) return tnull;
	else return '\0';
    }
    else return tnull;
}

/* constructor */
/**
 * @brief  󥹥ȥ饯
 */
fits_table_col::fits_table_col()
{
    c_memset(&(this->def_all_rec),0,sizeof(this->def_all_rec));
    c_memset(&(this->def_rec),0,sizeof(this->def_rec));

    this->type_rec = FITS::ANY_T;
    this->bytes_rec = 0;
    this->elem_size_rec = 0;
    this->dcol_size_rec = 0;
    this->full_bytes_rec = 0;
    this->heap_type_rec = FITS::ANY_T;
    this->vl_max_length_rec = -1;
    this->tzero_rec = 0;
    this->tscal_rec = 1;
    this->tnull_is_set_rec = false;
    this->set_tnull(NULL);

    this->protected_rec = false;
    this->manager = NULL;

    this->row_size_rec = 0;

    this->data_rec.init(1,false);
    this->data_rec.set_auto_resize(false);
    this->data_rec.register_extptr(&this->data_ptr_rec);	/* [void **] */

    this->heap_rec.init(1,false);
    this->heap_rec.set_auto_resize(false);
    this->heap_rec.register_extptr(&this->heap_ptr_rec);	/* [void **] */

    this->bit_size_telem.init(sizeof(int), true);
    this->str_buf = NULL;
    this->tmp_str_buf = NULL;
    this->fmt_str = NULL;
    this->fmt_nullstr = NULL;
    this->null_svalue_rec = "NULL";

    this->shallow_copy_ok = false;
    this->shallow_copy_dest_obj = NULL;
    this->shallow_copy_src_obj = NULL;
    this->__copying = false;

    return;
}

/* copy constructor */
/**
 * @brief  ԡ󥹥ȥ饯
 */
fits_table_col::fits_table_col(const fits_table_col &obj)
{
    c_memset(&(this->def_all_rec),0,sizeof(this->def_all_rec));
    c_memset(&(this->def_rec),0,sizeof(this->def_rec));

    this->type_rec = FITS::ANY_T;
    this->bytes_rec = 0;
    this->elem_size_rec = 0;
    this->dcol_size_rec = 0;
    this->full_bytes_rec = 0;
    this->heap_type_rec = FITS::ANY_T;
    this->vl_max_length_rec = -1;
    this->tzero_rec = 0;
    this->tscal_rec = 1;
    this->tnull_is_set_rec = false;
    this->set_tnull(NULL);

    this->protected_rec = false;
    this->manager = NULL;

    this->row_size_rec = 0;

    this->data_rec.init(1,false);
    this->data_rec.set_auto_resize(false);
    this->data_rec.register_extptr(&this->data_ptr_rec);	/* [void **] */

    this->heap_rec.init(1,false);
    this->heap_rec.set_auto_resize(false);
    this->heap_rec.register_extptr(&this->heap_ptr_rec);	/* [void **] */

    this->bit_size_telem.init(sizeof(int), true);
    this->str_buf = NULL;
    this->tmp_str_buf = NULL;
    this->fmt_str = NULL;
    this->fmt_nullstr = NULL;
    this->null_svalue_rec = "NULL";

    this->shallow_copy_ok = false;
    this->shallow_copy_dest_obj = NULL;
    this->shallow_copy_src_obj = NULL;
    this->__copying = false;

    this->init(obj);

    return;
}

/* destructor */
/**
 * @brief  ǥȥ饯
 */
fits_table_col::~fits_table_col()
{
    if ( this->str_buf != NULL ) {
	delete this->str_buf;
    }
    if ( this->tmp_str_buf != NULL ) {
	delete this->tmp_str_buf;
    }
    return;
}

/*
 * public
 */

/* initialize */
/**
 * @brief  ֥ȤΥԡ
 *
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::operator=(const fits_table_col &obj)
{
    this->init(obj);
    return *this;
}

/**
 * @brief  ֥Ȥν
 *
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::init()
{
    if ( this->manager != NULL ) {
	//const char *none[] = {NULL};
	//fits::table_def def = { NULL,"",none,none,
	//			"","","","1A", "","","","" };
	fits::table_def def = { NULL,"","","",
				"","","","1A", "","","","" };
	/* manager ͳ _define Ƥ */
	this->define(def);
	/* 줤äѤˤ */
	this->clean(0, this->length());
	/* svalue()  NULL ͤꥻå */
	this->assign_null_svalue(this->manager->default_null_svalue());
	return *this;
    }

    c_memset(&(this->def_all_rec),0,sizeof(this->def_all_rec));
    c_memset(&(this->def_rec),0,sizeof(this->def_rec));

    this->tany = NULL;
    this->talas = NULL;
    this->telem = NULL;
    this->telem_def = NULL;

    this->elem_index_rec.init();

    this->type_rec = FITS::ANY_T;
    this->bytes_rec = 0;
    this->elem_size_rec = 0;
    this->dcol_size_rec = 0;
    this->full_bytes_rec = 0;
    this->heap_type_rec = FITS::ANY_T;
    this->vl_max_length_rec = -1;
    this->tzero_rec = 0;
    this->tscal_rec = 1;
    this->tnull_is_set_rec = false;
    this->set_tnull(NULL);

    if ( this->manager == NULL ) {
	this->protected_rec = false;
    }

    this->row_size_rec = 0;

    this->bit_size_telem.init(sizeof(int), true);

    if ( this->str_buf != NULL ) {
	delete this->str_buf;
	this->str_buf = NULL;
    }
    if ( this->tmp_str_buf != NULL ) {
	delete this->tmp_str_buf;
	this->tmp_str_buf = NULL;
    }
    this->fmt_str = NULL;
    this->fmt_nullstr = NULL;
    this->null_svalue_rec = "NULL";

    this->data_rec.init(1,false);
    this->data_rec.set_auto_resize(false);

    this->heap_rec.init(1,false);
    this->heap_rec.set_auto_resize(false);

    return *this;
}

/**
 * @brief  ֥ȤΥԡ
 *
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::init(const fits_table_col &obj)
{
    long obj_full_bytes;
    long org_nrows = this->length();

    if ( &obj == this ) return *this;

    if ( this->manager == NULL ) {
	this->fits_table_col::init();
    }
    else {
	/* ̾TALAS ʤɤν manager ˤ */
	/* manager ͳ _define() Ƥ */
	this->define(obj.definition());
    }

    obj_full_bytes = obj.full_bytes_rec;

    try {
	this->data_rec.init(obj.data_rec);
    }
    catch (...) {
	err_throw(__FUNCTION__,"FATAL","this->data_rec.init() failed");
    }

    try {
	this->heap_rec.init(obj.heap_rec);
    }
    catch (...) {
	this->data_rec.init(1,false);
	err_throw(__FUNCTION__,"FATAL","this->heap_rec.init() failed");
    }

    if ( obj.str_buf != NULL ) {
	try {
	    this->str_buf = new tstring(obj.str_buf->max_length());
	}
	catch (...) {
	    this->data_rec.init(1,false);
	    this->heap_rec.init(1,false);
	    err_throw(__FUNCTION__,"FATAL","new failed");
	}
	try {
	    this->tmp_str_buf = new tstring(obj.tmp_str_buf->max_length());
	}
	catch (...) {
	    if ( this->str_buf != NULL ) delete this->str_buf;
	    this->str_buf = NULL;
	    this->data_rec.init(1,false);
	    this->heap_rec.init(1,false);
	    err_throw(__FUNCTION__,"FATAL","new failed");
	}
	this->str_buf->assign(obj.str_buf->cstr());
    }

    if ( 0 < obj.bit_size_telem.length() ) {
	try {
	    this->bit_size_telem.init(obj.bit_size_telem);
	}
	catch (...) {
	    if ( this->tmp_str_buf != NULL ) delete this->tmp_str_buf;
	    this->tmp_str_buf = NULL;
	    if ( this->str_buf != NULL ) delete this->str_buf;
	    this->str_buf = NULL;
	    this->data_rec.init(1,false);
	    this->heap_rec.init(1,false);
	    err_throw(__FUNCTION__,"FATAL","bit_size_telem.init() failed");
	}
    }

    try {
	this->fmt_str = obj.fmt_str;
	this->fmt_nullstr = obj.fmt_nullstr;
	this->null_svalue_rec = obj.null_svalue_rec;
	this->tany = obj.tany;
	this->talas = obj.talas;
	this->telem = obj.telem;
	this->telem_def = obj.telem_def;
	this->elem_index_rec = obj.elem_index_rec;
    }
    catch (...) {
	this->elem_index_rec.init();
	this->telem = NULL;
	this->telem_def = NULL;
	this->talas = NULL;
	this->tany = NULL;
	this->bit_size_telem.init(sizeof(int), true);
	if ( this->str_buf != NULL ) delete this->str_buf;
	this->str_buf = NULL;
	if ( this->tmp_str_buf != NULL ) delete this->tmp_str_buf;
	this->tmp_str_buf = NULL;
	this->data_rec.init(1,false);
	this->heap_rec.init(1,false);
	err_throw(__FUNCTION__,"FATAL","'=' failed");
    }

    this->update_def_rec_ptr();

    this->type_rec = obj.type_rec;
    this->bytes_rec = obj.bytes_rec;
    this->elem_size_rec = obj.elem_size_rec;
    this->dcol_size_rec = obj.dcol_size_rec;
    this->full_bytes_rec = obj.full_bytes_rec;
    this->heap_type_rec = obj.heap_type_rec;
    this->vl_max_length_rec = obj.vl_max_length_rec;
    this->tzero_rec = obj.tzero_rec;
    this->tscal_rec = obj.tscal_rec;
    this->tnull_is_set_rec = obj.tnull_is_set_rec;
    this->tnull_r_rec = obj.tnull_r_rec;
    this->tnull_w_rec = obj.tnull_w_rec;
    this->tnull_longlong_rec = obj.tnull_longlong_rec;
    this->tnull_long_rec = obj.tnull_long_rec;
    this->tnull_short_rec = obj.tnull_short_rec;
    this->tnull_byte_rec = obj.tnull_byte_rec;

    if ( this->manager == NULL ) {
	this->protected_rec = obj.protected_rec;
    }

    this->row_size_rec = obj.row_size_rec;

    if ( this->manager != NULL ) {	/* fits_table δξ */
	/* Υ礭᤹ */
	this->_resize(org_nrows);
	/* Ū _resize() Ƥ(ȥơ֥ꥵ) */
	//this->manager->resize_rows(obj.row_size_rec);
    }

    return *this;
}

/* swap internal data of two objects */
/**
 * @brief  2ĤΥ֥ȴ֤ǤƤΥå
 *
 *  ꤵ줿֥obj ƤȼȤƤؤޤ
 *  ǡ󡦥إåơ°٤Ƥξ֤ؤޤ
 *
 * @param   obj fits_table_col 饹Υ֥
 * @return  Ȥλ    
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::swap(fits_table_col &obj)
{
    fits::table_def_all def;
    tstring ttype1;
    //tarray_tstring talas1;
    tstring talas1;
    tstring ttype2;
    //tarray_tstring talas2;
    tstring talas2;

    if ( &obj == this ) return *this;

    c_memset(&def,0,sizeof(def));

    if ( this->manager != NULL ) {
	ttype2 = obj.tany.at(TTYPE_IDX);
	//talas2 = obj.talas;
	talas2 = obj.tany.cstr(TALAS_IDX);
    }
    if ( obj.manager != NULL ) {
	ttype1 = this->tany.at(TTYPE_IDX);
	//talas1 = this->talas;
	talas1 = this->tany.cstr(TALAS_IDX);
    }

    /* ޤ manager ¦򹹿 */
    if ( this->manager != NULL ) {
	def.ttype = ttype2.cstr();
	//def.talas = talas2.cstrarray();
	def.talas = talas2.cstr();
	/* manager ͳ _define Ƥ */
	this->define(def);
    }
    if ( obj.manager != NULL ) {
	def.ttype = ttype1.cstr();
	//def.talas = talas1.cstrarray();
	def.talas = talas1.cstr();
	/* manager ͳ _define Ƥ */
	obj.define(def);
    }

    this->tany.swap(obj.tany);
    this->talas.swap(obj.talas);
    this->telem.swap(obj.telem);
    this->telem_def.swap(obj.telem_def);
    this->elem_index_rec.swap(obj.elem_index_rec);

    tstring *tmp__str_buf;
    tstring *tmp__tmp_str_buf;

    this->bit_size_telem.swap(obj.bit_size_telem);

    tmp__str_buf = this->str_buf;
    this->str_buf = obj.str_buf;
    obj.str_buf = tmp__str_buf;

    tmp__tmp_str_buf = this->tmp_str_buf;
    this->tmp_str_buf = obj.tmp_str_buf;
    obj.tmp_str_buf = tmp__tmp_str_buf;

    this->fmt_str.swap(obj.fmt_str);
    this->fmt_nullstr.swap(obj.fmt_nullstr);
    this->null_svalue_rec.swap(obj.null_svalue_rec);

    int tmp__type_rec;
    long tmp__bytes_rec;
    long tmp__elem_size_rec;
    long tmp__dcol_size_rec;
    long tmp__full_bytes_rec;
    int tmp__heap_type_rec;
    long tmp__vl_max_length_rec;
    double tmp__tzero_rec;
    double tmp__tscal_rec;
    bool tmp__tnull_is_set_rec;
    long long tmp__tnull_r_rec;
    long long tmp__tnull_w_rec;
    long long tmp__tnull_longlong_rec;
    long tmp__tnull_long_rec;
    short tmp__tnull_short_rec;
    unsigned char tmp__tnull_byte_rec;
    long tmp__row_size_rec;

    tmp__type_rec = this->type_rec;
    this->type_rec = obj.type_rec;
    obj.type_rec = tmp__type_rec;

    tmp__bytes_rec = this->bytes_rec;
    this->bytes_rec = obj.bytes_rec;
    obj.bytes_rec = tmp__bytes_rec;

    tmp__elem_size_rec = this->elem_size_rec;
    this->elem_size_rec = obj.elem_size_rec;
    obj.elem_size_rec = tmp__elem_size_rec;

    tmp__dcol_size_rec = this->dcol_size_rec;
    this->dcol_size_rec = obj.dcol_size_rec;
    obj.dcol_size_rec = tmp__dcol_size_rec;

    tmp__full_bytes_rec = this->full_bytes_rec;
    this->full_bytes_rec = obj.full_bytes_rec;
    obj.full_bytes_rec = tmp__full_bytes_rec;

    tmp__heap_type_rec = this->heap_type_rec;
    this->heap_type_rec = obj.heap_type_rec;
    obj.heap_type_rec = tmp__heap_type_rec;

    tmp__vl_max_length_rec = this->vl_max_length_rec;
    this->vl_max_length_rec = obj.vl_max_length_rec;
    obj.vl_max_length_rec = tmp__vl_max_length_rec;

    tmp__tzero_rec = this->tzero_rec;
    this->tzero_rec = obj.tzero_rec;
    obj.tzero_rec = tmp__tzero_rec;

    tmp__tscal_rec = this->tscal_rec;
    this->tscal_rec = obj.tscal_rec;
    obj.tscal_rec = tmp__tscal_rec;

    tmp__tnull_is_set_rec = this->tnull_is_set_rec;
    this->tnull_is_set_rec = obj.tnull_is_set_rec;
    obj.tnull_is_set_rec = tmp__tnull_is_set_rec;

    tmp__tnull_r_rec = this->tnull_r_rec;
    this->tnull_r_rec = obj.tnull_r_rec;
    obj.tnull_r_rec = tmp__tnull_r_rec;

    tmp__tnull_w_rec = this->tnull_w_rec;
    this->tnull_w_rec = obj.tnull_w_rec;
    obj.tnull_w_rec = tmp__tnull_w_rec;

    tmp__tnull_longlong_rec = this->tnull_longlong_rec;
    this->tnull_longlong_rec = obj.tnull_longlong_rec;
    obj.tnull_longlong_rec = tmp__tnull_longlong_rec;

    tmp__tnull_long_rec = this->tnull_long_rec;
    this->tnull_long_rec = obj.tnull_long_rec;
    obj.tnull_long_rec = tmp__tnull_long_rec;

    tmp__tnull_short_rec = this->tnull_short_rec;
    this->tnull_short_rec = obj.tnull_short_rec;
    obj.tnull_short_rec = tmp__tnull_short_rec;

    tmp__tnull_byte_rec = this->tnull_byte_rec;
    this->tnull_byte_rec = obj.tnull_byte_rec;
    obj.tnull_byte_rec = tmp__tnull_byte_rec;

    tmp__row_size_rec = this->row_size_rec;
    this->row_size_rec = obj.row_size_rec;
    obj.row_size_rec = tmp__row_size_rec;

    this->data_rec.swap(obj.data_rec);

    this->heap_rec.swap(obj.heap_rec);

    this->update_def_rec_ptr();
    obj.update_def_rec_ptr();

    bool tmp__protected_rec;
    tmp__protected_rec = this->protected_rec;
    if ( this->manager == NULL ) {
	this->protected_rec = obj.protected_rec;
    }
    if ( obj.manager == NULL ) {
	obj.protected_rec = tmp__protected_rec;
    }

    if ( this->manager != NULL ) {
	def.ttype = ttype2.cstr();
	//def.talas = talas2.cstrarray();
	def.talas = talas2.cstr();
	this->_define(def);
    }
    if ( obj.manager != NULL ) {
	def.ttype = ttype1.cstr();
	//def.talas = talas1.cstrarray();
	def.talas = talas1.cstr();
	obj._define(def);
    }

    return *this;
}

/* 
 * Functions to obtain column information
 */

/**
 * @brief  ݸ°̵֤ͭ
 *
 * @note   桼γĥ饹ǻѤꡥ
 */
bool fits_table_col::is_protected() const
{
    return this->protected_rec;
}

/* to obtain all definition of a column */
/**
 * @brief   fits::table_def ¤ΤǼ
 */
const fits::table_def &fits_table_col::definition() const
{
    return this->def_rec;
}

/**
 * @brief   fits::table_def_all ¤ΤǼ
 */
const fits::table_def_all &fits_table_col::definition_all() const
{
    return this->def_all_rec;
}

/* get/set column name */
/**
 * @brief  ̾
 */
const char *fits_table_col::name() const
{
    return this->tany.cstr(TTYPE_IDX);
}

/**
 * @brief  ̾
 */
fits_table_col &fits_table_col::assign_name( const char *new_name )
{
    fits::table_def_all def = {NULL};
    def.ttype = new_name;
    this->define(def);
    return *this;
}

/**
 * @brief  ĹޤϲĹǿ
 *
 *  Ĺξ硤ͤ˴طʤ fits_table_col::elem_length() ֤
 *  Ʊ֤ͤޤ<br>
 *  elem_idx ϡĹ󵭽һҤʣ¸ߤ˻ꤷޤ
 *  㤨 TFORM  '4PE(999)' ξ elem_idx  0  3 ͤȤ
 *  Ǥޤ
 *
 * @param   row_idx ԥǥå
 * @param   elem_idx Ĺ󵭽һҤֹ (ά0)
 * @return  ǿ
 */
long fits_table_col::array_length( long row_idx, long elem_idx ) const
{
    long elem_len = this->elem_size_rec;
    if ( this->type_rec == FITS::LONGARRDESC_T ) {
      if ( this->data_ptr_rec == NULL ) return -1;
      else {
	  long ix = (elem_len * row_idx) + (elem_idx % elem_len);
	  return 
	      ((const fits::longarrdesc_t *)(this->data_ptr_rec))[ix].length;
      }
    }
    else if ( this->type_rec == FITS::LLONGARRDESC_T ) {
      if ( this->data_ptr_rec == NULL ) return -1;
      else {
	  long ix = (elem_len * row_idx) + (elem_idx % elem_len);
	  return 
	      ((const fits::llongarrdesc_t *)(this->data_ptr_rec))[ix].length;
      }
    }
    else return elem_len;
}

/**
 * @brief  ĹΥҡץեåȤ
 *
 *  elem_idx ϡĹ󵭽һҤʣ¸ߤ˻ꤷޤ
 *  㤨 TFORM  '4PE(999)' ξ elem_idx  0  3 ͤȤ
 *  Ǥޤ
 *
 * @param   row_idx ԥǥå
 * @param   elem_idx Ĺ󵭽һҤֹ (ά0)
 * @return  ҡץեå (Хñ)
 */
long fits_table_col::array_heap_offset( long row_idx, long elem_idx ) const
{
    long elem_len = this->elem_size_rec;
    if ( this->type_rec == FITS::LONGARRDESC_T ) {
      if ( this->data_ptr_rec == NULL ) return -1;
      else {
	  long ix = (elem_len * row_idx) + (elem_idx % elem_len);
	  return 
	      ((const fits::longarrdesc_t *)(this->data_ptr_rec))[ix].offset;
      }
    }
    else if ( this->type_rec == FITS::LLONGARRDESC_T ) {
      if ( this->data_ptr_rec == NULL ) return -1;
      else {
	  long ix = (elem_len * row_idx) + (elem_idx % elem_len);
	  return 
	      ((const fits::llongarrdesc_t *)(this->data_ptr_rec))[ix].offset;
      }
    }
    else return -1;
}


/*
 * NULL value to I/O of high-level APIs.  Default is "NULL".
 */

/**
 * @brief  NULLʸ
 *
 *  fits_table_col::svalue()fits_table_col::assign( const char *, ... ) 
 *  Ѥ NULL ʸͤꤷޤ<br>
 *  NULLʸνͤ "NULL" Ǥ
 *
 * @param   snull NULLʸ
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::assign_null_svalue( const char *snull )
{
    if ( snull != NULL ) {
	this->null_svalue_rec.assign(snull);
	/* */
	if ( this->str_buf != NULL ) {
	    size_t org_len = this->str_buf->max_length();
	    if ( org_len < this->null_svalue_rec.length() ) {
		size_t len_alc = this->null_svalue_rec.length();
		delete this->str_buf;
		this->str_buf = NULL;
		try {
		    this->str_buf = new tstring(len_alc);
		}
		catch (...) {
		    err_throw(__FUNCTION__,"FATAL","new failed");
		}
	    }
	}
	if ( this->tmp_str_buf != NULL ) {
	    size_t org_len = this->tmp_str_buf->max_length();
	    if ( org_len < this->null_svalue_rec.length() ) {
		size_t len_alc = this->null_svalue_rec.length();
		delete this->tmp_str_buf;
		this->tmp_str_buf = NULL;
		try {
		    this->tmp_str_buf = new tstring(len_alc);
		}
		catch (...) {
		    err_throw(__FUNCTION__,"FATAL","new failed");
		}
	    }
	}
    }
    return *this;
}

/**
 * @brief  NULLʸμ
 *
 *  fits_table_col::svalue()fits_table_col::assign( const char *, ... ) 
 *  Ѥ NULL ʸͤޤ
 *
 * @return  NULLʸ
 */
const char *fits_table_col::null_svalue() const
{
    return this->null_svalue_rec.cstr();
}

/*
 * Set default value when resizing
 * Note: default value will be resetted when changing attributes of column
 */

/* high-level */
/**
 * @brief  Υǥեͤ double 
 *
 *  fits_table::resize_rows() ǥơ֥ιԤĹ礭ο
 *  ͤꤷޤΥдؿǤϥإå TZEROnTSCALEnTNULLn ͤ
 *  ȿǤޤ<br>
 *  ǥեͤνͤϡξ 'F'ʸ󷿤ξ ' ' 
 *  ʤʸ󡤤¾ηξ 0 Ǥ<br>
 *  ǥեͤȤ NULLͤ򥻥åȤ NAN 򥻥åȤޤ
 *
 * @param   value 
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_default( double value )
{
    fits_table_col dummy;
    if ( this->heap_is_used() == true ) goto quit;

    dummy.define(this->definition());
    dummy.resize(1);
    dummy.assign(value, 0);

    this->data_rec.assign_default((const void *)(dummy.data_ptr_cs()));

 quit:
    return *this;
}

/**
 * @brief  Υǥեͤ float 
 *
 *  fits_table::resize_rows() ǥơ֥ιԤĹ礭ο
 *  ͤꤷޤΥдؿǤϥإå TZEROnTSCALEnTNULLn ͤ
 *  ȿǤޤ<br>
 *  ǥեͤνͤϡξ 'F'ʸ󷿤ξ ' ' 
 *  ʤʸ󡤤¾ηξ 0 Ǥ<br>
 *  ǥեͤȤ NULLͤ򥻥åȤ NAN 򥻥åȤޤ
 *
 * @param   value 
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_default( float value )
{
    fits_table_col dummy;
    if ( this->heap_is_used() == true ) goto quit;

    dummy.define(this->definition());
    dummy.resize(1);
    dummy.assign(value, 0);

    this->data_rec.assign_default((const void *)(dummy.data_ptr_cs()));

 quit:
    return *this;
}

/**
 * @brief  Υǥեͤ long long 
 *
 *  fits_table::resize_rows() ǥơ֥ιԤĹ礭ο
 *  ͤꤷޤΥдؿǤϥإå TZEROnTSCALEnTNULLn ͤ
 *  ȿǤޤ<br>
 *  ǥեͤνͤϡξ 'F'ʸ󷿤ξ ' ' 
 *  ʤʸ󡤤¾ηξ 0 Ǥ<br>
 *  ǥեͤȤ NULLͤ򥻥åȤ 
 *  fits_table_col::assign_default(double)  NAN 򥻥åȤޤ
 *
 * @param   value 
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_default( long long value )
{
    fits_table_col dummy;
    if ( this->heap_is_used() == true ) goto quit;

    dummy.define(this->definition());
    dummy.resize(1);
    dummy.assign(value, 0);

    this->data_rec.assign_default((const void *)(dummy.data_ptr_cs()));

 quit:
    return *this;
}

/**
 * @brief  Υǥեͤ long 
 *
 *  fits_table::resize_rows() ǥơ֥ιԤĹ礭ο
 *  ͤꤷޤΥдؿǤϥإå TZEROnTSCALEnTNULLn ͤ
 *  ȿǤޤ<br>
 *  ǥեͤνͤϡξ 'F'ʸ󷿤ξ ' ' 
 *  ʤʸ󡤤¾ηξ 0 Ǥ<br>
 *  ǥեͤȤ NULLͤ򥻥åȤ 
 *  fits_table_col::assign_default(double)  NAN 򥻥åȤޤ
 *
 * @param   value 
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_default( long value )
{
    fits_table_col dummy;
    if ( this->heap_is_used() == true ) goto quit;

    dummy.define(this->definition());
    dummy.resize(1);
    dummy.assign(value, 0);

    this->data_rec.assign_default((const void *)(dummy.data_ptr_cs()));

 quit:
    return *this;
}

/**
 * @brief  Υǥեͤ int 
 *
 *  fits_table::resize_rows() ǥơ֥ιԤĹ礭ο
 *  ͤꤷޤΥдؿǤϥإå TZEROnTSCALEnTNULLn ͤ
 *  ȿǤޤ<br>
 *  ǥեͤνͤϡξ 'F'ʸ󷿤ξ ' ' 
 *  ʤʸ󡤤¾ηξ 0 Ǥ<br>
 *  ǥեͤȤ NULLͤ򥻥åȤ 
 *  fits_table_col::assign_default(double)  NAN 򥻥åȤޤ
 *
 * @param   value 
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_default( int value )
{
    fits_table_col dummy;
    if ( this->heap_is_used() == true ) goto quit;

    dummy.define(this->definition());
    dummy.resize(1);
    dummy.assign(value, 0);

    this->data_rec.assign_default((const void *)(dummy.data_ptr_cs()));

 quit:
    return *this;
}

/**
 * @brief  Υǥեͤ char * 
 *
 *  fits_table::resize_rows() ǥơ֥ιԤĹ礭ο
 *  ͤꤷޤΥдؿǤϥإå TZEROnTSCALEnTNULLn ͤ
 *  ȿǤޤ<br>
 *  ǥեͤνͤϡξ 'F'ʸ󷿤ξ ' ' 
 *  ʤʸ󡤤¾ηξ 0 Ǥ<br>
 *  ǥեͤȤ NULLͤ򥻥åȤ 
 *  fits_table_col::assign_default(double)  NAN 򥻥åȤޤ
 *
 * @param   value 
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_default( const char *value )
{
    fits_table_col dummy;
    if ( this->heap_is_used() == true ) goto quit;

    dummy.define(this->definition());
    dummy.resize(1);
    dummy.assign(value, 0);

    this->data_rec.assign_default((const void *)(dummy.data_ptr_cs()));

 quit:
    return *this;
}

/* low-level */
/**
 * @brief  Υǥեͤ (٥)
 *
 *  fits_table::resize_rows() ǥơ֥ιԤĹ礭ο
 *  ͤꤷޤ٥ʥдؿǤΤǡإå TZEROnTSCALEn
 *  TNULLn ͤϹθޤ<br>
 *  ǥեͤνͤϡξ 'F'ʸ󷿤ξ ' ' 
 *  ʤʸ󡤤¾ηξ 0 Ǥ
 *
 * @param   value ͤΥɥ쥹
 * @note    value ؤƤΥХȿϡηΥХȿʤ
 *          fits_table_col::bytes() ֤͡פ˰פƤɬפޤ
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_default_value( const void *value )
{
    if ( this->heap_is_used() == true ) goto quit;

    this->data_rec.assign_default(value);

 quit:
    return *this;
}

/* */

/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
void *fits_table_col::heap_ptr()
{
    if ( this->manager != NULL ) return NULL;		      /* not enabled */

    return this->heap_ptr_rec;
}

#ifdef SLI__OVERLOAD_CONST_AT
/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
const void *fits_table_col::heap_ptr() const
{
    if ( this->manager != NULL ) return NULL;		      /* not enabled */

    return this->heap_ptr_rec;
}
#endif

/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
const void *fits_table_col::heap_ptr_cs() const
{
    if ( this->manager != NULL ) return NULL;		      /* not enabled */

    return this->heap_ptr_rec;
}

/*
 * Get or put raw data using C-like I/O args
 */

/**
 * @brief  ǡХåեإԡ
 *
 * @param   dest_buf ԡХåեΥɥ쥹
 * @param   buf_size dest_buf ΥХȥ
 * @return  : ХåեĹʬʾ˥ԡǤХȿ <br>
 *          (顼): ǥԡʤä
 * @attention  ĹϥݡȤƤޤ
 */
ssize_t fits_table_col::get_data( void *dest_buf, size_t buf_size ) const
{
    return this->get_data(0,dest_buf,buf_size);
}

/**
 * @brief  ǡХåեإԡ
 *
 * @param   row_index  ԡ(Хåե)ιԥǥå
 * @param   dest_buf ԡХåեΥɥ쥹
 * @param   buf_size dest_buf ΥХȥ
 * @return  : ХåեĹʬʾ˥ԡǤХȿ <br>
 *          (顼): ǥԡʤä
 * @attention  ĹϥݡȤƤޤ
 */
ssize_t fits_table_col::get_data( long row_index, 
				  void *dest_buf, size_t buf_size ) const
{
    if ( this->full_bytes_rec == 0 ) return -1;		/* not defined */

    return this->data_rec.getdata(dest_buf, buf_size, 0, row_index);
}

/**
 * @brief  ХåեΥǡ
 *
 * @param   src_buf ԡХåեΥɥ쥹
 * @param   buf_size src_buf ΥХȥ
 * @return  : src_bufΥХåեĹʬʾ˥ԡǤХȿ<br>
 *          : ǥԡʤä
 * @attention  ĹϥݡȤƤޤ
 */
ssize_t fits_table_col::put_data( const void *src_buf, size_t buf_size )
{
    return this->put_data(0,src_buf,buf_size);
}

/**
 * @brief  ХåեΥǡ
 *
 * @param   row_index ԡ(Хåե)ιԥǥå
 * @param   src_buf ԡХåեΥɥ쥹
 * @param   buf_size src_buf ΥХȥ
 * @return  : src_bufΥХåեĹʬʾ˥ԡǤХȿ<br>
 *          : ǥԡʤä
 * @attention  ĹϥݡȤƤޤ
 */
ssize_t fits_table_col::put_data( long row_index, 
				  const void *src_buf, size_t buf_size )
{
    if ( this->full_bytes_rec == 0 ) return -1;		/* not defined */

    return this->data_rec.putdata(src_buf, buf_size, 0, row_index);
}

/**
 * @brief  ǡƤ֥ȤλȤ֤ (ɼ)
 */
const mdarray &fits_table_col::data_array_cs() const
{
    return this->data_rec;
}

/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
ssize_t fits_table_col::get_heap( void *dest_buf, size_t buf_size ) const
{
    if ( this->manager != NULL ) return -1;		      /* not enabled */

    return this->get_heap(0,dest_buf,buf_size);
}

/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
ssize_t fits_table_col::get_heap( long offset, 
				  void *dest_buf, size_t buf_size ) const
{
    if ( this->manager != NULL ) return -1;		      /* not enabled */

    return this->heap_rec.getdata(dest_buf, buf_size, offset);
}

/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
ssize_t fits_table_col::put_heap( const void *buf, size_t buf_size )
{
    if ( this->manager != NULL ) return -1;		      /* not enabled */

    return this->put_heap(0,buf,buf_size);
}

/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
ssize_t fits_table_col::put_heap( long offset, 
				  const void *buf, size_t buf_size )
{
    if ( this->manager != NULL ) return -1;		      /* not enabled */

    return this->heap_rec.putdata(buf, buf_size, offset);
}

/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
const mdarray &fits_table_col::heap_array_cs() const
{
    return this->heap_rec;
}

/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
fits_table_col &fits_table_col::resize_heap( size_t sz )
{
    this->heap_rec.resize(sz);
    return *this;
}

/**
 * @deprecated  μΤΥдؿѤʤǤ
 */
size_t fits_table_col::heap_length()
{
    return this->heap_rec.length();
}

/**
 * @brief  TFORMnTDIMn ѡγƼ֤ ()
 *
 * @note    private ʴؿǤ
 */
static int parse_tform_and_tdim( const char *tform_in, const char *tdim_in,
				 int *ret_type, long *ret_sz_bytes, 
				 long *ret_elem_size, 
				 long *ret_dcol_size,
				 long *ret_full_bytes,
				 int *ret_heap_type, long *ret_heap_sz_bytes,
				 long *ret_vl_max_length )
{
    int status = 0;
    ssize_t p0, p1, p2;
    size_t l0, l1, l2;
    long new_elem_size;
    tstring tform;
    tstring tdim;
    int *tgt_type;
    long *tgt_sz_bytes;
    int tgt_ch;

    if ( tform_in == NULL ) {
	status = -1;					/* for ERROR */
	goto quit;
    }

    if ( tdim_in != NULL ) tdim.assign(tdim_in);
    tform.assign(tform_in);

    /* find starting position */
    p0 = tform.strpbrk("0123456789LXIJKAEDBPQCM");
    if ( p0 != 0 ) status = 1;				/* for WARNING */
    else if ( p0 < 0 ) {
	/* unsupported type */
	status = -1;					/* for ERROR */
	goto quit;
    }

    /* get element length */
    l0 = tform.strspn(p0, "0123456789");
    if ( l0 == 0 ) {
	new_elem_size = 1;
    }
    else {
	size_t endpos;
	new_elem_size = tform.strtol(p0,10,&endpos);
	if ( new_elem_size < 1 ) {
	    new_elem_size = 1;
	    status = 1;					/* for WARNING */
	}
    }

    /* get type */
    p1 = p0 + l0;
    l1 = tform.strspn(p1, "LXIJKAEDBPQCM");
    if ( l1 == 0 ) {
	/* unsupported type */
	status = -1;					/* for ERROR */
	goto quit;
    }
    else l1 = 1;

    /* using heap */
    if ( tform.cchr(p1) == 'P' || tform.cchr(p1) == 'Q' ) {
	/* ȤꤢʸȥӥåȷϥݡȤȤ */
	p2 = p1 + l1;
	l2 = tform.strspn(p2, "LIJKEDBCM");

	if ( l2 == 0 ) {
	    /* unsupported type */
	    status = -1;				/* for ERROR */
	    goto quit;
	}

	if ( tform.cchr(p1) == 'P' ) {
	    *ret_type  = FITS::LONGARRDESC_T;
	    *ret_sz_bytes = sizeof(fits::longarrdesc_t);
	}
	else if ( tform.cchr(p1) == 'Q' ) {
	    *ret_type  = FITS::LLONGARRDESC_T;
	    *ret_sz_bytes = sizeof(fits::llongarrdesc_t);
	}

	/* change target for next switch() statement */
	tgt_type = ret_heap_type;
	tgt_sz_bytes = ret_heap_sz_bytes;
	tgt_ch = tform.cchr(p2);

	if ( tform.cchr(p2 + 1) == '(' ) {
	    *ret_vl_max_length = tform.atol(p2 + 2);
	}
	else {
	    *ret_vl_max_length = -1;
	}
    }
    /* not using heap */
    else {
	*ret_heap_type = FITS::ANY_T;
	*ret_heap_sz_bytes = 0;
	*ret_vl_max_length = -1;
	/* change target for next switch() statement */
	tgt_type = ret_type;
	tgt_sz_bytes = ret_sz_bytes;
	tgt_ch = tform.cchr(p1);
    }

    /* set *ret_type or *ret_heap_type */
    switch ( tgt_ch ) {
    case 'L':
	*tgt_type  = FITS::LOGICAL_T;
	*tgt_sz_bytes = sizeof(fits::logical_t);
	break;
    case 'X':
	*tgt_type  = FITS::BIT_T;
	*tgt_sz_bytes = sizeof(fits::bit_t);
	break;
    case 'I':
	*tgt_type  = FITS::SHORT_T;
	*tgt_sz_bytes = sizeof(fits::short_t);
	break;
    case 'J':
	*tgt_type  = FITS::LONG_T;
	*tgt_sz_bytes = sizeof(fits::long_t);
	break;
    case 'K':
	*tgt_type  = FITS::LONGLONG_T;
	*tgt_sz_bytes = sizeof(fits::longlong_t);
	break;
    case 'A':
	*tgt_type  = FITS::STRING_T;
	*tgt_sz_bytes = sizeof(fits::byte_t);	/* for now */
	break;
    case 'E':
	*tgt_type  = FITS::FLOAT_T;
	*tgt_sz_bytes = -sizeof(fits::float_t);
	break;
    case 'D':
	*tgt_type  = FITS::DOUBLE_T;
	*tgt_sz_bytes = -sizeof(fits::double_t);
	break;
    case 'B':
	*tgt_type  = FITS::BYTE_T;
	*tgt_sz_bytes = sizeof(fits::byte_t);
	break;
    case 'P':
	*tgt_type  = FITS::LONGARRDESC_T;
	*tgt_sz_bytes = sizeof(fits::longarrdesc_t);
	break;
    case 'Q':
	*tgt_type  = FITS::LLONGARRDESC_T;
	*tgt_sz_bytes = sizeof(fits::llongarrdesc_t);
	break;
    case 'C':
	*tgt_type  = FITS::COMPLEX_T;
	*tgt_sz_bytes = sizeof(fits::complex_t);
	break;
    case 'M':
	*tgt_type  = FITS::DOUBLECOMPLEX_T;
	*tgt_sz_bytes = sizeof(fits::doublecomplex_t);
	break;
    }

    *ret_elem_size = new_elem_size;
    *ret_dcol_size = new_elem_size;

    if ( *ret_type == FITS::STRING_T ) {
	ssize_t t_pos, d_pos = -1;
	size_t t_endpos, d_endpos;
	long sz, dx, d0=new_elem_size, d1=1, d2=1;
	t_pos = tform.strpbrk(p1+1,"[0-9]");	/* e.g., 120A10 */
	if ( 0 <= t_pos ) {
	    dx = tform.strtol(t_pos,10,&t_endpos);
	    if ( 0 < dx ) {
		d1 = d0 / dx;
		d0 = dx;
	    }
	    if ( 0 < tdim.length() ) {
		d_pos = tdim.strpbrk("[0-9]");
		if ( d_pos < 0 ) {
		    /* not found dim:0 */
		    status = 1;			/* for WARNING */
		}
	    }
	    else {
		d_pos = -1;
	    }
	}
	else if ( 0 < tdim.length() ) {
	    d_pos = tdim.strpbrk("[0-9]");
	    if ( 0 <= d_pos ) {
		dx = tdim.strtol(d_pos,10,&d_endpos);
		if ( 0 < dx ) {
		    d1 = d0 / dx;
		    d0 = dx;
		}
		d_pos = tdim.strpbrk(d_endpos,"[0-9]");
	    }
	    else {
		/* not found dim:0 */
		status = 1;			/* for WARNING */
	    }
	}
	*ret_sz_bytes = sizeof(fits::byte_t) * d0;
	*ret_dcol_size = d1;
	*ret_elem_size = d1;
	if ( 0 <= d_pos ) {
	    dx = tdim.strtol(d_pos,10,&d_endpos);
	    if ( 0 < dx ) {
		d2 = d1 / dx;
		d1 = dx;
	    }
	    *ret_dcol_size = d1;
	    d_pos = tdim.strpbrk(d_endpos,"[0-9]");
	    if ( 0 <= d_pos ) {
		dx = tdim.strtol(d_pos,10,&d_endpos);
		if ( dx != d2 ) {
		    status = 1;			/* for WARNING */
		}
	    }
	}
	/* å */
	if ( *ret_elem_size < *ret_dcol_size ||
	     *ret_elem_size % *ret_dcol_size != 0 ) {
	    *ret_dcol_size = *ret_elem_size;
	    status = 1;				/* for WARNING */
	}
	sz = sizeof(fits::byte_t) * new_elem_size; 
	if ( (*ret_sz_bytes) * (*ret_elem_size) != sz ) {
	    *ret_sz_bytes = sizeof(fits::byte_t) * new_elem_size;
	    *ret_elem_size = 1;
	    *ret_dcol_size = 1;
	    status = 1;				/* for WARNING */
	}
    }
    else {	/* STRING ǤϤʤ */
	if ( 0 < tdim.length() ) {
	    ssize_t d_pos;
	    size_t d_endpos;
	    d_pos = tdim.strpbrk("[0-9]");
	    if ( 0 <= d_pos ) {
		long dx, d0=new_elem_size, d1=1;
		dx = tdim.strtol(d_pos,10,&d_endpos);
		if ( 0 < dx ) {
		    d1 = d0 / dx;
		    d0 = dx;
		}
		*ret_dcol_size = d0;
		d_pos = tdim.strpbrk(d_endpos,"[0-9]");
		if ( 0 <= d_pos ) {
		    dx = tdim.strtol(d_pos,10,&d_endpos);
		    if ( dx != d1 ) {
			status = 1;		/* for WARNING */
		    }
		}
	    }
	    else {
		/* not found dim:0 */
		status = 1;			/* for WARNING */
	    }
	    /* å */
	    if ( *ret_elem_size < *ret_dcol_size ||
		 *ret_elem_size % *ret_dcol_size != 0 ) {
		*ret_dcol_size = *ret_elem_size;
		status = 1;			/* for WARNING */
	    }
	}
    }

    if ( *ret_type == FITS::BIT_T ) {
	if ( *ret_elem_size == 0 ) *ret_full_bytes = 0;
	else *ret_full_bytes = 1 + (*ret_elem_size - 1) / 8;
    }
    else {
	*ret_full_bytes = abs_sz2z(*ret_sz_bytes) * (*ret_elem_size);
    }

    /* debug */
#if 0
    err_report1(__FUNCTION__,"DEBUG","*ret_type = %d",*ret_type);
    err_report1(__FUNCTION__,"DEBUG","*ret_sz_bytes = %ld",*ret_sz_bytes);
    err_report1(__FUNCTION__,"DEBUG","*ret_elem_size = %ld",*ret_elem_size);
    err_report1(__FUNCTION__,"DEBUG","*ret_dcol_size = %ld",*ret_dcol_size);
    err_report1(__FUNCTION__,"DEBUG","*ret_full_bytes = %ld",*ret_full_bytes);
    err_report1(__FUNCTION__,"DEBUG","*ret_heap_type = %d",*ret_heap_type);
    err_report1(__FUNCTION__,"DEBUG","*ret_heap_sz_bytes = %ld",
		*ret_heap_sz_bytes);
#endif

 quit:
    return status;
}

/* fits_table ϡΥдؿȤʤȡ*/
/* change the definition of a column */
/**
 * @brief   (fits::table_def ¤Τ)
 *
 *  ΥдؿϡȡѹͤƤ
 *  def.xxx == NULL ξϡιܤѹޤ
 *
 * @param   def 
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::define( const fits::table_def &def )
{
    fits::table_def_all def_all;

    def_all.ttype         = def.ttype;
    def_all.ttype_comment = def.ttype_comment;
    def_all.talas         = def.talas;
    def_all.talas_comment = NULL;
    def_all.telem         = def.telem;
    def_all.telem_comment = NULL;
    def_all.tunit         = def.tunit;
    def_all.tunit_comment = def.tunit_comment;
    def_all.tdisp         = def.tdisp;
    def_all.tdisp_comment = NULL;
    def_all.tform         = def.tform;
    def_all.tform_comment = NULL;
    def_all.tdim          = def.tdim;
    def_all.tdim_comment  = NULL;
    def_all.tnull         = def.tnull;
    def_all.tnull_comment = NULL;
    def_all.tzero         = def.tzero;
    def_all.tzero_comment = NULL;
    def_all.tscal         = def.tscal;
    def_all.tscal_comment = NULL;

    return this->define(def_all);
}

/* fits_table ϡΥдؿȤʤȡ*/
/**
 * @brief   (fits::table_def_all ¤Τ)
 *
 *  ΥдؿϡȡѹͤƤ
 *  def.xxx == NULL ξϡιܤѹޤ
 *
 * @param   def 
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::define( const fits::table_def_all &def )
{
    if ( this->manager == NULL ) {	/* fits_table δǤʤ */
	return this->_define(def);
    }
    else {				/* fits_table δξ */
	long i;
	for ( i=0 ; i < this->manager->col_length() ; i++ ) {
	    if ( &(this->manager->col(i)) == this ) {
		/* Ū _define() Ƥ */
		this->manager->define_a_col(i, def);
		break;
	    }
	}
	if ( i == this->manager->col_length() ) {
	    err_throw(__FUNCTION__,"FATAL",
		      "internal error; fits_table does not manage me...");
	}
    }
    return *this;
}

/**
 * @brief   (٥)
 *
 *  ΥдؿϡȡѹͤƤ
 *  def.xxx == NULL ξϡιܤѹޤ
 *
 * @param   def 
 * @return  Ȥλ
 * @note  fits_table ϡΥдؿȤȡ<br>
 *        Υдؿ protected Ǥ
 */
fits_table_col &fits_table_col::_define( const fits::table_def_all &def )
{
    tstring tmp_str;
    int status;

    if ( this->is_protected() != false ) {
	err_report1(__FUNCTION__,"WARNING",
		    "This column [%s] is protected, cannot modify definitions",
		    this->tany.cstr(TTYPE_IDX));
	goto quit;
    }

    if ( def.tzero != NULL ) {
	tmp_str.assign(def.tzero).strtrim();
	this->tany.put(TZERO_IDX,tmp_str.cstr(),1);
	if ( 0 < tmp_str.length() ) this->tzero_rec = tmp_str.atof();
	else this->tzero_rec = 0;
	this->update_zero_of_data_rec();
    }

    if ( def.tzero_comment != NULL ) {
        tmp_str.assign(def.tzero_comment);
	this->tany.put(TZERO_COMMENT_IDX,tmp_str.cstr(),1);
    }

    if ( def.tscal != NULL ) {
	tmp_str.assign(def.tscal).strtrim();
	if ( 0 < tmp_str.length() ) {
	    if ( tmp_str.atof() == 0 ) {
		err_report(__FUNCTION__,"WARNING",
			   "ZERO TSCAL cannot be set, ignored.");
	    }
	    else {
		this->tany.put(TSCAL_IDX,tmp_str.cstr(),1);
		this->tscal_rec = tmp_str.atof();
	    }
	}
	else {
	    this->tany.put(TSCAL_IDX,"",1);
	    this->tscal_rec = 1;
	}
	this->update_zero_of_data_rec();
    }

    if ( def.tscal_comment != NULL ) {
        tmp_str.assign(def.tscal_comment);
	this->tany.put(TSCAL_COMMENT_IDX,tmp_str.cstr(),1);
    }

    if ( def.tform != NULL || def.tdim != NULL ) {
	int new_type = 0;
	long new_sz_bytes = 0;
	long new_elem_size = 0;
	long new_dcol_size = 0;
	long new_full_bytes = 0;
	int new_heap_type = 0;
	long new_heap_sz_bytes = 0;
	long new_vl_max_length = -1;
	tstring tmp_tform;
	tstring tmp_tdim;
	bool do_tnull_reset;
	if ( def.tform != NULL ) {
	    tmp_tform.assign(def.tform).strtrim();
	}
	else {
	    tmp_tform.assign(this->tany.cstr(TFORM_IDX));
	}

	if ( def.tdim != NULL ) {
	    tmp_tdim.assign(def.tdim).strtrim();
	}
	else {
	    tmp_tdim.assign(this->tany.cstr(TDIM_IDX));
	}
	status = parse_tform_and_tdim( tmp_tform.cstr(), tmp_tdim.cstr(),
				       &new_type, &new_sz_bytes, 
				       &new_elem_size, 
				       &new_dcol_size,
				       &new_full_bytes,
				       &new_heap_type, &new_heap_sz_bytes,
				       &new_vl_max_length );
	if ( status < 0 ) {
	    err_throw1(__FUNCTION__,"ERROR","unsupported TFORM [%s]",
		       tmp_tform.cstr());
	}
	if ( status != 0 ) {
	    err_report(__FUNCTION__,"WARNING",
		       "TFORM or TDIM includes non-standard definitions");
	}

	if ( this->type_rec != new_type || 
	     this->full_bytes_rec != new_full_bytes ||
	     this->heap_type_rec != new_heap_type ) {
	    /* initialize internal memory buffer */
	    try {
		/* heap Ȥ */
		if ( new_type == FITS::LONGARRDESC_T || 
		     new_type == FITS::LLONGARRDESC_T ) {
		    const size_t nx[] = {new_elem_size, this->row_size_rec};
		    this->data_rec.init(new_sz_bytes, false, nx, 2, true);
		    this->heap_rec.init(new_heap_sz_bytes, false);
		}
		/* BIT_T ξ */
		else if ( new_type == FITS::BIT_T ) {
		    const size_t nx[] = {new_full_bytes, this->row_size_rec};
		    this->data_rec.init(1, false, nx, 2, true);
		    this->heap_rec.init(1, false);
		}
		/* ¾ξ(STRING_Tޤ) */
		else {
		    const size_t nx[] = {new_elem_size,	this->row_size_rec};
		    this->data_rec.init(new_sz_bytes, false, nx, 2, true);
		    this->heap_rec.init(1, false);
		}
		/* set default value of internal memory buffer */
		if ( new_type == FITS::LONGARRDESC_T || 
		     new_type == FITS::LLONGARRDESC_T ) {
		    this->update_zero_of_data_rec( new_heap_type, true );
		}
		else {
		    this->update_zero_of_data_rec( new_type, false );
		}
		/* clean up */
		this->data_rec.clean();
		this->heap_rec.clean();
	    }
	    catch (...) {
		err_throw(__FUNCTION__,"FATAL",
			  "data_rec.init() or heap_rec.init() failed");
	    }
	}

	if ( this->vl_max_length_rec != new_vl_max_length ) {
	    this->vl_max_length_rec = new_vl_max_length;
	}

	if ( this->dcol_size_rec != new_dcol_size ) {
	    const int c_1 = 1;
	    try {
		this->bit_size_telem.assign_default((const void *)&c_1);
		this->bit_size_telem.resize(new_dcol_size).clean();
	    }
	    catch (...) {
		err_throw(__FUNCTION__,"FATAL",
			  "bit_size_telem.resize() failed");
	    }
	}

	if ( new_type == FITS::STRING_T ) {
	    size_t len_alc;
	    if ( (long)(this->null_svalue_rec.length()) < new_sz_bytes )
		len_alc = new_sz_bytes;
	    else 		/* "NULL" ɽΤ */
		len_alc = this->null_svalue_rec.length();
	    /* */
	    if ( this->str_buf != NULL ) {
		delete this->str_buf;
		this->str_buf = NULL;
	    }
	    if ( this->tmp_str_buf != NULL ) {
		delete this->tmp_str_buf;
		this->tmp_str_buf = NULL;
	    }
	    try {
		this->str_buf = new tstring(len_alc);
		this->tmp_str_buf = new tstring(len_alc);
	    }
	    catch (...) {
		err_throw(__FUNCTION__,"FATAL","new failed");
	    }
	}
	else if (this->str_buf == NULL || this->str_buf->max_length() != 64) {
	    size_t len_alc = 64;
	    if ( this->str_buf != NULL ) {
		delete this->str_buf;
		this->str_buf = NULL;
	    }
	    if ( this->tmp_str_buf != NULL ) {
		delete this->tmp_str_buf;
		this->tmp_str_buf = NULL;
	    }
	    try {
		this->str_buf = new tstring(len_alc);
		this->tmp_str_buf = new tstring(len_alc);
	    }
	    catch (...) {
		err_throw(__FUNCTION__,"FATAL","new failed");
	    }
	}

	/* ̥ͤΥꥻåȤ˺ƤΤɲ 2011/12/2 */
	if ( (this->type_rec != new_type || 
	      this->heap_type_rec != new_heap_type)
	     && this->tany.length(TNULL_IDX) <= 0 )
	    do_tnull_reset = true;
	else do_tnull_reset = false;

	this->type_rec = new_type;
	this->bytes_rec = abs_sz2z(new_sz_bytes);
	this->elem_size_rec = new_elem_size;
	this->dcol_size_rec = new_dcol_size;
	this->full_bytes_rec = new_full_bytes;
	this->heap_type_rec = new_heap_type;

	/* ̥ͤΥꥻåȤ˺ƤΤɲ 2011/12/2 */
	if ( do_tnull_reset == true ) this->set_tnull(NULL);	/* ꥻå */

    }

    if ( def.talas != NULL ) {
	if ( def.talas[0] == '\0' /* NULL */ ) {
	    this->talas.resize(0);
	}
	else {
	    size_t i;
	    tarray_tstring tmp_arr;
	    tmp_arr.explode(def.talas,",",true);
	    this->talas.resize(0);
	    //for ( i=0 ; def.talas[i] != NULL ; i++ ) {
	    //    tmp_str.assign(def.talas[i]).strtrim();
	    //	this->talas.put(i,tmp_str.cstr(),1);
	    //}
	    for ( i=0 ; i < tmp_arr.length() ; i++ ) {
		tmp_arr[i].trim();
		this->talas.put(i,tmp_arr[i].cstr(),1);
	    }
	}
	tmp_str.assign(def.talas).trim();
	this->tany.put(TALAS_IDX,tmp_str.cstr(),1);
    }
    else {
	if ( this->talas.cstrarray() == NULL ) {
	    this->talas.resize(0);
	}
    }

    if ( def.talas_comment != NULL ) {
        tmp_str.assign(def.talas_comment);
	this->tany.put(TALAS_COMMENT_IDX,tmp_str.cstr(),1);
    }
    
    if ( def.telem != NULL ) {
	if ( def.telem[0] == '\0' /* NULL */ ) {
	    this->telem.resize(0);
	    this->telem_def.resize(0);
	    this->elem_index_rec.init();
	}
	else {
	    size_t i, j;
	    tarray_tstring tmp_arr;
	    tmp_arr.explode(def.telem,",",true);
	    this->telem.resize(0);
	    /* this->telem ʤȤ */
	    for ( i=0,j=0 ; i < tmp_arr.length() /* def.telem[i] != NULL */ ; i++ ) {
		ssize_t c_pos;
	        //tmp_str.assign(def.telem[i]).strtrim();
	        tmp_str.assign(tmp_arr[i]).trim();
		/* ӥåȥեȤ򰷤 */
		c_pos = tmp_str.strrchr(':');
		if ( 0 <= c_pos ) {
		    ssize_t k, n_bit = tmp_str.atoi(c_pos+1);
		    if ( 0 < n_bit ) {
			/* ':' 걦¦õ */
			tmp_str.erase(c_pos, tmp_str.length() - c_pos);
			for ( k=0 ; k < n_bit ; k++ ) {
			    this->telem.put(j,tmp_str.cstr(),1);
			    j++;
			}
		    }
		    else {
			err_report1(__FUNCTION__,"WARNING",
			       "Ignored invalid bitfield: %s", tmp_str.cstr());
			this->telem.put(j,tmp_str.cstr(),1);
			j++;
		    }
		}
		else {
		    this->telem.put(j,tmp_str.cstr(),1);
		    j++;
		}
	    }
	    /* this->telem_def ʤȤ */
	    this->telem_def.resize(0);
	    for ( i=0,j=0 ; i < this->telem.length() ; ) {
		size_t k, n_bit = 1;
		const tstring el_ref = this->telem.at_cs(i);
		if ( 0 < el_ref.length() ) {
		   for ( k=1 ; i+k < this->telem.length() ; k++ ) {
		      if ( el_ref.strcmp(this->telem.cstr(i+k)) == 0 ) n_bit++;
		      else break;
		   }
		}
		this->telem_def.put(j, this->telem.cstr(i),1);
		if ( 1 < n_bit ) {
		    this->telem_def.at(j).appendf(":%zu",n_bit);
		}
		j++;
		i += n_bit;
	    }
	    /* ǥåĥ */
	    this->elem_index_rec.init();
	    for ( i=0 ; i < this->telem.length() ; i++ ) {
		status = this->elem_index_rec.append(this->telem.cstr(i),i);
		if ( status < 0 ) {
		    err_throw(__FUNCTION__,"FATAL",
			       "this->elem_index_rec.append() failed");
		}
	    }
	}
	tmp_str.assign(def.telem).trim();
	this->tany.put(TELEM_IDX,tmp_str.cstr(),1);
    }
    else {
	if ( this->telem.cstrarray() == NULL ) {
	    this->telem.resize(0);
	}
	if ( this->telem_def.cstrarray() == NULL ) {
	    this->telem_def.resize(0);
	}
    }

    if ( def.telem_comment != NULL ) {
        tmp_str.assign(def.telem_comment);
	this->tany.put(TELEM_COMMENT_IDX,tmp_str.cstr(),1);
    }

    if ( def.tform != NULL || def.tdim != NULL || def.telem != NULL ) {
	long i;
	/* ӥåȤο */
	for ( i=0 ; i < this->dcol_size_rec ; i++ ) {
	    int j;
	    const char *e0 = this->telem.cstr(i);
	    if ( e0 == NULL ) {
		this->bit_size_telem.i(i) = 1;	/* default */
	    }
	    else {
		for ( j=0 ; ; j++ ) {
		    const char *e1 = this->telem.cstr(i+j);
		    if ( e1 == NULL ) break;
		    if ( this->telem.at(i).strcmp(e1) != 0 ) break;
		    //if ( c_strcmp(e0,e1) != 0 ) break;
		}
		this->bit_size_telem.i(i) = j;
	    }
	}
    }

    if ( def.ttype != NULL ) {
        tmp_str.assign(def.ttype).strtrim();
	this->tany.put(TTYPE_IDX,tmp_str.cstr(),1);
    }

    if ( def.ttype_comment != NULL ) {
        tmp_str.assign(def.ttype_comment);
	this->tany.put(TTYPE_COMMENT_IDX,tmp_str.cstr(),1);
    }

    if ( def.tunit != NULL ) {
        tmp_str.assign(def.tunit).strtrim();
	this->tany.put(TUNIT_IDX,tmp_str.cstr(),1);
    }

    if ( def.tunit_comment != NULL ) {
        tmp_str.assign(def.tunit_comment);
	this->tany.put(TUNIT_COMMENT_IDX,tmp_str.cstr(),1);
    }

    if ( def.tdisp != NULL ) {
	bool a_ok = true;
	bool oz_ok = true;
	bool fegd_ok = true;
        tmp_str.assign(def.tdisp).strtrim();
	this->tany.put(TDISP_IDX,tmp_str.cstr(),1);
	/* */
	this->fmt_str = NULL;
	this->fmt_nullstr = NULL;
	/* ƥǡǤɤ TDISP ΥפĤ뤫 */
	if ( this->type_rec != FITS::STRING_T &&
	     this->type_rec != FITS::LOGICAL_T ) a_ok = false;
	if ( this->type_rec == FITS::STRING_T ) oz_ok = false;
	if ( this->type_rec == FITS::LOGICAL_T ) oz_ok = false;
	if ( this->type_rec == FITS::LOGICAL_T ) fegd_ok = false;
	if ( this->type_rec == FITS::BIT_T ) fegd_ok = false;
	if ( 0 < tmp_str.length() ) {
	    ssize_t pos = tmp_str.strpbrk("[-0-9]");
	    int len0 = 0;
	    if ( 0 <= pos ) len0 = tmp_str.atoi(pos);
	    if ( (tmp_str.cchr(0) == 'A' || tmp_str.cchr(0) == 'L') && 
		 a_ok == true ) {
		if ( len0 != 0 ) {
		    if ( tmp_str.cchr(0) == 'A' ) {
			/*  */
			this->fmt_str.printf("%%%ds",-len0);
			this->fmt_nullstr.printf("%%%ds",-len0);
		    }
		    else {	/* for 'L' tdisp  */
			/*  */
			this->fmt_str.printf("%%%ds",len0);
			this->fmt_nullstr.printf("%%%ds",len0);
		    }
		}
		else {
		    this->fmt_str.printf("%%s");
		    this->fmt_nullstr.printf("%%s");
		}
		if ( this->str_buf == NULL || 
		     this->str_buf->max_length() < (size_t)len0 ) {
		    size_t len_alc;
		    if ( (int)(this->null_svalue_rec.length()) < len0 )
			len_alc = len0;
		    else	/* "NULL" ɽΤ */
			len_alc = this->null_svalue_rec.length();
		    /* */
		    if ( this->str_buf != NULL ) {
			delete this->str_buf;
			this->str_buf = NULL;
		    }
		    if ( this->tmp_str_buf != NULL ) {
			delete this->tmp_str_buf;
			this->tmp_str_buf = NULL;
		    }
		    try {
			this->str_buf = new tstring(len_alc);
			this->tmp_str_buf = new tstring(len_alc);
		    }
		    catch (...) {
			err_throw(__FUNCTION__,"FATAL","new failed");
		    }
		}
	    }
	    else if ( tmp_str.cchr(0) == 'O' && oz_ok == true ) {
		if ( 0 < len0 ) {
		    this->fmt_str.printf("0%%%dllo",len0+1);
		    this->fmt_nullstr.printf("%%%ds",len0+1);
		}
		else {
		    this->fmt_str.printf("0%%llo");
		    this->fmt_nullstr.printf("%%s");
		}
	    }
	    else if ( tmp_str.cchr(0) == 'Z' && oz_ok == true ) {
		if ( 0 < len0 ) {
		    this->fmt_str.printf("0x%%%dllx",len0+2);
		    this->fmt_nullstr.printf("%%%ds",len0+2);
		}
		else {
		    this->fmt_str.printf("0x%%llx");
		    this->fmt_nullstr.printf("%%s");
		}
	    }
	    else if ( tmp_str.cchr(0) == 'I' ) {
		if ( 0 < len0 ) {
		    this->fmt_str.printf("%%%dlld",len0);
		    this->fmt_nullstr.printf("%%%ds",len0);
		}
		else {
		    this->fmt_str.printf("%%lld");
		    this->fmt_nullstr.printf("%%s");
		}
	    }
	    /* 'B' is not supported */
	    else if ( (tmp_str.cchr(0) == 'F' || tmp_str.cchr(0) == 'E' ||
		       tmp_str.cchr(0) == 'G' || tmp_str.cchr(0) == 'D') &&
		      fegd_ok == true ) {
		ssize_t pos0 = tmp_str.strpbrk("[0-9]");
		ssize_t pos1 = tmp_str.strchr('.');
		int len0 = 0;
		int len1 = -1;
		if ( 0 <= pos0 ) {
		    if ( 0 <= pos1 ) {
			if ( pos0 < pos1 ) len0 = tmp_str.atoi(pos0);
		    }
		    else len0 = tmp_str.atoi(pos0);
		}
		if ( 0 <= pos1 ) {
		    pos1 = tmp_str.strpbrk(pos1,"[0-9]");
		    if ( 0 <= pos1 ) {
			len1 = tmp_str.atoi(pos1);
		    }
		}
		if ( tmp_str.cchr(0) == 'E' ) {
		    if ( 0 < len0 ) {
			if (0 <= len1) {
			    this->fmt_str.printf("%%%d.%dE",len0,len1);
			    this->fmt_nullstr.printf("%%%ds",len0);
			}
			else {
			    this->fmt_str.printf("%%%dE",len0);
			    this->fmt_nullstr.printf("%%%ds",len0);
			}
		    }
		    else {
			if (0 <= len1) {
			    this->fmt_str.printf("%%.%dE",len1);
			    this->fmt_nullstr.printf("%%s");
			}
			else {
			    this->fmt_str.printf("%%E");
			    this->fmt_nullstr.printf("%%s");
			}
		    }
		}
		else if ( tmp_str.cchr(0) == 'G' ) {
		    if ( 0 < len0 ) {
			if (0 <= len1) {
			    this->fmt_str.printf("%%%d.%dG",len0,len1);
			    this->fmt_nullstr.printf("%%%ds",len0);
			}
			else {
			    this->fmt_str.printf("%%%dG",len0);
			    this->fmt_nullstr.printf("%%%ds",len0);
			}
		    }
		    else {
			if (0 <= len1) {
			    this->fmt_str.printf("%%.%dG",len1);
			    this->fmt_nullstr.printf("%%s");
			}
			else {
			    this->fmt_str.printf("%%G");
			    this->fmt_nullstr.printf("%%s");
			}
		    }
		}
		/* 'D' ΰϤ? */
		/* (fv ưǤ %f FORTRANŪˤ %E Ǹ D ѹ) */
		else if ( tmp_str.cchr(0) == 'D' ) {
		    if ( 0 < len0 ) {
			if (0 <= len1) {
			    this->fmt_str.printf("%%%d.%dE",len0,len1);
			    this->fmt_nullstr.printf("%%%ds",len0);
			}
			else {
			    this->fmt_str.printf("%%%d.15E",len0);
			    this->fmt_nullstr.printf("%%%ds",len0);
			}
		    }
		    else {
			if (0 <= len1) {
			    this->fmt_str.printf("%%.%dE",len1);
			    this->fmt_nullstr.printf("%%s");
			}
			else {
			    this->fmt_str.printf("%%.15E");
			    this->fmt_nullstr.printf("%%s");
			}
		    }
		}
		/* 'F' ξ */
		else {
		    if ( 0 < len0 ) {
			if (0 <= len1) {
			    this->fmt_str.printf("%%%d.%df",len0,len1);
			    this->fmt_nullstr.printf("%%%ds",len0);
			}
			else {
			    this->fmt_str.printf("%%%df",len0);
			    this->fmt_nullstr.printf("%%%ds",len0);
			}
		    }
		    else {
			if (0 <= len1) {
			    this->fmt_str.printf("%%.%df",len1);
			    this->fmt_nullstr.printf("%%s");
			}
			else {
			    this->fmt_str.printf("%%f");
			    this->fmt_nullstr.printf("%%s");
			}
		    }
		}
	    }
	}
    }
	
    if ( def.tdisp_comment != NULL ) {
        tmp_str.assign(def.tdisp_comment);
	this->tany.put(TDISP_COMMENT_IDX,tmp_str.cstr(),1);
    }

    if ( def.tform != NULL ) {
        tmp_str.assign(def.tform).strtrim();
	this->tany.put(TFORM_IDX,tmp_str.cstr(),1);
    }

    if ( def.tform_comment != NULL ) {
        tmp_str.assign(def.tform_comment);
	this->tany.put(TFORM_COMMENT_IDX,tmp_str.cstr(),1);
    }

    if ( def.tdim != NULL ) {
        tmp_str.assign(def.tdim).strtrim();
	this->tany.put(TDIM_IDX,tmp_str.cstr(),1);
    }

    if ( def.tdim_comment != NULL ) {
        tmp_str.assign(def.tdim_comment);
	this->tany.put(TDIM_COMMENT_IDX,tmp_str.cstr(),1);
    }

    /*
     * TNULL νϺ𤷤䤹Τ
     *  def.tnull = NULL   -> ⤷ʤ
     *  def.tnull = ""     -> õ
     *  def.tnull = " "    -> this->bytes() Ĥζ NULL ͤȤ
     */
    if ( def.tnull != NULL ) {
	if ( this->type_rec != FITS::STRING_T ) {
	    tmp_str.assign(def.tnull).strtrim();
	    if ( 0 < tmp_str.length() ) {
		long long tnl = tmp_str.atoll();
		this->set_tnull(&tnl);			/* å */
	    }
	    else this->set_tnull(NULL);			/* ꥻå */
	}
	else {
	    tmp_str.assign(def.tnull);
	    if ( 0 < tmp_str.length() ) {
		/* Changed 2011/12/1 */
		//tmp_str.resize(this->full_bytes_rec);
		if ( (long)(tmp_str.length()) <= this->bytes() ) {
		    tmp_str.resize(this->bytes());
		}
		/* NULL ͤĹ礭硤*/
		/* ǷٹФƤɤۤäƤ */
	    }
	    /*  NULL ꥻåȤƤ */
	    this->set_tnull(NULL);
	}
	this->tany.put(TNULL_IDX,tmp_str.cstr(),1);
	if ( 0 < this->tany.length(TNULL_IDX) ) {
	    this->tnull_is_set_rec = true;
	}
	else {
	    this->tnull_is_set_rec = false;
	}
    }

    if ( def.tnull_comment != NULL ) {
        tmp_str.assign(def.tnull_comment);
	this->tany.put(TNULL_COMMENT_IDX,tmp_str.cstr(),1);
    }

    /* ̥ݤϤʤ褦ˤ */
    this->tany.resize(TANY_ARR_SIZE);

    this->update_def_rec_ptr();

 quit:
    return *this;
}

/*
 * Functions to convert numeric columns
 */

/**
 * @brief  ǡѴѹ
 *
 *  ĹǤϤʤ뤤ϼ¿Υ(TFORMn λ꤬ 'B''I'
 *  'J''K''E''D'ޤ)η new_type ѹޤ
 *  ɬפ˱ơХåեΥѹޤ<br>
 *  new_type ˻ǤͤϡFITS::DOUBLE_TFITS::FLOAT_T
 *  FITS::LONGLONG_TFITS::LONG_TFITS::SHORT_TFITS::BYTE_T Τ줫
 *  Ǥ<br>
 *  new_zeronew_scalenew_null ꤵ줿ϡإå TZEROn
 *  TSCALnTNULLn ѹͤȿǤǡѴޤ<br>
 *  new_null ͭʤΤϡnew_type ξΤߤǤ<br>
 *  ʸ󷿤ΥϤΥдؿǤѴǤޤ
 *
 * @param  new_type ѴΥǡ
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::convert_type( int new_type )
{
    return this->convert_col_type( new_type, NULL, NULL, NULL );
}

/**
 * @brief  ǡѴѹ
 *
 *  ĹǤϤʤ뤤ϼ¿Υ(TFORMn λ꤬ 'B''I'
 *  'J''K''E''D'ޤ)η new_type ѹޤ
 *  ɬפ˱ơХåեΥѹޤ<br>
 *  new_type ˻ǤͤϡFITS::DOUBLE_TFITS::FLOAT_T
 *  FITS::LONGLONG_TFITS::LONG_TFITS::SHORT_TFITS::BYTE_T Τ줫
 *  Ǥ<br>
 *  إå TZEROnѹͤȿǤǡѴޤ<br>
 *  ʸ󷿤ΥϤΥдؿǤѴǤޤ
 *
 * @param  new_type ѴΥǡ
 * @param  new_zero ѴTZEROn
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::convert_type( int new_type, double new_zero )
{
    return this->convert_col_type( new_type, &new_zero, NULL, NULL );
}

/**
 * @brief  ǡѴѹ
 *
 *  ĹǤϤʤ뤤ϼ¿Υ(TFORMn λ꤬ 'B''I'
 *  'J''K''E''D'ޤ)η new_type ѹޤ
 *  ɬפ˱ơХåեΥѹޤ<br>
 *  new_type ˻ǤͤϡFITS::DOUBLE_TFITS::FLOAT_T
 *  FITS::LONGLONG_TFITS::LONG_TFITS::SHORT_TFITS::BYTE_T Τ줫
 *  Ǥ<br>
 *  إå TZEROnTSCALnѹͤȿǤǡѴ
 *  <br>
 *  ʸ󷿤ΥϤΥдؿǤѴǤޤ
 *
 * @param  new_type ѴΥǡ
 * @param  new_zero ѴTZEROn
 * @param  new_scale ѴTSCALn
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::convert_type( int new_type, double new_zero,
					      double new_scale )
{
    return this->convert_col_type( new_type, &new_zero, &new_scale, NULL );
}

/**
 * @brief  ǡѴѹ
 *
 *  ĹǤϤʤ뤤ϼ¿Υ(TFORMn λ꤬ 'B''I'
 *  'J''K''E''D'ޤ)η new_type ѹޤ
 *  ɬפ˱ơХåեΥѹޤ<br>
 *  new_type ˻ǤͤϡFITS::DOUBLE_TFITS::FLOAT_T
 *  FITS::LONGLONG_TFITS::LONG_TFITS::SHORT_TFITS::BYTE_T Τ줫
 *  Ǥ<br>
 *  إå TZEROnTSCALnTNULLn ѹͤȿǤǡ
 *  Ѵޤ<br>
 *  new_null ͭʤΤϡnew_type ξΤߤǤ<br>
 *  ʸ󷿤ΥϤΥдؿǤѴǤޤ
 *
 * @param  new_type ѴΥǡ
 * @param  new_zero ѴTZEROn
 * @param  new_scale ѴTSCALn
 * @param  new_null ѴTNULLn 
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::convert_type( int new_type, double new_zero, 
					      double new_scale,
					      long long new_null )
{
    return this->convert_col_type( new_type, &new_zero, &new_scale, &new_null);
}

/*
 * Functions to set TNULL and scaling parameters
 */

/**
 * @brief  ʪñ(TUNITn )
 */
fits_table_col &fits_table_col::assign_tunit( const char *unit )
{
    fits::table_def_all def = {NULL};
    def.tunit = unit;
    return this->_define(def);
}

/**
 * @brief  ʪñ(TUNITn)ξõ
 *
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::erase_tunit()
{
    fits::table_def_all def = {NULL};
    def.tunit = "";
    return this->_define(def);
}

/**
 * @brief  ΥФ NULL(TNULLn ) long long 
 */
fits_table_col &fits_table_col::assign_tnull( long long null )
{
    tstring sval;
    fits::table_def_all def = {NULL};
    sval.printf("%lld",null);
    def.tnull = sval.cstr();
    return this->_define(def);
}

/**
 * @brief  ΥФ NULL(TNULLn )ʸ
 */
fits_table_col &fits_table_col::assign_tnull( const char *null )
{
    if ( null != NULL ) {
	fits::table_def_all def = {NULL};
	def.tnull = null;
	return this->_define(def);
    }
    else return *this;
}

/**
 * @brief  ΥФ NULL(TNULLn)ξõ
 *
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::erase_tnull()
{
    fits::table_def_all def = {NULL};
    def.tnull = "";
    return this->_define(def);
}

/**
 * @brief  (TZEROn )
 *
 * @param   zero TZEROn 
 * @param   prec ()ά15
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::assign_tzero( double zero, int prec )
{
    tstring val, fmt;
    fits::table_def_all def = {NULL};
    ssize_t ep;
    fmt.printf("%%.%dG",prec);
    val.printf(fmt.cstr(),zero);
    ep = val.strchr('E');
    if ( ep < 0 && val.strchr('.') < 0 ) val.append(".");
    if ( 0 <= ep ) val.at(ep) = 'D';
    def.tzero = val.cstr();
    return this->_define(def);
}

/**
 * @brief  (TZEROn)ξõ
 *
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::erase_tzero()
{
    fits::table_def_all def = {NULL};
    def.tzero = "";
    return this->_define(def);
}

/**
 * @brief  󥰥ե(TSCALn )
 *
 * @param   scal TSCALn 
 * @param   prec ()ά15
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::assign_tscal( double scal, int prec )
{
    tstring val, fmt;
    fits::table_def_all def = {NULL};
    ssize_t ep;
    fmt.printf("%%.%dG",prec);
    val.printf(fmt.cstr(),scal);
    ep = val.strchr('E');
    if ( ep < 0 && val.strchr('.') < 0 ) val.append(".");
    if ( 0 <= ep ) val.at(ep) = 'D';
    def.tscal = val.cstr();
    return this->_define(def);
}

/**
 * @brief  󥰥ե(TSCALn)ξõ
 *
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::erase_tscal()
{
    fits::table_def_all def = {NULL};
    def.tscal = "";
    return this->_define(def);
}

/*
 * Functions to resize, erase and insert rows.
 */

/* fits_table_col, fits_table ϡΥдؿȤʤ */
/**
 * @brief  Կѹ
 *
 *  ˤϡǥեͤåȤޤǥեͤνͤϡ
 *  ξ 'F'ʸ󷿤ξ ' ' ʤʸ󡤤¾ηξ
 *  0 Ǥǥեͤ fits_table_col::assign_default() дؿѹ
 *  ޤ<br>
 *  Ȥ fits_table ֥Ȥδˤ硤¾ΥƱ
 *  ¹Ԥޤ
 *
 * @param   num_rows ѹιԿ
 * @return  Ȥλ
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::resize( long num_rows )
{
    if ( this->manager == NULL ) {	/* fits_table δǤʤ */
	return this->_resize(num_rows);
    }
    else {				/* fits_table δξ */
	/* Ū _resize() Ƥ */
	this->manager->resize_rows(num_rows);
	return *this;
    }
}

/**
 * @brief  Կѹ (٥)
 *
 * @param   num_rows ѹιԿ
 * @return  Ȥλ
 * @note    fits_table ϡΥдؿȤ<br>
 *          Υдؿ protected Ǥ
 */
fits_table_col &fits_table_col::_resize( long num_rows )
{
    if ( this->full_bytes_rec == 0 ) return *this;	/* not defined */
    if ( num_rows < 0 ) return *this;			/* invalid */

    this->data_rec.resize(1, num_rows);
    this->row_size_rec = this->data_rec.row_length();

    return *this;
}

/*
 * Functions to edit rows
 */

/**
 * @brief  ԤԤؤΥԡ
 *
 * @param  src_index ԡ򼨤ԥǥå
 * @param  num_rows ԡԤԿ
 * @param  dest_index ԡ򼨤ԥǥå
 * @return  Ȥλ
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::move( long src_index, long num_rows, 
				      long dest_index )
{
    if ( this->full_bytes_rec == 0 ) return *this;	/* not defined */
    if ( num_rows < 0 ) return *this;			/* invalid */

    this->data_rec.move(1, src_index, num_rows, dest_index, false);

    return *this;
}

/**
 * @brief  ԤȹԤȤؤ
 *
 *  index0 ǻꤵ줿Ԥ num_rows ĤιԤindex1 ǻꤵ줿Ԥ
 *  num_cols ĤιԤؤޤ<br>
 *  num_cols ˤäơŤʤԤϡnum_cols ͤ򸺤餷ؤ
 *  Ԥʤޤ
 *
 * @param  index0 ؤ򼨤ԥǥå
 * @param  num_rows ؤԤԿ
 * @param  index1 ؤ򼨤ԥǥå
 * @return  Ȥλ
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::swap( long index0, long num_rows, 
				      long index1 )
{
    if ( this->full_bytes_rec == 0 ) return *this;	/* not defined */
    if ( num_rows < 0 ) return *this;			/* invalid */

    this->data_rec.swap(1, index0, num_rows, index1);

    return *this;
}

/**
 * @brief  Ԥνȿž
 *
 * @param   index ȿžϰ֤򼨤ԥǥå
 * @param   num_rows ȿžԿ
 * @return  Ȥλ
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::flip( long index, long num_rows )
{
    if ( this->full_bytes_rec == 0 ) return *this;	/* not defined */
    if ( num_rows < 0 ) return *this;			/* invalid */

    this->data_rec.flip(1, index, num_rows);

    return *this;
}

/* fits_table_col, fits_table ϡΥдؿȤʤ */
/**
 * @brief  Ԥξõ
 *
 *  Ȥ fits_table ֥Ȥδˤ硤¾ΥƱ
 *  ¹Ԥޤ
 *
 * @param   index ϰ֤򼨤ԥǥå
 * @param   num_rows Կ
 * @return  Ȥλ
 * @attention  ơ֥βΥХȿȹԿ礭 index 
 *             ۤɡΥԡμ¹ԥȤ礷ޤ
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::erase( long index, long num_rows )
{
    if ( this->manager == NULL ) {	/* fits_table δǤʤ */
	return this->_erase(index, num_rows);
    }
    else {				/* fits_table δξ */
	/* Ū _erase() Ƥ */
	this->manager->erase_rows(index, num_rows);
	return *this;
    }
}

/**
 * @brief  Ԥξõ (٥)
 *
 * @param   index ϰ֤򼨤ԥǥå
 * @param   num_rows Կ
 * @return  Ȥλ
 * @note    fits_table ϡΥдؿȤ
 *          Υдؿ protected Ǥ
 */
fits_table_col &fits_table_col::_erase( long index, long num_rows )
{
    if ( this->full_bytes_rec == 0 ) return *this;	/* not defined */
    if ( num_rows < 0 ) return *this;			/* invalid */

    this->data_rec.erase(1, index, num_rows);
    this->row_size_rec = this->data_rec.row_length();

    return *this;
}

/* fits_table_col, fits_table ϡΥдؿȤʤ */
/**
 * @brief  Ԥ
 *
 *  ˤϡǥեͤåȤޤǥեͤνͤϡ
 *  ξ 'F'ʸ󷿤ξ ' ' ʤʸ󡤤¾ηξ
 *  0 Ǥǥեͤ fits_table_col::assign_default() дؿѹ
 *  ޤ<br>
 *  Ȥ fits_table ֥Ȥδˤ硤¾ΥƱ
 *  ¹Ԥޤ
 *
 * @param   index ֤򼨤ԥǥå
 * @param   num_rows Կ
 * @return  Ȥλ
 * @attention  ơ֥βΥХȿȹԿ礭 index 
 *             ۤɡΥԡμ¹ԥȤ礷ޤ
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::insert( long index, long num_rows )
{
    if ( this->manager == NULL ) {	/* fits_table δǤʤ */
	return this->_insert(index,num_rows);
    }
    else {				/* fits_table δξ */
	/* Ū _insert() Ƥ */
	this->manager->insert_rows(index,num_rows);
	return *this;
    }
}

/**
 * @brief  Ԥ (٥)
 *
 * @return  Ȥλ
 * @note    fits_table ϡΥдؿȤ
 *          Υдؿ protected Ǥ
 */
fits_table_col &fits_table_col::_insert( long index, long num_rows )
{
    if ( this->full_bytes_rec == 0 ) return *this;	/* not defined */
    if ( num_rows < 0 ) return *this;			/* invalid */

    this->data_rec.insert(1, index, num_rows);
    this->row_size_rec = this->data_rec.row_length();

    return *this;
}

/**
 * @brief  μΤΥдؿ
 *
 * @note   Υдؿ protected Ǥ
 */
void **fits_table_col::data_ptr_mng_ptr()
{
    return &this->data_ptr_rec;
}

/**
 * @brief  μΤΥдؿ
 *
 * @note   Υдؿ protected Ǥ
 */
void **fits_table_col::heap_ptr_mng_ptr()
{
    return &this->heap_ptr_rec;
}

/**
 * @brief  ४֥ȤΥݡ
 *
 *  fits_table_col ֥ from  idx_begin  num_rows ĤιԤ
 *  Ȥ dest_index ǻꤵ줿Ԥ num_rowsĤιԤإݡȤޤ<br>
 *  from λĥ֥ȤΥηϡפƤɬפ
 *  ޤ󡥰פʤϡͤѴƥݡȤޤ
 *
 * @param  dest_index ݡιԥǥå
 * @param  from ݡȸΥơ֥륪֥
 * @param  idx_begin ݡȸιԥǥå
 * @param  num_rows ݡȤԤԿ
 * @return  Ȥλ
 * @note  ݡ()ιԿѹϹԤʤޤ󡥹Ԥη»ʤݡ
 *        ˤϡ餫ώʬʹԿݤƤɬפޤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::import( long dest_index, 
				    const fits_table_col &from, long idx_begin,
				    long num_rows )
{
    if ( this->full_bytes_rec == 0 ) return *this;	/* not defined */
    if ( from.full_bytes_rec == 0 ) return *this;	/* not defined */

    /* reject variable length array */
    if ( this->heap_is_used() == true ) {
	err_report1(__FUNCTION__,"WARNING", 
		    "cannot use this function for TFORM = %s",
		    this->tany.cstr(TFORM_IDX));
	return *this;
    }
    if ( from.heap_is_used() == true ) {
	err_report1(__FUNCTION__,"WARNING", 
		    "cannot import TFORM = %s",from.tany.cstr(TFORM_IDX));
	return *this;
    }

    /* check args */
    if ( num_rows == FITS::ALL )
	num_rows = from.length() - idx_begin;

    if ( dest_index < 0 || this->row_size_rec <= dest_index || 
	 num_rows < 0 || idx_begin < 0 ) {
	goto quit;	/* invalid */
    }
    if ( from.length() <= idx_begin ) {
	goto quit;	/* invalid */
    }
    if ( num_rows == 0 ) {
	goto quit;
    }
    
    if ( this->row_size_rec < dest_index + num_rows ) {
	num_rows = this->row_size_rec - dest_index;
    }
    if ( from.length() < idx_begin + num_rows ) {
	num_rows = from.length() - idx_begin;
    }

    /* import */
    if ( this->type() == from.type() &&
	 this->bytes() == from.bytes() &&
	 this->elem_length() == from.elem_length() &&
	 this->tzero() == from.tzero() &&
	 this->tscal() == from.tscal() &&
	 this->tnull_is_set() == from.tnull_is_set() &&
	 this->tnull() == from.tnull() ) {
	long full_bytes = this->full_bytes_rec;
	this->data_rec.putdata(
		(const char *)from.data_ptr_cs() + full_bytes * idx_begin,
		full_bytes * num_rows,
		0, dest_index);
    }
    else {
	long i,j,ix,id;
	/* &from == this Ǥɤ褦 */
	if ( dest_index <= idx_begin ) {
	    ix = 0;
	    id = 1;
	}
	else {
	    ix = num_rows - 1;
	    id = -1;
	}
	if ( this->type_rec == FITS::STRING_T ) {
	    tstring sval;
	    for ( i=0 ; i < num_rows ; i++, ix += id ) {
		for ( j=0 ; j < this->elem_size_rec ; j++ ) {
		    const char *v = from.get_svalue(idx_begin+ix,j,sval);
		    this->assign(v,dest_index+ix,j);
		}
	    }
	}
	//else if ( this->type_rec == FITS::LOGICAL_T ) {
	//    if ( from.type() == FITS::STRING_T ||
	//	 from.type() == FITS::LOGICAL_T ) {
	//	for ( i=0 ; i < num_rows ; i++, ix += id ) {
	//	    for ( j=0 ; j < this->elem_size_rec ; j++ ) {
	//		int v = from.logical_value(idx_begin+ix,j);
	//		this->assign_logical(v,dest_index+ix,j);
	//	    }
	//	}
	//   }
	//   else {
	//	for ( i=0 ; i < num_rows ; i++, ix += id ) {
	//	    for ( j=0 ; j < this->elem_size_rec ; j++ ) {
	//		bool v = from.bvalue(idx_begin+ix,j);
	//		this->assign(v,dest_index+ix,j);
	//	    }
	//	}
	//    }
	//}
	if ( this->type_rec == FITS::LONGLONG_T ) {
	    for ( i=0 ; i < num_rows ; i++, ix += id ) {
		for ( j=0 ; j < this->elem_size_rec ; j++ ) {
		    long long v = from.llvalue(idx_begin+ix,j);
		    this->assign(v,dest_index+ix,j);
		}
	    }
	}
	else {
	    for ( i=0 ; i < num_rows ; i++, ix += id ) {
		for ( j=0 ; j < this->elem_size_rec ; j++ ) {
		    double v = from.dvalue(idx_begin+ix,j);
		    this->assign(v,dest_index+ix,j);
		}
	    }
	}
    }

 quit:
    return *this;
}

/* fill all or part of rows using default value */
/**
 * @brief  ιԤΥͤ(ǥեͤ򥻥å)
 *
 *  ǥեͤνͤϡξ 'F'ʸ󷿤ξ ' ' 
 *  ʤʸ󡤤¾ηξ 0 Ǥǥեͤ 
 *  fits_table_col::assign_default() дؿѹǤޤ
 *
 * @param   index ϰ֤򼨤ԥǥå
 * @param   num_rows Կ
 * @return  Ȥλ
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::clean( long index, long num_rows )
{
    if ( this->full_bytes_rec == 0 ) return *this;	/* not defined */
    if ( num_rows < 0 ) return *this;			/* invalid */

    this->data_rec.clean( 0, this->data_rec.col_length(),
			  index, num_rows );

    return *this;
}

/**
 * @brief  ƤιԤΥͤ(ǥեͤ򥻥å)
 *
 *  ǥեͤνͤϡξ 'F'ʸ󷿤ξ ' ' 
 *  ʤʸ󡤤¾ηξ 0 Ǥǥեͤ 
 *  fits_table_col::assign_default() дؿѹǤޤ
 *
 * @return  Ȥλ
 * @attention  Ĺξ硤󵭽һҤФȤʤޤ
 */
fits_table_col &fits_table_col::clean()
{
    return this->clean( 0, this->row_size_rec );
}

#define REMOVE_SPC(ts)	{ \
    ssize_t i_s; \
    ts.strtrim(' '); \
    for ( i_s=0 ; 0 <= (i_s=ts.strchr(i_s,' ')) ; i_s++ ) { \
        ts.erase(i_s,ts.strspn(i_s,' ')); \
    } \
}

/*
 * high-level APIs to read a cell.
 *  - TSCAL and TZERO are applied.
 *  - NULL value is handled.
 */

/**
 * @brief  ͤ¿ͤǼ (٥)
 *
 *  ͤˡإå TZEROn  TSCALn ȿǤ¿ͤޤ<br>
 *  NULLͤξϡNAN֤ޤХʥơ֥䥢
 *  ơ֥ TNULLn ͤˡͤפNULLͤȤޤ<br>
 *  Υξϡͤ 'T' ʤ 1 ͤ 'F' ʤ 0 򡤤ʳ
 *   NAN ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥ TFORMn
 *  ɽ򼨤ʤˤϡʸ¿ͤѴͤ򤽤Τޤ
 *  ֤ޤ<br>
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλNULLͤξNAN <br>
 *          NAN: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
double fits_table_col::dvalue( long row_index ) const
{
    double v;
    if ( this->type_rec == FITS::BIT_T || 
	 this->type_rec == FITS::LOGICAL_T )
	v = this->double_value(row_index);
    else if ( this->type_rec == FITS::STRING_T ) {
	if ( row_index < 0 || this->row_size_rec <= row_index ) return NAN;
	tstring sval;
	this->get_string_value(row_index,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) return NAN;
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 )
	    return NAN;
	else {
	    REMOVE_SPC(sval);
	    v = sval.atof();
	    if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGDI") == 0 ) {
		v = this->tzero_rec + v * this->tscal_rec;
	    }
	}
    }
    else if ( this->tnull_is_set_rec == true ) {
	if ( row_index < 0 || this->row_size_rec <= row_index ) return NAN;
	if ( this->type_rec == FITS::LONG_T ||
	     this->type_rec == FITS::SHORT_T ||
	     this->type_rec == FITS::BYTE_T ||
	     this->type_rec == FITS::LONGLONG_T ) {
	    long long llv = this->longlong_value(row_index);
	    if ( llv == this->tnull_r_rec ) return NAN;
	    else v = this->tzero_rec + llv * this->tscal_rec;
	}
	else {
	    v = this->tzero_rec + 
		this->double_value(row_index) * this->tscal_rec;
	}
    }
    else
	v = this->tzero_rec + this->double_value(row_index) * this->tscal_rec;
    return v;
}

/**
 * @brief  ͤ¿ͤǼ (٥)
 *
 *  ͤˡإå TZEROn  TSCALn ȿǤ¿ͤޤ<br>
 *  NULLͤξϡNAN֤ޤХʥơ֥䥢
 *  ơ֥ TNULLn ͤˡͤפNULLͤȤޤ<br>
 *  Υξϡͤ 'T' ʤ 1 ͤ 'F' ʤ 0 򡤤ʳ
 *   NAN ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥ TFORMn
 *  ɽ򼨤ʤˤϡʸ¿ͤѴͤ򤽤Τޤ
 *  ֤ޤ<br>
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλNULLͤξNAN <br>
 *          NAN: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
double fits_table_col::dvalue( long row_index,
			     const char *elem_name, long repetition_idx ) const
{
    double v;
    if ( this->type_rec == FITS::BIT_T || 
	 this->type_rec == FITS::LOGICAL_T )
	v = this->double_value(row_index,elem_name,repetition_idx);
    else if ( this->type_rec == FITS::STRING_T ) {
	long elem_index = this->elem_index_rec.index(elem_name,0);
	if ( row_index < 0 || this->row_size_rec <= row_index ) return NAN;
	if ( elem_index < 0 || repetition_idx < 0 ) return NAN;
	tstring sval;
	this->get_string_value(row_index,elem_index,repetition_idx, sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) return NAN;
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 )
	    return NAN;
	else {
	    REMOVE_SPC(sval);
	    v = sval.atof();
	    if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGDI") == 0 ) {
		v = this->tzero_rec + v * this->tscal_rec;
	    }
	}
    }
    else if ( this->tnull_is_set_rec == true ) {
	long elem_index = this->elem_index_rec.index(elem_name,0);
	if ( row_index < 0 || this->row_size_rec <= row_index ) return NAN;
	if ( elem_index < 0 || repetition_idx < 0 ) return NAN;
	if ( this->type_rec == FITS::LONG_T ||
	     this->type_rec == FITS::SHORT_T ||
	     this->type_rec == FITS::BYTE_T ||
	     this->type_rec == FITS::LONGLONG_T ) {
	    long long llv = this->longlong_value(row_index,
						 elem_index,repetition_idx);
	    if ( llv == this->tnull_r_rec ) return NAN;
	    else v = this->tzero_rec + llv * this->tscal_rec;
	}
	else {
	    v = this->tzero_rec + 
      this->double_value(row_index,elem_name,repetition_idx) * this->tscal_rec;
	}
    }
    else
	v = this->tzero_rec + 
      this->double_value(row_index,elem_name,repetition_idx) * this->tscal_rec;
    return v;
}

/**
 * @brief  ͤ¿ͤǼ (٥)
 *
 *  ͤˡإå TZEROn  TSCALn ȿǤ¿ͤޤ<br>
 *  NULLͤξϡNAN֤ޤХʥơ֥䥢
 *  ơ֥ TNULLn ͤˡͤפNULLͤȤޤ<br>
 *  Υξϡͤ 'T' ʤ 1 ͤ 'F' ʤ 0 򡤤ʳ
 *   NAN ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥ TFORMn
 *  ɽ򼨤ʤˤϡʸ¿ͤѴͤ򤽤Τޤ
 *  ֤ޤ<br>
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @return  : ｪλNULLͤξNAN <br>
 *          NAN: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
double fits_table_col::dvalue( long row_index,
			      long elem_index, long repetition_idx ) const
{
    double v;
    if ( this->type_rec == FITS::BIT_T || 
	 this->type_rec == FITS::LOGICAL_T )
	v = this->double_value(row_index,elem_index,repetition_idx);
    else if ( this->type_rec == FITS::STRING_T ) {
	if ( row_index < 0 || this->row_size_rec <= row_index ) return NAN;
	if ( elem_index < 0 || repetition_idx < 0 ) return NAN;
	tstring sval;
	this->get_string_value(row_index,elem_index,repetition_idx, sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) return NAN;
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 )
	    return NAN;
	else {
	    REMOVE_SPC(sval);
	    v = sval.atof();
	    if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGDI") == 0 ) {
		v = this->tzero_rec + v * this->tscal_rec;
	    }
	}
    }
    else if ( this->tnull_is_set_rec == true ) {
	if ( row_index < 0 || this->row_size_rec <= row_index ) return NAN;
	if ( elem_index < 0 || repetition_idx < 0 ) return NAN;
	if ( this->type_rec == FITS::LONG_T ||
	     this->type_rec == FITS::SHORT_T ||
	     this->type_rec == FITS::BYTE_T ||
	     this->type_rec == FITS::LONGLONG_T ) {
	    long long llv = this->longlong_value(row_index,
						 elem_index,repetition_idx);
	    if ( llv == this->tnull_r_rec ) return NAN;
	    else v = this->tzero_rec + llv * this->tscal_rec;
	}
	else {
	    v = this->tzero_rec + 
     this->double_value(row_index,elem_index,repetition_idx) * this->tscal_rec;
	}
    }
    else
	v = this->tzero_rec + 
     this->double_value(row_index,elem_index,repetition_idx) * this->tscal_rec;
    return v;
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ͤˡإå TZEROn  TSCALn ȿǤ¿ͤ˺Ǥᤤͤ
 *  ֤ޤͤ long ɽǤϰϤʤ硤INDEF_LONG
 *  ֤ޤ<br>
 *  NULLͤξϡINDEF_LONG ֤ޤХʥơ֥
 *  ơ֥ TNULLn ͤˡͤפNULLͤȤ
 *  <br>
 *  Υξϡͤ 'T' ʤ 1 ͤ 'F' ʤ 0 򡤤ʳ
 *   INDEF_LONG ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥
 *  TFORMn ɽ򼨤ʤˤϡʸ¿ͤѴͤ
 *  Ǥᤤ֤ͤޤ
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @return  ͡NULLͤξINDEF_LONG <br>
 *          INDEF_LONG: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
long fits_table_col::lvalue( long row_index ) const
{
    double dv = this->dvalue(row_index);
    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LONG <= dv && dv <= MAX_DOUBLE_ROUND_LONG )
	return round_d2l(dv);
    else return INDEF_LONG;
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ͤˡإå TZEROn  TSCALn ȿǤ¿ͤ˺Ǥᤤͤ
 *  ֤ޤͤ long ɽǤϰϤʤ硤INDEF_LONG
 *  ֤ޤ<br>
 *  NULLͤξϡINDEF_LONG ֤ޤХʥơ֥
 *  ơ֥ TNULLn ͤˡͤפNULLͤȤ
 *  <br>
 *  Υξϡͤ 'T' ʤ 1 ͤ 'F' ʤ 0 򡤤ʳ
 *   INDEF_LONG ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥
 *  TFORMn ɽ򼨤ʤˤϡʸ¿ͤѴͤ
 *  Ǥᤤ֤ͤޤ
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @return  ͡NULLͤξINDEF_LONG <br>
 *          INDEF_LONG: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
long fits_table_col::lvalue( long row_index,
			     const char *elem_name, long repetition_idx ) const
{
    double dv = this->dvalue(row_index,elem_name,repetition_idx);
    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LONG <= dv && dv <= MAX_DOUBLE_ROUND_LONG )
	return round_d2l(dv);
    else return INDEF_LONG;
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ͤˡإå TZEROn  TSCALn ȿǤ¿ͤ˺Ǥᤤͤ
 *  ֤ޤͤ long ɽǤϰϤʤ硤INDEF_LONG
 *  ֤ޤ<br>
 *  NULLͤξϡINDEF_LONG ֤ޤХʥơ֥
 *  ơ֥ TNULLn ͤˡͤפNULLͤȤ
 *  <br>
 *  Υξϡͤ 'T' ʤ 1 ͤ 'F' ʤ 0 򡤤ʳ
 *   INDEF_LONG ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥
 *  TFORMn ɽ򼨤ʤˤϡʸ¿ͤѴͤ
 *  Ǥᤤ֤ͤޤ
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @return  ͡NULLͤξINDEF_LONG <br>
 *          INDEF_LONG: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
long fits_table_col::lvalue( long row_index,
			     long elem_index, long repetition_idx ) const
{
    double dv = this->dvalue(row_index,elem_index,repetition_idx);
    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LONG <= dv && dv <= MAX_DOUBLE_ROUND_LONG )
	return round_d2l(dv);
    else return INDEF_LONG;
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ͤˡإå TZEROn  TSCALn ȿǤ¿ͤ˺Ǥᤤͤ
 *  ֤ޤͤ long long ɽǤϰϤʤ硤
 *  INDEF_LLONG ֤ޤ<br>
 *  NULLͤξϡINDEF_LLONG ֤ޤХʥơ֥
 *  ơ֥ TNULLn ͤˡͤפNULLͤȤ
 *  <br>
 *  Υξϡͤ 'T' ʤ 1 ͤ 'F' ʤ 0 򡤤ʳ
 *   INDEF_LLONG ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥
 *  TFORMn ɽ򼨤ʤˤϡʸ¿ͤѴͤ
 *  Ǥᤤ֤ͤޤ
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @return  ͡NULLͤξINDEF_LLONG <br>
 *          INDEF_LLONG: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
long long fits_table_col::llvalue( long row_index ) const
{
    if ( this->type_rec == FITS::LONGLONG_T ) {
	if ( row_index < 0 || this->row_size_rec <= row_index )
	    return INDEF_LLONG;
	long long v = this->longlong_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v ) {
	    return INDEF_LLONG;
	}
	else if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return v;
	}
	else {
	    double dv = this->tzero_rec + v * this->tscal_rec;
	    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG )
		return round_d2ll(dv);
	    else return INDEF_LLONG;
	}
    }
    else {
	double dv = this->dvalue(row_index);
	if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(dv);
	else return INDEF_LLONG;
    }
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ͤˡإå TZEROn  TSCALn ȿǤ¿ͤ˺Ǥᤤͤ
 *  ֤ޤͤ long long ɽǤϰϤʤ硤
 *  INDEF_LLONG ֤ޤ<br>
 *  NULLͤξϡINDEF_LLONG ֤ޤХʥơ֥
 *  ơ֥ TNULLn ͤˡͤפNULLͤȤ
 *  <br>
 *  Υξϡͤ 'T' ʤ 1 ͤ 'F' ʤ 0 򡤤ʳ
 *   INDEF_LLONG ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥
 *  TFORMn ɽ򼨤ʤˤϡʸ¿ͤѴͤ
 *  Ǥᤤ֤ͤޤ
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @return  ͡NULLͤξINDEF_LLONG <br>
 *          INDEF_LLONG: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
long long fits_table_col::llvalue( long row_index,
			     const char *elem_name, long repetition_idx ) const
{
    if ( this->type_rec == FITS::LONGLONG_T ) {
	long elem_index = this->elem_index_rec.index(elem_name,0);
	if ( row_index < 0 || this->row_size_rec <= row_index )
	    return INDEF_LLONG;
	if ( elem_index < 0 || repetition_idx < 0 ) return INDEF_LLONG;
	long long v = this->longlong_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v ) {
	    return INDEF_LLONG;
	}
	else if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return v;
	}
	else {
	    double dv = this->tzero_rec + v * this->tscal_rec;
	    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG )
		return round_d2ll(dv);
	    else return INDEF_LLONG;
	}
    }
    else {
	double dv = this->dvalue(row_index,elem_name,repetition_idx);
	if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(dv);
	else return INDEF_LLONG;
    }
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ͤˡإå TZEROn  TSCALn ȿǤ¿ͤ˺Ǥᤤͤ
 *  ֤ޤͤ long long ɽǤϰϤʤ硤
 *  INDEF_LLONG ֤ޤ<br>
 *  NULLͤξϡINDEF_LLONG ֤ޤХʥơ֥
 *  ơ֥ TNULLn ͤˡͤפNULLͤȤ
 *  <br>
 *  Υξϡͤ 'T' ʤ 1 ͤ 'F' ʤ 0 򡤤ʳ
 *   INDEF_LLONG ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥
 *  TFORMn ɽ򼨤ʤˤϡʸ¿ͤѴͤ
 *  Ǥᤤ֤ͤޤ
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @return  ͡NULLͤξINDEF_LLONG <br>
 *          INDEF_LLONG: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
long long fits_table_col::llvalue( long row_index,
				   long elem_index, long repetition_idx ) const
{
    if ( this->type_rec == FITS::LONGLONG_T ) {
	if ( row_index < 0 || this->row_size_rec <= row_index )
	    return INDEF_LLONG;
	if ( elem_index < 0 || repetition_idx < 0 ) return INDEF_LLONG;
	long long v = this->longlong_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v ) {
	    return INDEF_LLONG;
	}
	else if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return v;
	}
	else {
	    double dv = this->tzero_rec + v * this->tscal_rec;
	    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG )
		return round_d2ll(dv);
	    else return INDEF_LLONG;
	}
    }
    else {
	double dv = this->dvalue(row_index,elem_index,repetition_idx);
	if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(dv);
	else return INDEF_LLONG;
    }
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ֤ͤͤޤ֤ͤϡtrue  false ǡԤ NULLͤξ
 *  ޤߤޤХʥơ֥䥢ơ֥ TNULLn 
 *  ˡͤפNULLͤȤޤ<br>
 *  'T''F''\0' 3ͤɬפȤϡ
 *  fits_table_col::logical_value() дؿȤäƤ<br>
 *  Υξ硤ͤ 'T' ʤ true 򡤤Ǥʤʤ false ֤
 *  ޤ<br>
 *  ¿Υξ硤ͤˡإå TZEROn  TSCAL n 
 *  ȿǤ¿ͤ˺Ǥᤤͤ 0 ʤ false 0 Ǥʤʤ true 
 *  ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥ TFORMn 
 *  ɽ򼨤ʤˤϡʸμ¿ͤؤѴԤѴǽ
 *  ʾϤμ¿ͤ˺Ǥᤤͤ 0 ʤ false 0 Ǥʤʤ true 
 *  ֤ޤѴԲǽʾϡƬʸ(Ͻ) 'T' ޤ 't' ξ
 *  ˤ true 򡤤Ǥʤʤ false ֤ޤ<br>
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @return  ͡NULLͤξfalse <br>
 *          false: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
bool fits_table_col::bvalue( long row_index ) const
{
    if ( this->type_rec == FITS::LOGICAL_T ||
	 this->type_rec == FITS::BIT_T )
	return (this->logical_value(row_index) == 'T') ? true : false;
    else if ( this->type_rec == FITS::STRING_T ) {
	if ( row_index < 0 || this->row_size_rec <= row_index ) return false;
	tstring sval;
	double dval;
	size_t epos;
	this->get_string_value(row_index,sval);
	if (sval.length() == 0 && this->tnull_is_set_rec == false) return false;
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return false;
	}
	REMOVE_SPC(sval);
	dval = sval.strtod(&epos);
	if ( epos == 0 ) {
	    if ( sval.cchr(0) == 'T' || sval.cchr(0) == 't' ) return true;
	    else return false;
	}
	else {
	    if ( !isfinite(dval) ) return false;
	    if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGDI") == 0 ) {
		dval = this->tzero_rec + dval * this->tscal_rec;
	    }
	    return (round(dval) != 0.0) ? true : false;
	}
    }
    else {
	double v = this->dvalue(row_index);
	if ( !isfinite(v) ) return false;
	else if ( round(v) != 0.0 ) return true;
	else return false;
    }
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ֤ͤͤޤ֤ͤϡtrue  false ǡԤ NULLͤξ
 *  ޤߤޤХʥơ֥䥢ơ֥ TNULLn 
 *  ˡͤפNULLͤȤޤ<br>
 *  'T''F''\0' 3ͤɬפȤϡ
 *  fits_table_col::logical_value() дؿȤäƤ<br>
 *  Υξ硤ͤ 'T' ʤ true 򡤤Ǥʤʤ false ֤
 *  ޤ<br>
 *  ¿Υξ硤ͤˡإå TZEROn  TSCAL n 
 *  ȿǤ¿ͤ˺Ǥᤤͤ 0 ʤ false 0 Ǥʤʤ true 
 *  ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥ TFORMn 
 *  ɽ򼨤ʤˤϡʸμ¿ͤؤѴԤѴǽ
 *  ʾϤμ¿ͤ˺Ǥᤤͤ 0 ʤ false 0 Ǥʤʤ true 
 *  ֤ޤѴԲǽʾϡƬʸ(Ͻ) 'T' ޤ 't' ξ
 *  ˤ true 򡤤Ǥʤʤ false ֤ޤ<br>
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @return  ͡NULLͤξfalse <br>
 *          false: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
bool fits_table_col::bvalue( long row_index,
			     const char *elem_name, long repetition_idx ) const
{
    if ( this->type_rec == FITS::LOGICAL_T ||
	 this->type_rec == FITS::BIT_T )
	return (this->logical_value(row_index,elem_name,repetition_idx) == 'T') ? true : false;
    else if ( this->type_rec == FITS::STRING_T ) {
	long elem_index = this->elem_index_rec.index(elem_name,0);
	if ( row_index < 0 || this->row_size_rec <= row_index ) return false;
	if ( elem_index < 0 || repetition_idx < 0 ) return false;
	tstring sval;
	double dval;
	size_t epos;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if (sval.length() == 0 && this->tnull_is_set_rec == false) return false;
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return false;
	}
	REMOVE_SPC(sval);
	dval = sval.strtod(&epos);
	if ( epos == 0 ) {
	    if ( sval.cchr(0) == 'T' || sval.cchr(0) == 't' ) return true;
	    else return false;
	}
	else {
	    if ( !isfinite(dval) ) return false;
	    if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGDI") == 0 ) {
		dval = this->tzero_rec + dval * this->tscal_rec;
	    }
	    return (round(dval) != 0.0) ? true : false;
	}
    }
    else {
	double v = this->dvalue(row_index,elem_name,repetition_idx);
	if ( !isfinite(v) ) return false;
	else if ( round(v) != 0.0 ) return true;
	else return false;
    }
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ֤ͤͤޤ֤ͤϡtrue  false ǡԤ NULLͤξ
 *  ޤߤޤХʥơ֥䥢ơ֥ TNULLn 
 *  ˡͤפNULLͤȤޤ<br>
 *  'T''F''\0' 3ͤɬפȤϡ
 *  fits_table_col::logical_value() дؿȤäƤ<br>
 *  Υξ硤ͤ 'T' ʤ true 򡤤Ǥʤʤ false ֤
 *  ޤ<br>
 *  ¿Υξ硤ͤˡإå TZEROn  TSCAL n 
 *  ȿǤ¿ͤ˺Ǥᤤͤ 0 ʤ false 0 Ǥʤʤ true 
 *  ֤ޤ<br>
 *  Хʥơ֥ʸ󷿤Υξȡơ֥ TFORMn 
 *  ɽ򼨤ʤˤϡʸμ¿ͤؤѴԤѴǽ
 *  ʾϤμ¿ͤ˺Ǥᤤͤ 0 ʤ false 0 Ǥʤʤ true 
 *  ֤ޤѴԲǽʾϡƬʸ(Ͻ) 'T' ޤ 't' ξ
 *  ˤ true 򡤤Ǥʤʤ false ֤ޤ<br>
 *  Υǥåϡ0ϤޤǤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @return  ͡NULLͤξfalse <br>
 *          false: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
bool fits_table_col::bvalue( long row_index,
			     long elem_index, long repetition_idx ) const
{
    if ( this->type_rec == FITS::LOGICAL_T ||
	 this->type_rec == FITS::BIT_T )
	return (this->logical_value(row_index,elem_index,repetition_idx) == 'T') ? true : false;
    else if ( this->type_rec == FITS::STRING_T ) {
	if ( row_index < 0 || this->row_size_rec <= row_index ) return false;
	if ( elem_index < 0 || repetition_idx < 0 ) return false;
	tstring sval;
	double dval;
	size_t epos;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if (sval.length() == 0 && this->tnull_is_set_rec == false) return false;
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return false;
	}
	REMOVE_SPC(sval);
	dval = sval.strtod(&epos);
	if ( epos == 0 ) {
	    if ( sval.cchr(0) == 'T' || sval.cchr(0) == 't' ) return true;
	    else return false;
	}
	else {
	    if ( !isfinite(dval) ) return false;
	    if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGDI") == 0 ) {
		dval = this->tzero_rec + dval * this->tscal_rec;
	    }
	    return (round(dval) != 0.0) ? true : false;
	}
    }
    else {
	double v = this->dvalue(row_index,elem_index,repetition_idx);
	if ( !isfinite(v) ) return false;
	else if ( round(v) != 0.0 ) return true;
	else return false;
    }
}


#define DO_LL(ll_v) { \
	    if ( this->tany.cchr(TDISP_IDX,0) == 'O' ) {	/* Octal */ \
		this->tmp_str_buf->printf(this->fmt_str.cstr(),ll_v); \
		size_t spn = this->tmp_str_buf->strspn(1,' '); \
		if ( 0 < spn ) { \
		    this->tmp_str_buf->at(spn) = '0'; \
		    this->str_buf->assign(this->tmp_str_buf->cstr() + 1); \
		} \
		else { \
		    this->str_buf->assign(this->tmp_str_buf->cstr()); \
		} \
	    } \
	    else if ( this->tany.cchr(TDISP_IDX,0) == 'Z' ) {	/* Hexi */ \
		this->tmp_str_buf->printf(this->fmt_str.cstr(),ll_v); \
		size_t spn = this->tmp_str_buf->strspn(2,' '); \
		if ( 1 < spn ) { \
		    this->tmp_str_buf->at(spn) = '0'; \
		    this->tmp_str_buf->at(spn+1) = 'x'; \
		    this->str_buf->assign(this->tmp_str_buf->cstr() + 2); \
		} \
		else if ( 1 == spn ) { \
		    this->tmp_str_buf->at(spn) = '0'; \
		    this->tmp_str_buf->at(spn+1) = 'x'; \
		    this->str_buf->assign(this->tmp_str_buf->cstr() + 1); \
		} \
		else { \
		    this->str_buf->assign(this->tmp_str_buf->cstr()); \
		} \
	    } \
	    else { \
		this->str_buf->printf(this->fmt_str.cstr(),ll_v); \
	    } \
}

#define DO_LL_TS(ll_v,ts) { \
	    if ( this->tany.cchr(TDISP_IDX,0) == 'O' ) {	/* Octal */ \
		ts.printf(this->fmt_str.cstr(),ll_v); \
		size_t spn = ts.strspn(1,' '); \
		if ( 0 < spn ) { \
		    ts.at(spn) = '0'; \
		    ts.erase(0,1); \
		} \
	    } \
	    else if ( this->tany.cchr(TDISP_IDX,0) == 'Z' ) {	/* Hexi */ \
		ts.printf(this->fmt_str.cstr(),ll_v); \
		size_t spn = ts.strspn(2,' '); \
		if ( 1 < spn ) { \
		    ts.at(spn) = '0';  ts.at(spn+1) = 'x'; \
		    ts.erase(0,2); \
		} \
		else if ( 1 == spn ) { \
		    ts.at(spn) = '0';  ts.at(spn+1) = 'x'; \
		    ts.erase(0,1); \
		} \
	    } \
	    else { \
		ts.printf(this->fmt_str.cstr(),ll_v); \
	    } \
}

#define DO_IZOFEGD(val) { \
	if ( this->tany.cchr(TDISP_IDX,0) == 'I' || \
	     this->tany.cchr(TDISP_IDX,0) == 'Z' || \
	     this->tany.cchr(TDISP_IDX,0) == 'O' ) { \
	    double dv = this->tzero_rec + val * this->tscal_rec; \
	    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG ) { \
	        long long ll_v = round_d2ll(dv); \
	        DO_LL(ll_v); \
	    } \
	    else { \
		goto ret_null; \
	    } \
	} \
	else {	/* 'F','E','G' and 'D' */ \
	    double v = this->tzero_rec + val * this->tscal_rec; \
	    if ( isfinite(v) ) { \
		this->str_buf->printf(this->fmt_str.cstr(),v); \
		if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) { \
		    ssize_t ep = this->str_buf->strchr('E'); \
		    if ( 0 <= ep ) this->str_buf->at(ep) = 'D'; \
		} \
	    } \
	    else { \
		goto ret_null; \
	    } \
	} \
}

#define DO_IZOFEGD_TS(val,ts) { \
	if ( this->tany.cchr(TDISP_IDX,0) == 'I' || \
	     this->tany.cchr(TDISP_IDX,0) == 'Z' || \
	     this->tany.cchr(TDISP_IDX,0) == 'O' ) { \
	    double dv = this->tzero_rec + val * this->tscal_rec; \
	    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG ) { \
		long long ll_v = round_d2ll(dv); \
		DO_LL_TS(ll_v,ts); \
	    } \
	    else { \
		goto ret_null; \
	    } \
	} \
	else {	/* 'F','E','G' and 'D' */ \
	    double v = this->tzero_rec + val * this->tscal_rec; \
	    if ( isfinite(v) ) { \
		ts.printf(this->fmt_str.cstr(),v); \
		if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) { \
		    ssize_t ep = ts.strchr('E'); \
		    if ( 0 <= ep ) ts.at(ep) = 'D'; \
		} \
	    } \
	    else { \
		goto ret_null; \
	    } \
	} \
}


/* string */
/**
 * @brief  ͤʸͤǼ (٥)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @return  : ｪλNULLͤξ NULLʸ <br>
 *          NULLʸ: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 * @attention  ĹϥݡȤƤޤ
 */
const char *fits_table_col::svalue( long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto ret_null;

    if ( this->type_rec == FITS::STRING_T ) {
	const char *d_ptr = (const char *)this->data_ptr_rec;
	this->tmp_str_buf->assign(d_ptr +
		this->bytes_rec * this->elem_size_rec * row_index,
		this->bytes_rec);
	/* */
	if ( (this->tmp_str_buf->cchr(0) == '\0' && this->tnull_is_set_rec == false) ||
	     (this->tnull_is_set_rec == true &&
	      this->tany.at_cs(TNULL_IDX).strcmp(this->tmp_str_buf->cstr()) == 0) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.cchr(TDISP_IDX,0) == 'A' ) {
		this->str_buf->printf(this->fmt_str.cstr(),
				      this->tmp_str_buf->cstr());
	    }
	    else if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		double v;
		tstring sv;
		sv.assign(this->tmp_str_buf->cstr());
		REMOVE_SPC(sv);
		v = this->tzero_rec + sv.atof() * this->tscal_rec;
		if ( isfinite(v) ) {
		    this->str_buf->printf(this->fmt_str.cstr(),v);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->str_buf->strchr('E');
			if ( 0 <= ep ) this->str_buf->at(ep) = 'D';
		    }
		}
		else {
		    goto ret_null;
		}
	    }
	    else if ( this->tany.at_cs(TDISP_IDX).strpbrk("I") == 0 ) {
		double v;
		tstring sv;
		sv.assign(this->tmp_str_buf->cstr());
		REMOVE_SPC(sv);
		v = this->tzero_rec + sv.atof() * this->tscal_rec;
		if ( isfinite(v) && MIN_DOUBLE_ROUND_LLONG <= v && v <= MAX_DOUBLE_ROUND_LLONG ) {
		    this->str_buf->printf(this->fmt_str.cstr(),round_d2ll(v));
		}
		else {
		    goto ret_null;
		}
	    }
	    else if ( this->tany.cchr(TDISP_IDX,0) == 'L' ) {
		this->str_buf->printf(this->fmt_str.cstr(),
				      this->tmp_str_buf->cstr());
	    }
	    else {
		this->str_buf->assign(this->tmp_str_buf->cstr());
	    }
	}
	else {
	    this->str_buf->assign(this->tmp_str_buf->cstr());
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long val = this->long_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%ld",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.10G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long val = this->longlong_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%lld",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.15G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short val = this->short_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%hd",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double val = this->double_value(row_index);
	if ( !isfinite(val) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    double v = this->tzero_rec + val * this->tscal_rec;
	    if ( isfinite(v) ) this->str_buf->printf("%.15G",v);
	    else goto ret_null;
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float val = this->float_value(row_index);
	if ( !isfinite(val) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%.7G",(double)val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char val = this->byte_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%hhu",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( 0 < this->fmt_str.length() ) {
	    if ( v == 'T' ) {
		this->str_buf->printf(this->fmt_str.cstr(),"T");
	    }
	    else if ( v == 'F' ) {
		this->str_buf->printf(this->fmt_str.cstr(),"F");
	    }
	    else {
		this->str_buf->printf(this->fmt_nullstr.cstr(),
				      this->null_svalue());
	    }
	}
	else {
	    if (v == 'T') this->str_buf->assign("T");
	    else if (v == 'F') this->str_buf->assign("F");
	    else this->str_buf->assign(this->null_svalue());
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v = this->bit_value(row_index);
	if ( 0 < this->fmt_str.length() ) {
	    long long ll_v=v; DO_LL(ll_v);
	}
	else {
	    this->str_buf->printf("%lld",v);
	}
	return this->str_buf->cstr();
    }
 ret_null:
    if ( 0 < this->fmt_nullstr.length() ) 
	this->str_buf->printf(this->fmt_nullstr.cstr(), this->null_svalue());
    else
	this->str_buf->assign(this->null_svalue());
    return this->str_buf->cstr();
}

/**
 * @brief  ͤʸͤǼ (٥)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @return  : ｪλNULLͤξ NULLʸ <br>
 *          NULLʸ: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 * @attention  ĹϥݡȤƤޤ
 */
const char *fits_table_col::svalue( long row_index,
				    const char *elem_name, 
				    long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( row_index < 0 || this->row_size_rec <= row_index ) goto ret_null;
	if ( elem_index < 0 || repetition_idx < 0 ) goto ret_null;
	v = this->bit_value( row_index, elem_index, repetition_idx, 0 );
	if ( 0 < this->fmt_str.length() ) {
	    long long ll_v=v;
	    DO_LL(ll_v);
	}
	else {
	    this->str_buf->printf("%lld",v);
	}
	return this->str_buf->cstr();
    }
    else {
	return this->svalue( row_index, elem_index, repetition_idx );
    }
 ret_null:
    if ( 0 < this->fmt_nullstr.length() ) 
	this->str_buf->printf(this->fmt_nullstr.cstr(), this->null_svalue());
    else
	this->str_buf->assign(this->null_svalue());
    return this->str_buf->cstr();
}

/**
 * @brief  ͤʸͤǼ (٥)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @return  : ｪλNULLͤξ NULLʸ <br>
 *          NULLʸ: ʾ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 * @attention  ĹϥݡȤƤޤ
 */
const char *fits_table_col::svalue( long row_index,
				    long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto ret_null;
    if ( elem_index < 0 || repetition_idx < 0 ) goto ret_null;

    if ( this->type_rec == FITS::STRING_T ) {
	const char *d_ptr = (const char *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	this->tmp_str_buf->assign(d_ptr + 
	      this->bytes_rec * (this->elem_size_rec * row_index + e_idx),
	      this->bytes_rec);
	/* */
	if ( (this->tmp_str_buf->cchr(0) == '\0' && this->tnull_is_set_rec == false) ||
	     (this->tnull_is_set_rec == true &&
	      this->tany.at_cs(TNULL_IDX).strcmp(this->tmp_str_buf->cstr()) == 0) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.cchr(TDISP_IDX,0) == 'A' ) {
		this->str_buf->printf(this->fmt_str.cstr(),
				      this->tmp_str_buf->cstr());
	    }
	    else if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		double v;
		tstring sv;
		sv.assign(this->tmp_str_buf->cstr());
		REMOVE_SPC(sv);
		v = this->tzero_rec + sv.atof() * this->tscal_rec;
		if ( isfinite(v) ) {
		    this->str_buf->printf(this->fmt_str.cstr(),v);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->str_buf->strchr('E');
			if ( 0 <= ep ) this->str_buf->at(ep) = 'D';
		    }
		}
		else {
		    goto ret_null;
		}
	    }
	    else if ( this->tany.at_cs(TDISP_IDX).strpbrk("I") == 0 ) {
		double v;
		tstring sv;
		sv.assign(this->tmp_str_buf->cstr());
		REMOVE_SPC(sv);
		v = this->tzero_rec + sv.atof() * this->tscal_rec;
		if ( isfinite(v) && MIN_DOUBLE_ROUND_LLONG <= v && v <= MAX_DOUBLE_ROUND_LLONG ) {
		    this->str_buf->printf(this->fmt_str.cstr(),round_d2ll(v));
		}
		else {
		    goto ret_null;
		}
	    }
	    else if ( this->tany.cchr(TDISP_IDX,0) == 'L' ) {
		this->str_buf->printf(this->fmt_str.cstr(),
				      this->tmp_str_buf->cstr());
	    }
	    else {
		this->str_buf->assign(this->tmp_str_buf->cstr());
	    }
	}
	else {
	    this->str_buf->assign(this->tmp_str_buf->cstr());
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long val = this->long_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%ld",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.10G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long val = 
	    this->longlong_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%lld",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.15G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short val = this->short_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%hd",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double val = this->double_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(val) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    double v = this->tzero_rec + val * this->tscal_rec;
	    if ( isfinite(v) ) this->str_buf->printf("%.15G",v);
	    else goto ret_null;
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float val = this->float_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(val) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%.7G",(double)val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char val = 
	    this->byte_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD(val);
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		this->str_buf->printf("%hhu",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) this->str_buf->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( 0 < this->fmt_str.length() ) {
	    if ( v == 'T' ) {
		this->str_buf->printf(this->fmt_str.cstr(),"T");
	    }
	    else if ( v == 'F' ) {
		this->str_buf->printf(this->fmt_str.cstr(),"F");
	    }
	    else {
		this->str_buf->printf(this->fmt_nullstr.cstr(),
				      this->null_svalue());
	    }
	}
	else {
	    if (v == 'T') this->str_buf->assign("T");
	    else if (v == 'F') this->str_buf->assign("F");
	    else this->str_buf->assign(this->null_svalue());
	}
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v = this->bit_value(row_index,elem_index,repetition_idx);
	if ( 0 < this->fmt_str.length() ) {
	    long long ll_v=v; DO_LL(ll_v);
	}
	else {
	    this->str_buf->printf("%lld",v);
	}
	return this->str_buf->cstr();
    }
 ret_null:
    if ( 0 < this->fmt_nullstr.length() ) 
	this->str_buf->printf(this->fmt_nullstr.cstr(), this->null_svalue());
    else
	this->str_buf->assign(this->null_svalue());
    return this->str_buf->cstr();
}

/* */

/* get string */

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   dest ͤγǼ (return)
 * @return  : ｪλNULLͤξ NULLʸ <br>
 *          NULLʸ: ʾ <br>
 *          NULL: dest  NULL ξ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
const char *fits_table_col::get_svalue( long row_index,
					tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    if ( row_index < 0 || this->row_size_rec <= row_index ) goto ret_null;

    if ( this->type_rec == FITS::STRING_T ) {
	const char *d_ptr = (const char *)this->data_ptr_rec;
	dest->assign(d_ptr + 
		    this->bytes_rec * this->elem_size_rec * row_index,
		    this->bytes_rec);
	if ( dest->length() == 0 && this->tnull_is_set_rec == false ) {
	    goto ret_null;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(dest->cstr()) == 0 ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.cchr(TDISP_IDX,0) == 'A' ) {
		dest->assignf(this->fmt_str.cstr(),dest->cstr());
	    }
	    else if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		double v;
		REMOVE_SPC((*dest));
		v = this->tzero_rec + dest->atof() * this->tscal_rec;
		if ( isfinite(v) ) {
		    dest->assignf(this->fmt_str.cstr(),v);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = dest->strchr('E');
			if ( 0 <= ep ) dest->at(ep) = 'D';
		    }
		}
		else {
		    goto ret_null;
		}
	    }
	    else if ( this->tany.at_cs(TDISP_IDX).strpbrk("I") == 0 ) {
		double v;
		REMOVE_SPC((*dest));
		v = this->tzero_rec + dest->atof() * this->tscal_rec;
		if ( isfinite(v) && MIN_DOUBLE_ROUND_LLONG <= v && v <= MAX_DOUBLE_ROUND_LLONG ) {
		    dest->assignf(this->fmt_str.cstr(),round_d2ll(v));
		}
		else {
		    goto ret_null;
		}
	    }
	    else if ( this->tany.cchr(TDISP_IDX,0) == 'L' ) {
		dest->assignf(this->fmt_str.cstr(),dest->cstr());
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long val = this->long_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%ld",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.10G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long val = this->longlong_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%lld",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.15G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short val = this->short_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%hd",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double val = this->double_value(row_index);
	if ( !isfinite(val) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    double v = this->tzero_rec + val * this->tscal_rec;
	    if ( isfinite(v) ) dest->printf("%.15G",v);
	    else goto ret_null;
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float val = this->float_value(row_index);
	if ( !isfinite(val) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%.7G",(double)val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char val = this->byte_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%hhu",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( 0 < this->fmt_str.length() ) {
	    if ( v == 'T' ) {
		dest->printf(this->fmt_str.cstr(),"T");
	    }
	    else if ( v == 'F' ) {
		dest->printf(this->fmt_str.cstr(),"F");
	    }
	    else {
		dest->printf(this->fmt_nullstr.cstr(), this->null_svalue());
	    }
	}
	else {
	    if (v == 'T') dest->assign("T");
	    else if (v == 'F') dest->assign("F");
	    else dest->assign(this->null_svalue());
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v = this->bit_value(row_index);
	if ( 0 < this->fmt_str.length() ) {
	    long long ll_v=v; DO_LL_TS(ll_v,(*dest));
	}
	else {
	    dest->printf("%lld",v);
	}
	return dest->cstr();
    }
 ret_null:
    if ( 0 < this->fmt_nullstr.length() ) 
	dest->assignf(this->fmt_nullstr.cstr(), this->null_svalue());
    else
	dest->assign(this->null_svalue());
    return dest->cstr();
}

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   dest ͤγǼ (return)
 * @return  : ｪλNULLͤξ NULLʸ <br>
 *          NULLʸ: ʾ <br>
 *          NULL: dest  NULL ξ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
const char *fits_table_col::get_svalue( long row_index,
					const char *elem_name, 
					tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    return this->get_svalue( row_index, elem_name, 0, dest );
}

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @param   dest ͤγǼ (return)
 * @return  : ｪλNULLͤξ NULLʸ <br>
 *          NULLʸ: ʾ <br>
 *          NULL: dest  NULL ξ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
const char *fits_table_col::get_svalue( long row_index,
					const char *elem_name, 
					long repetition_idx,
					tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( row_index < 0 || this->row_size_rec <= row_index ) goto ret_null;
	if ( elem_index < 0 || repetition_idx < 0 ) goto ret_null;
	v = this->bit_value( row_index, elem_index, repetition_idx, 0 );
	if ( 0 < this->fmt_str.length() ) {
	    long long ll_v=v;
	    DO_LL_TS(ll_v,(*dest));
	}
	else {
	    dest->printf("%lld",v);
	}
	return dest->cstr();
    }
    else {
	return this->get_svalue( row_index, elem_index, repetition_idx, dest );
    }
 ret_null:
    if ( 0 < this->fmt_nullstr.length() ) 
	dest->assignf(this->fmt_nullstr.cstr(), this->null_svalue());
    else
	dest->assign(this->null_svalue());
    return dest->cstr();
}

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   dest ͤγǼ (return)
 * @return  : ｪλNULLͤξ NULLʸ <br>
 *          NULLʸ: ʾ <br>
 *          NULL: dest  NULL ξ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
const char *fits_table_col::get_svalue( long row_index,
					long elem_index, tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    return this->get_svalue( row_index, elem_index, 0, dest );
}

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @param   dest ͤγǼ (return)
 * @return  : ｪλNULLͤξ NULLʸ <br>
 *          NULLʸ: ʾ <br>
 *          NULL: dest  NULL ξ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
const char *fits_table_col::get_svalue( long row_index,
					long elem_index, 
					long repetition_idx,
					tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    if ( row_index < 0 || this->row_size_rec <= row_index ) goto ret_null;
    if ( elem_index < 0 || repetition_idx < 0 ) goto ret_null;

    if ( this->type_rec == FITS::STRING_T ) {
	const char *d_ptr = (const char *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	dest->assign(d_ptr + 
	      this->bytes_rec * (this->elem_size_rec * row_index + e_idx),
	      this->bytes_rec);
	if ( dest->length() == 0 && this->tnull_is_set_rec == false ) {
	    goto ret_null;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(dest->cstr()) == 0 ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.cchr(TDISP_IDX,0) == 'A' ) {
		dest->assignf(this->fmt_str.cstr(),dest->cstr());
	    }
	    else if ( this->tany.at_cs(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		double v;
		REMOVE_SPC((*dest));
		v = this->tzero_rec + dest->atof() * this->tscal_rec;
		if ( isfinite(v) ) {
		    dest->assignf(this->fmt_str.cstr(),v);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = dest->strchr('E');
			if ( 0 <= ep ) dest->at(ep) = 'D';
		    }
		}
		else {
		    goto ret_null;
		}
	    }
	    else if ( this->tany.at_cs(TDISP_IDX).strpbrk("I") == 0 ) {
		double v;
		REMOVE_SPC((*dest));
		v = this->tzero_rec + dest->atof() * this->tscal_rec;
		if ( isfinite(v) && MIN_DOUBLE_ROUND_LLONG <= v && v <= MAX_DOUBLE_ROUND_LLONG ) {
		    dest->assignf(this->fmt_str.cstr(),round_d2ll(v));
		}
		else {
		    goto ret_null;
		}
	    }
	    else if ( this->tany.cchr(TDISP_IDX,0) == 'L' ) {
		dest->assignf(this->fmt_str.cstr(),dest->cstr());
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long val = this->long_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%ld",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.10G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long val = 
	    this->longlong_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%lld",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.15G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short val = 
	    this->short_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%hd",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double val = this->double_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(val) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    double v = this->tzero_rec + val * this->tscal_rec;
	    if ( isfinite(v) ) dest->printf("%.15G",v);
	    else goto ret_null;
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float val = this->float_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(val) ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%.7G",(double)val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char val = 
	    this->byte_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec ) {
	    goto ret_null;
	}
	else if ( 0 < this->fmt_str.length() ) {
	    DO_IZOFEGD_TS(val,(*dest));
	}
	else {
	    if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
		dest->printf("%hhu",val);
	    }
	    else {
		double v = this->tzero_rec + val * this->tscal_rec;
		if ( isfinite(v) ) dest->printf("%.7G",v);
		else goto ret_null;
	    }
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( 0 < this->fmt_str.length() ) {
	    if ( v == 'T' ) {
		dest->printf(this->fmt_str.cstr(),"T");
	    }
	    else if ( v == 'F' ) {
		dest->printf(this->fmt_str.cstr(),"F");
	    }
	    else {
		dest->printf(this->fmt_nullstr.cstr(), this->null_svalue());
	    }
	}
	else {
	    if (v == 'T') dest->assign("T");
	    else if (v == 'F') dest->assign("F");
	    else dest->assign(this->null_svalue());
	}
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v = this->bit_value(row_index,elem_index,repetition_idx);
	if ( 0 < this->fmt_str.length() ) {
	    long long ll_v=v; DO_LL_TS(ll_v,(*dest));
	}
	else {
	    dest->printf("%lld",v);
	}
	return dest->cstr();
    }
 ret_null:
    if ( 0 < this->fmt_nullstr.length() ) 
	dest->assignf(this->fmt_nullstr.cstr(), this->null_svalue());
    else
	dest->assign(this->null_svalue());
    return dest->cstr();
}

/**
 * @deprecated  侩<br>
 *              fits_table_col::get_svalue( long, tstring * ) <br>
 *              򤪻Ȥ
 */
const char *fits_table_col::get_svalue( long row_index, 
					tstring &dest ) const
{
    return this->get_svalue(row_index, &dest);
}

/**
 * @deprecated  侩<br>
 *             fits_table_col::get_svalue( long, const char *, tstring * ) <br>
 *             򤪻Ȥ
 */
const char *fits_table_col::get_svalue( long row_index,
					const char *elem_name, 
					tstring &dest ) const
{
    return this->get_svalue(row_index, elem_name, 0, &dest);
}

/**
 * @deprecated  侩<br>
 *       fits_table_col::get_svalue( long, const char *, long, tstring * ) <br>
 *       򤪻Ȥ
 */
const char *fits_table_col::get_svalue( long row_index,
					const char *elem_name, 
					long repetition_idx,
					tstring &dest ) const
{
    return this->get_svalue(row_index, elem_name, repetition_idx, &dest);
}

/**
 * @deprecated  侩<br>
 *              fits_table_col::get_svalue( long, long, tstring * ) <br>
 *              򤪻Ȥ
 */
const char *fits_table_col::get_svalue( long row_index,
					long elem_index, tstring &dest ) const
{
    return this->get_svalue(row_index, elem_index, 0, &dest);
}

/**
 * @deprecated  侩<br>
 *              fits_table_col::get_svalue( long, long, long, tstring * ) <br>
 *              򤪻Ȥ
 */
const char *fits_table_col::get_svalue( long row_index,
					long elem_index, 
					long repetition_idx,
					tstring &dest ) const
{
    return this->get_svalue(row_index, elem_index, repetition_idx, &dest);
}

#undef DO_IZOFEGD_TS
#undef DO_IZOFEGD
#undef DO_LL_TS
#undef DO_LL

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
ssize_t fits_table_col::get_svalue( long row_index, 
				    char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_svalue( row_index, &sval );
    return sval.getstr(dest_buf,buf_size);
}

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
ssize_t fits_table_col::get_svalue( long row_index,
				    const char *elem_name, 
				    char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_svalue( row_index, elem_name, 0, &sval );
    return sval.getstr(dest_buf,buf_size);
}

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
ssize_t fits_table_col::get_svalue( long row_index,
				    const char *elem_name, long repetition_idx,
				    char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_svalue( row_index, elem_name, repetition_idx, &sval );
    return sval.getstr(dest_buf,buf_size);
}

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
ssize_t fits_table_col::get_svalue( long row_index, long elem_index, 
				    char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_svalue( row_index, elem_index, 0, &sval );
    return sval.getstr(dest_buf,buf_size);
}

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  FITS˴ŤƼѴԤʤäͤʸ֤ͤޤ<br>
 *  NULLͤξ硤fits_table_col::assign_null_svalue() ǻꤵ줿 NULL ʸ
 *  ֤ޤ(NULLʸνͤ "NULL" Ǥ)Хʥơ֥
 *  䥢ơ֥ TNULLn ͤˡͤפ
 *  NULLͤȤޤ<br>
 *  Хʥơ֥ξ <br>
 *  ¿Υξ硤ͤإå TZEROn  TSCALn ͤ
 *  ѴʸѴ֤ͤޤ<br>
 *  TDISPn λ꤬ϡTDISPn λ˽äƥեޥåȤ줿ʸ
 *  ֤ޤ<br>
 *  TDISPn λ̵꤬硤¿Υ TZEROnTSCALn ꤵƤ
 *  ΥǤϡʲprintf()ؿΥեޥåȤ˽äѴʸ
 *  ֤ޤ<br>
 *  FITS::FLOAT_T                    ... "%.7G" <br>
 *  FITS::DOUBLE_T, FITS::LONGLONG_T ... "%.15G" <br>
 *  FITS::LONG_T                     ... "%.10G" <br>
 *  嵭ʳΥॿ           ... "%.7G" <br>
 *  ơ֥ξ <br>
 *  TFORMn ɽ򼨤ʤˤϡʸ TFORMn ǥեޥå
 *  ʸ֤ޤ(TFORMn λ̵꤬ϡʸ򤽤Τޤ֤
 *  ޤ)<br>
 *  TFORMn ɽ򼨤ˤϡʸ¿ͤѴͤ
 *  TZEROn  TSCALn Ѵͤ TFORMn ǥեޥåȤʸ֤
 *  ޤ<br>
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx TDIMn 2ܤΥǥå
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
ssize_t fits_table_col::get_svalue( long row_index,
				    long elem_index, long repetition_idx,
				    char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_svalue( row_index, elem_index, repetition_idx, &sval );
    return sval.getstr(dest_buf,buf_size);
}

/*
 * low-level APIs to read a cell.
 */

/* logical */
/**
 * @brief  ͤͤǼ (٥)
 *
 *  ֤ͤϡ'T''F''\0' (NULL) 3Ǥ<br>
 *  η FITS::LOGICAL_T (TFORMn λ L ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
int fits_table_col::logical_value( long row_index ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) return '\0';

    if ( this->type_rec == FITS::LOGICAL_T ) {
	fits::logical_t v;
	const fits::logical_t *d_ptr = (fits::logical_t *)this->data_ptr_rec;
	v = d_ptr[this->elem_size_rec * row_index];
	if ( v == 'T' ) return 'T';
	else if ( v == 'F' ) return 'F';
	else return '\0';
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long val = this->long_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec )
	    return '\0';
	return (val != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short val = this->short_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec )
	    return '\0';
	return (val != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long val = this->longlong_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec )
	    return '\0';
	return (val != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char val = this->byte_value(row_index);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec )
	    return '\0';
	return (val != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v = this->double_value(row_index);
	if ( !isfinite(v) ) return '\0';
	else return (round(v) != 0.0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v = this->float_value(row_index);
	if ( !isfinite(v) ) return '\0';
	else return (round(v) != 0.0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return (this->bit_value(row_index) != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double dval;
	size_t epos;
	this->get_string_value(row_index,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return '\0';
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return '\0';
	}
	REMOVE_SPC(sval);
	dval = sval.strtod(&epos);
	if ( epos == 0 ) {
	    if ( sval.cchr(0) == 'T' || sval.cchr(0) == 't' ) return 'T';
	    else if ( sval.cchr(0) == 'F' || sval.cchr(0) == 'f' ) return 'F';
	    else return '\0';
	}
	else {
	    if ( !isfinite(dval) ) return '\0';
	    else return (round(dval) != 0.0) ? 'T' : 'F';
	}
    }
    else return '\0';
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ֤ͤϡ'T''F''\0' (NULL) 3Ǥ<br>
 *  η FITS::LOGICAL_T (TFORMn λ L ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
int fits_table_col::logical_value( long row_index,
			     const char *elem_name, long repetition_idx ) const
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	if ( row_index < 0 || this->row_size_rec <= row_index ) return '\0';
	if ( elem_index < 0 || repetition_idx < 0 ) return '\0';
	return (this->bit_value(row_index,elem_index,repetition_idx,0) != 0) ? 'T' : 'F';
    }
    else {
	return this->logical_value( row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ͤͤǼ (٥)
 *
 *  ֤ͤϡ'T''F''\0' (NULL) 3Ǥ<br>
 *  η FITS::LOGICAL_T (TFORMn λ L ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
int fits_table_col::logical_value( long row_index,
				   long elem_index, long repetition_idx ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) return '\0';
    if ( elem_index < 0 || repetition_idx < 0 ) return '\0';

    if ( this->type_rec == FITS::LOGICAL_T ) {
	fits::logical_t v;
	const fits::logical_t *d_ptr = (fits::logical_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	v = d_ptr[this->elem_size_rec * row_index + e_idx];
	if ( v == 'T' ) return 'T';
	else if ( v == 'F' ) return 'F';
	else return '\0';
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long val = this->long_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec )
	    return '\0';
	return (val != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short val = this->short_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec )
	    return '\0';
	return (val != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long val = this->longlong_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec )
	    return '\0';
	return (val != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char val = this->byte_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && val == this->tnull_r_rec )
	    return '\0';
	return (val != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v = this->double_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(v) ) return '\0';
	else return (round(v) != 0.0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v = this->float_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(v) ) return '\0';
	else return (round(v) != 0.0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return (this->bit_value(row_index,elem_index,repetition_idx) != 0) ? 'T' : 'F';
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double dval;
	size_t epos;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return '\0';
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return '\0';
	}
	REMOVE_SPC(sval);
	dval = sval.strtod(&epos);
	if ( epos == 0 ) {
	    if ( sval.cchr(0) == 'T' || sval.cchr(0) == 't' ) return 'T';
	    else if ( sval.cchr(0) == 'F' || sval.cchr(0) == 'f' ) return 'F';
	    else return '\0';
	}
	else {
	    if ( !isfinite(dval) ) return '\0';
	    else return (round(dval) != 0.0) ? 'T' : 'F';
	}
    }
    else return '\0';
}


/* short */
/**
 * @brief  ͤ(short)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (short)֤ޤ<br>
 *  η FITS::SHORT_T (TFORMn λ 'I' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_SHORT Ǥ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
short fits_table_col::short_value( long row_index ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;

    if ( this->type_rec == FITS::SHORT_T ) {
	const fits::short_t *d_ptr = (fits::short_t *)this->data_ptr_rec;
	return d_ptr[this->elem_size_rec * row_index];
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v0 = this->long_value(row_index);
	if ( MIN_SHORT <= v0 && v0 <= MAX_SHORT ) return v0;
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index);
	if ( MIN_SHORT <= v0 && v0 <= MAX_SHORT ) return v0;
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	return this->byte_value(row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_SHORT <= v0 && v0 <= MAX_DOUBLE_ROUND_SHORT )
	    return (short)round_d2l(v0);
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_SHORT <= v0 && v0 <= MAX_FLOAT_ROUND_SHORT )
	    return (short)round_f2l(v0);
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v0 = this->bit_value(row_index);
	if ( MIN_SHORT <= v0 && v0 <= MAX_SHORT ) return v0;
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_short_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_short_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_SHORT <= v0 && v0 <= MAX_DOUBLE_ROUND_SHORT )
	    return (short)round_d2l(v0);
	else return this->tnull_short_rec;
    }
 invalid:
    return this->tnull_short_rec;
}

/**
 * @brief  ͤ(short)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (short)֤ޤ<br>
 *  η FITS::SHORT_T (TFORMn λ 'I' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_SHORT Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
short fits_table_col::short_value( long row_index,
			     const char *elem_name, long repetition_idx ) const
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	long long v0 = this->bit_value(row_index,elem_index,repetition_idx,0);
	if ( MIN_SHORT <= v0 && v0 <= MAX_SHORT ) return v0;
	else return this->tnull_short_rec;
    }
    else {
	return this->short_value( row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ͤ(short)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (short)֤ޤ<br>
 *  η FITS::SHORT_T (TFORMn λ 'I' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_SHORT Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
short fits_table_col::short_value( long row_index,
				   long elem_index, long repetition_idx ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;
    if ( elem_index < 0 || repetition_idx < 0 ) goto invalid;

    if ( this->type_rec == FITS::SHORT_T ) {
	const fits::short_t *d_ptr = (fits::short_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	return d_ptr[this->elem_size_rec * row_index + e_idx];
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v0 = this->long_value(row_index,elem_index,repetition_idx);
	if ( MIN_SHORT <= v0 && v0 <= MAX_SHORT ) return v0;
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index,elem_index,repetition_idx);
	if ( MIN_SHORT <= v0 && v0 <= MAX_SHORT ) return v0;
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	return this->byte_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_SHORT <= v0 && v0 <= MAX_DOUBLE_ROUND_SHORT )
	    return (short)round_d2l(v0);
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_SHORT <= v0 && v0 <= MAX_FLOAT_ROUND_SHORT )
	    return (short)round_f2l(v0);
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v0 = this->bit_value(row_index,elem_index,repetition_idx);
	if ( MIN_SHORT <= v0 && v0 <= MAX_SHORT ) return v0;
	else return this->tnull_short_rec;
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_short_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_short_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_SHORT <= v0 && v0 <= MAX_DOUBLE_ROUND_SHORT )
	    return (short)round_d2l(v0);
	else return this->tnull_short_rec;
    }
 invalid:
    return this->tnull_short_rec;
}


/* long */
/**
 * @brief  ͤ(long)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (long)֤ޤ<br>
 *  η FITS::LONG_T (TFORMn λ 'J' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_LONG Ǥ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 * @note  long 32-bit OS Ǥ 32-bit64-bit OS Ǥ 64-bit Ļ
 *        ۤȤɤǤ
 */
long fits_table_col::long_value( long row_index ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;

    if ( this->type_rec == FITS::LONG_T ) {
	const fits::long_t *d_ptr = (fits::long_t *)this->data_ptr_rec;
	return d_ptr[this->elem_size_rec * row_index];
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	return this->short_value(row_index);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	return this->byte_value(row_index);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index);
	if ( MIN_LONG <= v0 && v0 <= MAX_LONG ) return v0;
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LONG )
	    return round_d2l(v0);
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_LONG <= v0 && v0 <= MAX_FLOAT_ROUND_LONG )
	    return round_f2l(v0);
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v0 = this->bit_value(row_index);
	if ( MIN_LONG <= v0 && v0 <= MAX_LONG ) return v0;
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_long_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_long_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LONG )
	    return round_d2l(v0);
	else return this->tnull_long_rec;
    }
 invalid:
    return this->tnull_long_rec;
}

/**
 * @brief  ͤ(long)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (long)֤ޤ<br>
 *  η FITS::LONG_T (TFORMn λ 'J' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_LONG Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 * @note  long 32-bit OS Ǥ 32-bit64-bit OS Ǥ 64-bit Ļ
 *        ۤȤɤǤ
 */
long fits_table_col::long_value( long row_index,
			     const char *elem_name, long repetition_idx ) const
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	long long v0 = this->bit_value(row_index,elem_index,repetition_idx,0);
	if ( MIN_LONG <= v0 && v0 <= MAX_LONG ) return v0;
	else return this->tnull_long_rec;
    }
    else {
	return this->long_value( row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ͤ(long)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (long)֤ޤ<br>
 *  η FITS::LONG_T (TFORMn λ 'J' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_LONG Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 * @note  long 32-bit OS Ǥ 32-bit64-bit OS Ǥ 64-bit Ļ
 *        ۤȤɤǤ
 */
long fits_table_col::long_value( long row_index,
				 long elem_index, long repetition_idx ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;
    if ( elem_index < 0 || repetition_idx < 0 ) goto invalid;

    if ( this->type_rec == FITS::LONG_T ) {
	const fits::long_t *d_ptr = (fits::long_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	return d_ptr[this->elem_size_rec * row_index + e_idx];
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	return this->short_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	return this->byte_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index,elem_index,repetition_idx);
	if ( MIN_LONG <= v0 && v0 <= MAX_LONG ) return v0;
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LONG )
	    return round_d2l(v0);
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_LONG <= v0 && v0 <= MAX_FLOAT_ROUND_LONG )
	    return round_f2l(v0);
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v0 = this->bit_value(row_index,elem_index,repetition_idx);
	if ( MIN_LONG <= v0 && v0 <= MAX_LONG ) return v0;
	else return this->tnull_long_rec;
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_long_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_long_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LONG )
	    return round_d2l(v0);
	else return this->tnull_long_rec;
    }
 invalid:
    return this->tnull_long_rec;
}


/* longlong */
/**
 * @brief  ͤ(long long)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (long long)֤ޤ<br>
 *  η FITS::LONGLONG_T (TFORMn λ 'K' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_LLONG Ǥ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
long long fits_table_col::longlong_value( long row_index ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;

    if ( this->type_rec == FITS::LONGLONG_T ) {
	const fits::longlong_t *d_ptr = (fits::longlong_t *)this->data_ptr_rec;
	return d_ptr[this->elem_size_rec * row_index];
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	return this->long_value(row_index);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	return this->short_value(row_index);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	return this->byte_value(row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LLONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(v0);
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_LLONG <= v0 && v0 <= MAX_FLOAT_ROUND_LLONG )
	    return round_f2ll(v0);
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->bit_value(row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_longlong_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_longlong_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LLONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(v0);
	else return this->tnull_longlong_rec;
    }
 invalid:
    return this->tnull_longlong_rec;
}

/**
 * @brief  ͤ(long long)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (long long)֤ޤ<br>
 *  η FITS::LONGLONG_T (TFORMn λ 'K' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_LLONG Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
long long fits_table_col::longlong_value( long row_index,
					  const char *elem_name, 
					  long repetition_idx ) const
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	return this->bit_value(row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->longlong_value( row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ͤ(long long)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (long long)֤ޤ<br>
 *  η FITS::LONGLONG_T (TFORMn λ 'K' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_LLONG Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
long long fits_table_col::longlong_value( long row_index,
				   long elem_index, long repetition_idx ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;
    if ( elem_index < 0 || repetition_idx < 0 ) goto invalid;

    if ( this->type_rec == FITS::LONGLONG_T ) {
	const fits::longlong_t *d_ptr = (fits::longlong_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	return d_ptr[this->elem_size_rec * row_index + e_idx];
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	return this->long_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	return this->short_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	return this->byte_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LLONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(v0);
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_LLONG <= v0 && v0 <= MAX_FLOAT_ROUND_LLONG )
	    return round_f2ll(v0);
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->bit_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_longlong_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_longlong_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LLONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(v0);
	else return this->tnull_longlong_rec;
    }
 invalid:
    return this->tnull_longlong_rec;
}


/* float */
/**
 * @brief  ͤ¿(float)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)¿
 *  (float)֤ޤ<br>
 *  η FITS::FLOAT_T (TFORMn λ 'E' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  NULLͤξϡNAN ֤ޤ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλ<br>
 *          NAN: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
float fits_table_col::float_value( long row_index ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;

    if ( this->type_rec == FITS::FLOAT_T ) {
	const fits::float_t *d_ptr = (fits::float_t *)this->data_ptr_rec;
	return d_ptr[this->elem_size_rec * row_index];
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	return this->double_value(row_index);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v0 = this->long_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v0 = this->short_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v0 = this->byte_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return NAN;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->bit_value(row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	this->get_string_value(row_index,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return NAN;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return NAN;
	}
	REMOVE_SPC(sval);
	return (float)(sval.atof());
    }
 invalid:
    return NAN;
}

/**
 * @brief  ͤ¿(float)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)¿
 *  (float)֤ޤ<br>
 *  η FITS::FLOAT_T (TFORMn λ 'E' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  NULLͤξϡNAN ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NAN: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
float fits_table_col::float_value( long row_index,
			     const char *elem_name, long repetition_idx ) const
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	return this->bit_value(row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->float_value( row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ͤ¿(float)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)¿
 *  (float)֤ޤ<br>
 *  η FITS::FLOAT_T (TFORMn λ 'E' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  NULLͤξϡNAN ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NAN: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
float fits_table_col::float_value( long row_index,
				   long elem_index, long repetition_idx ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;
    if ( elem_index < 0 || repetition_idx < 0 ) goto invalid;

    if ( this->type_rec == FITS::FLOAT_T ) {
	const fits::float_t *d_ptr = (fits::float_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	return d_ptr[this->elem_size_rec * row_index + e_idx];
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	return this->double_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v0 = this->long_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v0 = this->short_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v0 = this->byte_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return NAN;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->bit_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return NAN;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return NAN;
	}
	REMOVE_SPC(sval);
	return (float)(sval.atof());
    }
 invalid:
    return NAN;
}


/* double */
/**
 * @brief  ͤ¿(double)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)¿
 *  (double)֤ޤ<br>
 *  η FITS::DOUBLE_T (TFORMn λ 'D' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  NULLͤξϡNAN ֤ޤ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλ<br>
 *          NAN: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
double fits_table_col::double_value( long row_index ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;

    if ( this->type_rec == FITS::DOUBLE_T ) {
	const fits::double_t *d_ptr = (fits::double_t *)this->data_ptr_rec;
	return d_ptr[this->elem_size_rec * row_index];
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	return this->float_value(row_index);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v0 = this->long_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v0 = this->short_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v0 = this->byte_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return NAN;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->bit_value(row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	this->get_string_value(row_index,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return NAN;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return NAN;
	}
	REMOVE_SPC(sval);
	return sval.atof();
    }
 invalid:
    return NAN;
}

/**
 * @brief  ͤ¿(double)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)¿
 *  (double)֤ޤ<br>
 *  η FITS::DOUBLE_T (TFORMn λ 'D' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  NULLͤξϡNAN ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NAN: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
double fits_table_col::double_value( long row_index,
			     const char *elem_name, long repetition_idx ) const
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	return this->bit_value(row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->double_value( row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ͤ¿(double)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)¿
 *  (double)֤ޤ<br>
 *  η FITS::DOUBLE_T (TFORMn λ 'D' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  NULLͤξϡNAN ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NAN: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
double fits_table_col::double_value( long row_index,
				   long elem_index, long repetition_idx ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;
    if ( elem_index < 0 || repetition_idx < 0 ) goto invalid;

    if ( this->type_rec == FITS::DOUBLE_T ) {
	const fits::double_t *d_ptr = (fits::double_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	return d_ptr[this->elem_size_rec * row_index + e_idx];
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	return this->float_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v0 = this->long_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v0 = this->short_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v0 = this->byte_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v0 )
	    return NAN;
	else return v0;
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return NAN;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->bit_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return NAN;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return NAN;
	}
	REMOVE_SPC(sval);
	return sval.atof();
    }
 invalid:
    return NAN;
}


/* byte */
/**
 * @brief  ͤ(byte)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (unsigned char)֤ޤ<br>
 *  η FITS::BYTE_T (TFORMn λ 'B' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_UCHAR Ǥ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
unsigned char fits_table_col::byte_value( long row_index ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;

    if ( this->type_rec == FITS::BYTE_T ) {
	const fits::byte_t *d_ptr = (fits::byte_t *)this->data_ptr_rec;
	return d_ptr[this->elem_size_rec * row_index];
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v0 = this->short_value(row_index);
	if ( MIN_UCHAR <= v0 && v0 <= MAX_UCHAR ) return v0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v0 = this->long_value(row_index);
	if ( MIN_UCHAR <= v0 && v0 <= MAX_UCHAR ) return v0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index);
	if ( MIN_UCHAR <= v0 && v0 <= MAX_UCHAR ) return v0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_UCHAR <= v0 && v0 <= MAX_FLOAT_ROUND_UCHAR )
	    return round_f2l(v0);
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_UCHAR <= v0 && v0 <= MAX_DOUBLE_ROUND_UCHAR )
	    return round_d2l(v0);
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v0 = this->bit_value(row_index);
	if ( MIN_UCHAR <= v0 && v0 <= MAX_UCHAR ) return v0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_byte_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_byte_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_UCHAR <= v0 && v0 <= MAX_DOUBLE_ROUND_UCHAR )
	    return round_d2l(v0);
	else return this->tnull_byte_rec;
    }
 invalid:
    return this->tnull_byte_rec;
}

/**
 * @brief  ͤ(byte)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (unsigned char)֤ޤ<br>
 *  η FITS::BYTE_T (TFORMn λ 'B' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_UCHAR Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
unsigned char fits_table_col::byte_value( long row_index, 
					  const char *elem_name, 
					  long repetition_idx ) const
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	long long v0 = this->bit_value(row_index,elem_index,repetition_idx,0);
	if ( MIN_UCHAR <= v0 && v0 <= MAX_UCHAR ) return v0;
	else return this->tnull_byte_rec;
    }
    else {
	return this->byte_value( row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ͤ(byte)Ǽ (٥)
 *
 *  (إå TZEROn  TSCALn ͤȿǤƤʤ)
 *  (unsigned char)֤ޤ<br>
 *  η FITS::BYTE_T (TFORMn λ 'B' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  ֤ NULLͤϡإå TNULLn ͤ¸ߤϤ͡¸ߤʤ
 *   INDEF_UCHAR Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
unsigned char fits_table_col::byte_value( long row_index,
				   long elem_index, long repetition_idx ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;
    if ( elem_index < 0 || repetition_idx < 0 ) goto invalid;

    if ( this->type_rec == FITS::BYTE_T ) {
	const fits::byte_t *d_ptr = (fits::byte_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	return d_ptr[this->elem_size_rec * row_index + e_idx];
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v0 = this->short_value(row_index,elem_index,repetition_idx);
	if ( MIN_UCHAR <= v0 && v0 <= MAX_UCHAR ) return v0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v0 = this->long_value(row_index,elem_index,repetition_idx);
	if ( MIN_UCHAR <= v0 && v0 <= MAX_UCHAR ) return v0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v0 = this->longlong_value(row_index,elem_index,repetition_idx);
	if ( MIN_UCHAR <= v0 && v0 <= MAX_UCHAR ) return v0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_UCHAR <= v0 && v0 <= MAX_FLOAT_ROUND_UCHAR )
	    return round_f2l(v0);
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_UCHAR <= v0 && v0 <= MAX_DOUBLE_ROUND_UCHAR )
	    return round_d2l(v0);
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v0 = this->bit_value(row_index,elem_index,repetition_idx);
	if ( MIN_UCHAR <= v0 && v0 <= MAX_UCHAR ) return v0;
	else return this->tnull_byte_rec;
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_byte_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_byte_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_UCHAR <= v0 && v0 <= MAX_DOUBLE_ROUND_UCHAR )
	    return round_d2l(v0);
	else return this->tnull_byte_rec;
    }
 invalid:
    return this->tnull_byte_rec;
}


/* bit */
/**
 * @brief  ͤ(bit)֤ (٥)
 *
 *  ͤ(bit)֤ޤ<br>
 *  η FITS::BIT_T (TFORMn λ 'X' ޤ) ξ˺®
 *  ǤΥХǤκüΥӥåȤ֤ͤޤ<br>
 *  η FITS::BIT_T ʳξ֤ NULLͤϡإå TNULLn 
 *  ͤ¸ߤϤ͡¸ߤʤ INDEF_LONG Ǥ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 */
long long fits_table_col::bit_value( long row_index ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;

    if ( this->type_rec == FITS::BIT_T ) {
	/* MSB򸫤 */
	const unsigned char *d_ptr = (const unsigned char *)this->data_ptr_rec;
	unsigned char v = d_ptr[this->full_bytes_rec * row_index];
	if ( (v & 0x080) != 0 ) return 1;
	else return 0;
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	return this->byte_value(row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	return this->short_value(row_index);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	return this->long_value(row_index);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->longlong_value(row_index);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_LLONG <= v0 && v0 <= MAX_FLOAT_ROUND_LLONG )
	    return round_f2ll(v0);
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LLONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(v0);
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_longlong_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_longlong_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LLONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(v0);
	else return this->tnull_longlong_rec;
    }
 invalid:
    return this->tnull_longlong_rec;
}

/**
 * @brief  ͤ(bit)֤ (٥)
 *
 *  ͤ(bit)֤ޤ<br>
 *  η FITS::BIT_T (TFORMn λ 'X' ޤ) ξ˻Ѥ
 *  дؿǡǤդΥӥåǤ鱦ǤդΥӥåȿͤФ
 *  ޤ<br>
 *  η FITS::BIT_T ʳξ֤ NULLͤϡإå TNULLn 
 *  ͤ¸ߤϤ͡¸ߤʤ INDEF_LONG Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @param   nbit ӥå (ά TELEMn ͤΥӥåȥեɻŬ)
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 * @note  ӥåȱ黻Ԥʤᡤ¹ԥȤ礭⤷ޤ
 */
long long fits_table_col::bit_value( long row_index, const char *elem_name,
				     long repetition_idx, int nbit ) const
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    return this->bit_value( row_index, elem_index, repetition_idx, nbit );
}

/**
 * @brief  ͤ(bit)֤ (٥)
 *
 *  ͤ(bit)֤ޤ<br>
 *  η FITS::BIT_T (TFORMn λ 'X' ޤ) ξ˻Ѥ
 *  дؿǡǤդΥӥåǤ鱦ǤդΥӥåȿͤФ
 *  ޤ<br>
 *  η FITS::BIT_T ʳξ֤ NULLͤϡإå TNULLn 
 *  ͤ¸ߤϤ͡¸ߤʤ INDEF_LONG Ǥ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @param   nbit ӥå (ά 1 Ŭ)
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 * @note  ӥåȱ黻Ԥʤᡤ¹ԥȤ礭⤷ޤ
 */
long long fits_table_col::bit_value( long row_index,
				     long elem_index, long repetition_idx,
				     int nbit ) const
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;
    if ( elem_index < 0 || repetition_idx < 0 ) goto invalid;

    if ( this->type_rec == FITS::BIT_T ) {
	const unsigned char *d_ptr = (const unsigned char *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	int n_bit;
	long e_mod, e_div, e_align;
	long long ret_val = 0;
	int i;
	if ( nbit < 1 ) 
	    n_bit = this->bit_size_telem.i_cs(elem_index % this->dcol_size_rec);
	else
	    n_bit = nbit;
	if ( this->elem_size_rec <= e_idx ) {
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	}
	d_ptr += this->full_bytes_rec * row_index;
	e_align = (n_bit-1) % 8;
	e_mod = e_idx % 8;
	e_div = e_idx / 8;
	    /*  
	     *   [case A]
	     *
	     *   01234567 01234567
	     *   xxxXXXXx xxxxxxxx  e_mod:3  n_bit:4
	     *      3210            e_align:3
	     *
	     *   01234567 01234567
	     *   xxxXYYYY YYYYxxxx  e_mod:3  n_bit:9
	     *      07654 3210      e_align:0 n_shift:4
	     *
	     *   01234567 01234567
	     *   xxxxXXYY YYYYYYxx  e_mod:4  n_bit:10
	     *       1076 543210    e_align:1 n_shift:2
	     *
	     *   [case B]
	     *
	     *   01234567 01234567
	     *   xxxxxxxY Yxxxxxxx  e_mod:7  n_bit:2
	     *          1 0         e_align:1
	     *
	     *   01234567 01234567
	     *   xxxxxYYY YYYYxxxx  e_mod:5  n_bit:7
	     *        654 3210      e_align:6 n_shift:4
	     *
	     *   01234567 01234567
	     *   xxxxxYYY YYYYYxxx  e_mod:5  n_bit:8
	     *        765 43210     e_align:7 n_shift:5
	     */
	if ( e_mod + e_align < 8 ) {		/* case A : إեȤ */
	    int n_shift = 7 - (e_mod + e_align);	/* եȤbit */
	    unsigned char mask;
	    i = n_bit;					/* ̤Ͽbit */
	    mask = (unsigned char)0xff >> e_mod;
	    ret_val = (d_ptr[e_div] & mask) >> n_shift;	i -= 1 + e_align;
	    while ( 0 < i ) {
		ret_val <<= 8;
		ret_val |= d_ptr[e_div] << (8-n_shift);	i -= n_shift;
		e_div ++;
		ret_val |= d_ptr[e_div] >> n_shift;	i -= 8 - n_shift;
	    }
	}
	else {					/* case B : إեȤ */
	    int n_shift = e_mod + e_align - 7;		/* եȤbit */
	    unsigned char mask;
	    i = n_bit;
	    mask = (unsigned char)0xff >> e_mod;
	    ret_val= (d_ptr[e_div] & mask) << n_shift;	i -= 8 - e_mod;
	    e_div ++;
	    ret_val |= d_ptr[e_div] >> (8-n_shift);	i -= n_shift;
	    while ( 0 < i ) {
		ret_val <<= 8;
		ret_val |= d_ptr[e_div] << n_shift;	i -= 8 - n_shift;
		e_div ++;
		ret_val |= d_ptr[e_div] >> (8-n_shift);	i -= n_shift;
	    }
	}
	return ret_val;
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	return this->byte_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( v == 'T' ) return 1;
	else if ( v == 'F' ) return 0;
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	return this->short_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	return this->long_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->longlong_value(row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v0 = this->float_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_FLOAT_ROUND_LLONG <= v0 && v0 <= MAX_FLOAT_ROUND_LLONG )
	    return round_f2ll(v0);
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v0 = this->double_value(row_index,elem_index,repetition_idx);
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LLONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(v0);
	else return this->tnull_longlong_rec;
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring sval;
	double v0;
	this->get_string_value(row_index,elem_index,repetition_idx,sval);
	if ( sval.length() == 0 && this->tnull_is_set_rec == false ) {
	    return this->tnull_longlong_rec;
	}
	else if ( this->tnull_is_set_rec == true &&
		  this->tany.at_cs(TNULL_IDX).strcmp(sval.cstr()) == 0 ) {
	    return this->tnull_longlong_rec;
	}
	REMOVE_SPC(sval);
	v0 = sval.atof();
	if ( isfinite(v0) && MIN_DOUBLE_ROUND_LLONG <= v0 && v0 <= MAX_DOUBLE_ROUND_LLONG )
	    return round_d2ll(v0);
	else return this->tnull_longlong_rec;
    }
 invalid:
    return this->tnull_longlong_rec;
}


/* string */
/**
 * @brief  ͤʸͤǼ (٥)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
const char *fits_table_col::string_value( long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;

    if ( this->type_rec == FITS::STRING_T ) {
	const char *d_ptr = (const char *)this->data_ptr_rec;
	this->str_buf->assign(d_ptr +
		this->bytes_rec * this->elem_size_rec * row_index,
		this->bytes_rec);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( v == 'T' ) this->str_buf->assign("T");
	else if ( v == 'F' ) this->str_buf->assign("F");
	else this->str_buf->assign("");
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v = this->long_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    this->str_buf->assign("");
	else this->str_buf->printf("%ld",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v = this->longlong_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    this->str_buf->assign("");
	else this->str_buf->printf("%lld",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v = this->short_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    this->str_buf->assign("");
	else this->str_buf->printf("%hd",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v = this->double_value(row_index);
	if ( !isfinite(v) ) this->str_buf->assign("");
	else this->str_buf->printf("%.15G",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v = this->float_value(row_index);
	if ( !isfinite(v) ) this->str_buf->assign("");
	else this->str_buf->printf("%.7G",(double)v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v = this->byte_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    this->str_buf->assign("");
	else this->str_buf->printf("%hhu",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v = this->bit_value(row_index);
	this->str_buf->printf("%lld",v);
	return this->str_buf->cstr();
    }
 invalid:
    return "";
}

/**
 * @brief  ͤʸͤǼ (٥)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
const char *fits_table_col::string_value( long row_index,
					  const char *elem_name, 
					  long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( row_index < 0 || this->row_size_rec <= row_index ) return "";
	if ( elem_index < 0 || repetition_idx < 0 ) return "";
	v = this->bit_value(row_index,elem_index,repetition_idx,0);
	this->str_buf->printf("%lld",v);
	return this->str_buf->cstr();
    }
    else {
	return this->string_value( row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ͤʸͤǼ (٥)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  : ｪλ<br>
 *          NULL: ʾ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
const char *fits_table_col::string_value( long row_index,
					 long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;
    if ( elem_index < 0 || repetition_idx < 0 ) goto invalid;

    if ( this->type_rec == FITS::STRING_T ) {
	const char *d_ptr = (const char *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	this->str_buf->assign(d_ptr + 
	      this->bytes_rec * (this->elem_size_rec * row_index +e_idx ),
	      this->bytes_rec);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( v == 'T' ) this->str_buf->assign("T");
	else if ( v == 'F' ) this->str_buf->assign("F");
	else this->str_buf->assign("");
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v = this->long_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    this->str_buf->assign("");
	else this->str_buf->printf("%ld",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v=this->longlong_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    this->str_buf->assign("");
	else this->str_buf->printf("%lld",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v = this->short_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    this->str_buf->assign("");
	else this->str_buf->printf("%hd",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v = this->double_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(v) ) this->str_buf->assign("");
	else this->str_buf->printf("%.15G",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v = this->float_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(v) ) this->str_buf->assign("");
	else this->str_buf->printf("%.7G",(double)v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v =this->byte_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    this->str_buf->assign("");
	else this->str_buf->printf("%hhu",v);
	return this->str_buf->cstr();
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v = this->bit_value(row_index,elem_index,repetition_idx);
	this->str_buf->printf("%lld",v);
	return this->str_buf->cstr();
    }
 invalid:
    return "";
}

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   dest ͤγǼ (return)
 * @return  : ｪλ<br>
 *          NULL: ʾ硥<br>
 *          NULL: dest  NULL ξ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
const char *fits_table_col::get_string_value( long row_index,
					      tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;

    if ( this->type_rec == FITS::STRING_T ) {
	const char *d_ptr = (const char *)this->data_ptr_rec;
	dest->assign(d_ptr + 
		    this->bytes_rec * this->elem_size_rec * row_index,
		    this->bytes_rec);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index);
	if ( v == 'T' ) dest->assign("T");
	else if ( v == 'F' ) dest->assign("F");
	else dest->assign("");
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v = this->long_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    dest->assign("");
	else dest->printf("%ld",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v = this->longlong_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    dest->assign("");
	else dest->printf("%lld",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v = this->short_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    dest->assign("");
	else dest->printf("%hd",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v = this->double_value(row_index);
	if ( !isfinite(v) ) dest->assign("");
	else dest->printf("%.15G",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v = this->float_value(row_index);
	if ( !isfinite(v) ) dest->assign("");
	else dest->printf("%.7G",(double)v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v = this->byte_value(row_index);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    dest->assign("");
	else dest->printf("%hhu",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v = this->bit_value(row_index);
	dest->printf("%lld",v);
	return dest->cstr();
    }
 invalid:
    dest->assign("");
    return dest->cstr();
}

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   dest ͤγǼ (return)
 * @return  : ｪλ<br>
 *          NULL: ʾ硥<br>
 *          NULL: dest  NULL ξ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
const char *fits_table_col::get_string_value( long row_index,
					 long elem_index, tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    return this->get_string_value( row_index, elem_index, 0, dest );
}

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @param   dest ͤγǼ (return)
 * @return  : ｪλ<br>
 *          NULL: ʾ硥<br>
 *          NULL: dest  NULL ξ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
const char *fits_table_col::get_string_value( long row_index,
		    long elem_index, long repetition_idx, tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    if ( row_index < 0 || this->row_size_rec <= row_index ) goto invalid;
    if ( elem_index < 0 || repetition_idx < 0 ) goto invalid;

    if ( this->type_rec == FITS::STRING_T ) {
	const char *d_ptr = (const char *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	dest->assign(d_ptr + 
	     this->bytes_rec * (this->elem_size_rec * row_index + e_idx ),
	     this->bytes_rec);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v = this->logical_value(row_index,elem_index,repetition_idx);
	if ( v == 'T' ) dest->assign("T");
	else if ( v == 'F' ) dest->assign("F");
	else dest->assign("");
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v = this->long_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    dest->assign("");
	else dest->printf("%ld",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v=this->longlong_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    dest->assign("");
	else dest->printf("%lld",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v = this->short_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    dest->assign("");
	else dest->printf("%hd",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v = this->double_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(v) ) dest->assign("");
	else dest->printf("%.15G",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v = this->float_value(row_index,elem_index,repetition_idx);
	if ( !isfinite(v) ) dest->assign("");
	else dest->printf("%.7G",(double)v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v =this->byte_value(row_index,elem_index,repetition_idx);
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == v )
	    dest->assign("");
	else dest->printf("%hhu",v);
	return dest->cstr();
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v = this->bit_value(row_index,elem_index,repetition_idx);
	dest->printf("%lld",v);
	return dest->cstr();
    }
 invalid:
    dest->assign("");
    return dest->cstr();
}

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   dest ͤγǼ (return)
 * @return  : ｪλ<br>
 *          NULL: ʾ硥<br>
 *          NULL: dest  NULL ξ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
const char *fits_table_col::get_string_value( long row_index,
					      const char *elem_name, 
					      tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    return this->get_string_value( row_index, elem_name, 0, dest );
}

/**
 * @brief  ͤʸͤǼ (٥롦tstringǼ)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @param   dest ͤγǼ (return)
 * @return  : ｪλ<br>
 *          NULL: ʾ硥<br>
 *          NULL: dest  NULL ξ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
const char *fits_table_col::get_string_value( long row_index,
					      const char *elem_name, 
					      long repetition_idx,
					      tstring *dest ) const
{
    if ( dest == NULL ) return NULL;

    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( row_index < 0 || this->row_size_rec <= row_index ) {
	    dest->assign("");
	    return dest->cstr();
	}
	if ( elem_index < 0 || repetition_idx < 0 ) {
	    dest->assign("");
	    return dest->cstr();
	}
	v = this->bit_value(row_index,elem_index,repetition_idx,0);
	dest->printf("%lld",v);
	return dest->cstr();
    }
    else {
	return this->get_string_value( row_index, elem_index, repetition_idx,
				       dest );
    }
}

/**
 * @deprecated  侩<br>
 *              fits_table_col::get_string_value( long, tstring * ) <br>
 *              򤪻Ȥ
 */
const char *fits_table_col::get_string_value( long row_index, 
					      tstring &dest ) const
{
    return this->get_string_value(row_index, &dest);
}

/**
 * @deprecated  侩<br>
 *              fits_table_col::get_string_value( long, long, tstring * ) <br>
 *              򤪻Ȥ
 */
const char *fits_table_col::get_string_value( long row_index,
					 long elem_index, tstring &dest ) const
{
    return this->get_string_value(row_index, elem_index, 0, &dest);
}

/**
 * @deprecated  侩<br>
 *         fits_table_col::get_string_value( long, long, long, tstring * ) <br>
 *         򤪻Ȥ
 */
const char *fits_table_col::get_string_value( long row_index,
		    long elem_index, long repetition_idx, tstring &dest ) const
{
    return 
	this->get_string_value(row_index, elem_index, repetition_idx, &dest);
}

/**
 * @deprecated  侩<br>
 *       fits_table_col::get_string_value( long, const char *, tstring * ) <br>
 *       򤪻Ȥ
 */
const char *fits_table_col::get_string_value( long row_index,
					      const char *elem_name, 
					      tstring &dest ) const
{
    return this->get_string_value(row_index, elem_name, 0, &dest);
}

/**
 * @deprecated  侩<br>
 * fits_table_col::get_string_value( long, const char *, long, tstring * ) <br>
 * 򤪻Ȥ
 */
const char *fits_table_col::get_string_value( long row_index,
					      const char *elem_name, 
					      long repetition_idx,
					      tstring &dest ) const
{
    return this->get_string_value(row_index, elem_name, repetition_idx, &dest);
}

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ʳξ֤ NULLͤϡ"" Ǥ
 *
 * @param   row_index ԥǥå
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
ssize_t fits_table_col::get_string_value( long row_index, 
					char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_string_value( row_index, &sval );
    return sval.getstr(dest_buf,buf_size);
}

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
ssize_t fits_table_col::get_string_value( long row_index,
					const char *elem_name, 
					char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_string_value( row_index, elem_name, 0, &sval );
    return sval.getstr(dest_buf,buf_size);
}

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
ssize_t fits_table_col::get_string_value( long row_index,
				    const char *elem_name, long repetition_idx,
				    char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_string_value( row_index, elem_name, repetition_idx, &sval );
    return sval.getstr(dest_buf,buf_size);
}

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
ssize_t fits_table_col::get_string_value( long row_index, long elem_index, 
					char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_string_value( row_index, elem_index, 0, &sval );
    return sval.getstr(dest_buf,buf_size);
}

/**
 * @brief  ͤʸͤǼ (٥롦桼Хåե˼)
 *
 *  ͤʸ֤ͤޤ<br>
 *  η FITS::ASCII_T (TFORMn λ 'A' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  η FITS::ASCII_T ξ硤뤬ʸ٥NULL͡פʤ
 *  ΥдؿNULLͤȤ "" ֤ޤ(TNULLn ȤӤϹԤʤޤ)
 *  FITS::ASCII_T ʳξ硤뤬FITSǤNULLͤǤСΥдؿ
 *   NULLͤȤ "" ֤ޤ
 *
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @param   dest_buf ͤγǼ (return)
 * @param   buf_size dest_buf ΥХåե
 * @return  : ХåեĹʬʾ˥ԡǤʸ
 *                    ('\0'ϴޤޤʤ) <br>
 *          (顼): ǥԡʤä
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
ssize_t fits_table_col::get_string_value( long row_index,
					long elem_index, long repetition_idx,
					char *dest_buf, size_t buf_size ) const
{
    tstring sval;
    this->get_string_value( row_index, elem_index, repetition_idx, &sval );
    return sval.getstr(dest_buf,buf_size);
}


/*
 * high-level APIs to write a cell.
 *  - TSCAL and TZERO are applied.
 *  - NULL value is handled.
 */

/* double arg */

/**
 * @brief  ¿(double)ǥͤ (٥)
 *
 *  ¿ value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  value  NAN ͿϡNULLͤͿ줿ΤȤޤλХ
 *  ʥơ֥䥢ơ֥ξ TNULLn ͤ
 *  ͤ򥻥ޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( double value, long row_index )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( isfinite(value) ) {
	    if ( 0 < this->fmt_str.length() ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		    tstring buf(64);
		    buf.printf("%.15G",value);
		    this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
		double dvalue = value;
		dvalue -= this->tzero_rec;
		dvalue /= this->tscal_rec;
		if ( !isfinite(dvalue) ) goto set_null_str;
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dvalue);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    if ( MIN_DOUBLE_ROUND_LLONG <= dvalue && dvalue <= MAX_DOUBLE_ROUND_LLONG ) {
			this->tmp_str_buf->printf(this->fmt_str.cstr(),
						  round_d2ll(dvalue));
			return this->assign_string(this->tmp_str_buf->cstr(),
						   row_index);
		    }
		    else goto set_null_str;
		}
	    }
	}
	else {
	set_null_str:
	    if ( this->tnull_is_set_rec == true ) {
		return this->assign_string(this->tany.cstr(TNULL_IDX),
					   row_index);
	    }
	    else return this->assign_string(NULL, row_index);
	}
	return this->assign_double(value, row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_double(value, row_index);
    }
    else if ( !isfinite(value) &&
	      ( this->type_rec == FITS::LONG_T ||
		this->type_rec == FITS::SHORT_T ||
		this->type_rec == FITS::BYTE_T ||
		this->type_rec == FITS::LONGLONG_T ) ) {
	return this->assign_longlong(this->tnull_w_rec, row_index);
    }
    else {
	value -= this->tzero_rec;
	value /= this->tscal_rec;
	return this->assign_double(value, row_index);
    }
}

/**
 * @brief  ¿(double)ǥͤ (٥)
 *
 *  ¿ value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  value  NAN ͿϡNULLͤͿ줿ΤȤޤλХ
 *  ʥơ֥䥢ơ֥ξ TNULLn ͤ
 *  ͤ򥻥ޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( double value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( isfinite(value) ) {
	    if ( 0 < this->fmt_str.length() ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		    tstring buf(64);
		    buf.printf("%.15G",value);
		    this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		    return this->assign_string(this->tmp_str_buf->cstr(),
					  row_index, elem_name, repetition_idx);
		}
		double dvalue = value;
		dvalue -= this->tzero_rec;
		dvalue /= this->tscal_rec;
		if ( !isfinite(dvalue) ) goto set_null_str;
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dvalue);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				   row_index, elem_name, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    if ( MIN_DOUBLE_ROUND_LLONG <= dvalue && dvalue <= MAX_DOUBLE_ROUND_LLONG ) {
			this->tmp_str_buf->printf(this->fmt_str.cstr(),
						  round_d2ll(dvalue));
			return this->assign_string(this->tmp_str_buf->cstr(),
				       row_index, elem_name, repetition_idx);
		    }
		    else goto set_null_str;
		}
	    }
	}
	else {
	set_null_str:
	    if ( this->tnull_is_set_rec == true ) {
		return this->assign_string(this->tany.cstr(TNULL_IDX),
				row_index, elem_name, repetition_idx);
	    }
	    else return this->assign_string(NULL, row_index,
					    elem_name, repetition_idx);
	}
	return this->assign_double(value, row_index,
				   elem_name, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_double(value, row_index,
				   elem_name, repetition_idx);
    }
    else if ( !isfinite(value) &&
	      ( this->type_rec == FITS::LONG_T ||
		this->type_rec == FITS::SHORT_T ||
		this->type_rec == FITS::BYTE_T ||
		this->type_rec == FITS::LONGLONG_T ) ) {
	return this->assign_longlong(this->tnull_w_rec, row_index,
				     elem_name, repetition_idx);
    }
    else {
	value -= this->tzero_rec;
	value /= this->tscal_rec;
	return this->assign_double(value, row_index,
				   elem_name, repetition_idx);
    }
}

/**
 * @brief  ¿(double)ǥͤ (٥)
 *
 *  ¿ value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  value  NAN ͿϡNULLͤͿ줿ΤȤޤλХ
 *  ʥơ֥䥢ơ֥ξ TNULLn ͤ
 *  ͤ򥻥ޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( double value, long row_index,
					long elem_index, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( isfinite(value) ) {
	    if ( 0 < this->fmt_str.length() ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		    tstring buf(64);
		    buf.printf("%.15G",value);
		    this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		    return this->assign_string(this->tmp_str_buf->cstr(),
					 row_index, elem_index, repetition_idx);
		}
		double dvalue = value;
		dvalue -= this->tzero_rec;
		dvalue /= this->tscal_rec;
		if ( !isfinite(dvalue) ) goto set_null_str;
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dvalue);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				  row_index, elem_index, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    if ( MIN_DOUBLE_ROUND_LLONG <= dvalue && dvalue <= MAX_DOUBLE_ROUND_LLONG ) {
			this->tmp_str_buf->printf(this->fmt_str.cstr(),
						  round_d2ll(dvalue));
			return this->assign_string(this->tmp_str_buf->cstr(),
				      row_index, elem_index, repetition_idx);
		    }
		    else goto set_null_str;
		}
	    }
	}
	else {
	set_null_str:
	    if ( this->tnull_is_set_rec == true ) {
		return this->assign_string(this->tany.cstr(TNULL_IDX),
			       row_index, elem_index, repetition_idx);
	    }
	    else return this->assign_string(NULL, row_index,
					    elem_index, repetition_idx);
	}
	return this->assign_double(value, row_index,
				   elem_index, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_double(value, row_index,
				   elem_index, repetition_idx);
    }
    else if ( !isfinite(value) &&
	      ( this->type_rec == FITS::LONG_T ||
		this->type_rec == FITS::SHORT_T ||
		this->type_rec == FITS::BYTE_T ||
		this->type_rec == FITS::LONGLONG_T ) ) {
	return this->assign_longlong(this->tnull_w_rec, row_index,
				     elem_index, repetition_idx);
    }
    else {
	value -= this->tzero_rec;
	value /= this->tscal_rec;
	return this->assign_double(value, row_index,
				   elem_index, repetition_idx);
    }
}

/* float arg */

/**
 * @brief  ¿(float)ǥͤ (٥)
 *
 *  ¿ value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  value  NAN ͿϡNULLͤͿ줿ΤȤޤλХ
 *  ʥơ֥䥢ơ֥ξ TNULLn ͤ
 *  ͤ򥻥ޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( float value, long row_index )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( isfinite(value) ) {
	    if ( 0 < this->fmt_str.length() ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		    tstring buf(64);
		    buf.printf("%.7G",(double)value);
		    this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
		double dvalue = value;
		dvalue -= this->tzero_rec;
		dvalue /= this->tscal_rec;
		if ( !isfinite(dvalue) ) goto set_null_str;
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dvalue);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    if ( MIN_DOUBLE_ROUND_LLONG <= dvalue && dvalue <= MAX_DOUBLE_ROUND_LLONG ) {
			this->tmp_str_buf->printf(this->fmt_str.cstr(),
						  round_d2ll(dvalue));
			return this->assign_string(this->tmp_str_buf->cstr(),
						   row_index);
		    }
		    else goto set_null_str;
		}
	    }
	}
	else {
	set_null_str:
	    if ( this->tnull_is_set_rec == true ) {
		return this->assign_string(this->tany.cstr(TNULL_IDX),
					   row_index);
	    }
	    else return this->assign_string(NULL, row_index);
	}
	return this->assign_float(value, row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_double(value, row_index);
    }
    else if ( !isfinite(value) &&
	      ( this->type_rec == FITS::LONG_T ||
		this->type_rec == FITS::SHORT_T ||
		this->type_rec == FITS::BYTE_T ||
		this->type_rec == FITS::LONGLONG_T ) ) {
	return this->assign_longlong(this->tnull_w_rec, row_index);
    }
    else {
	double dvalue = value;
	dvalue -= this->tzero_rec;
	dvalue /= this->tscal_rec;
	return this->assign_double(dvalue, row_index);
    }
}

/**
 * @brief  ¿(float)ǥͤ (٥)
 *
 *  ¿ value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  value  NAN ͿϡNULLͤͿ줿ΤȤޤλХ
 *  ʥơ֥䥢ơ֥ξ TNULLn ͤ
 *  ͤ򥻥ޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( float value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( isfinite(value) ) {
	    if ( 0 < this->fmt_str.length() ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		    tstring buf(64);
		    buf.printf("%.7G",(double)value);
		    this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		    return this->assign_string(this->tmp_str_buf->cstr(),
				          row_index, elem_name, repetition_idx);
		}
		double dvalue = value;
		dvalue -= this->tzero_rec;
		dvalue /= this->tscal_rec;
		if ( !isfinite(dvalue) ) goto set_null_str;
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dvalue);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				   row_index, elem_name, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    if ( MIN_DOUBLE_ROUND_LLONG <= dvalue && dvalue <= MAX_DOUBLE_ROUND_LLONG ) {
			this->tmp_str_buf->printf(this->fmt_str.cstr(),
						  round_d2ll(dvalue));
			return this->assign_string(this->tmp_str_buf->cstr(),
				       row_index, elem_name, repetition_idx);
		    }
		    else goto set_null_str;
		}
	    }
	}
	else {
	set_null_str:
	    if ( this->tnull_is_set_rec == true ) {
		return this->assign_string(this->tany.cstr(TNULL_IDX),
				row_index, elem_name, repetition_idx);
	    }
	    else return this->assign_string(NULL, row_index,
					    elem_name, repetition_idx);
	}
	return this->assign_float(value, row_index,
				  elem_name, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_double(value, row_index,
				   elem_name, repetition_idx);
    }
    else if ( !isfinite(value) &&
	      ( this->type_rec == FITS::LONG_T ||
		this->type_rec == FITS::SHORT_T ||
		this->type_rec == FITS::BYTE_T ||
		this->type_rec == FITS::LONGLONG_T ) ) {
	return this->assign_longlong(this->tnull_w_rec, row_index,
				     elem_name, repetition_idx);
    }
    else {
	double dvalue = value;
	dvalue -= this->tzero_rec;
	dvalue /= this->tscal_rec;
	return this->assign_double(dvalue, row_index,
				   elem_name, repetition_idx);
    }
}

/**
 * @brief  ¿(double)ǥͤ (٥)
 *
 *  ¿ value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  value  NAN ͿϡNULLͤͿ줿ΤȤޤλХ
 *  ʥơ֥䥢ơ֥ξ TNULLn ͤ
 *  ͤ򥻥ޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( float value, long row_index,
					long elem_index, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( isfinite(value) ) {
	    if ( 0 < this->fmt_str.length() ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		    tstring buf(64);
		    buf.printf("%.7G",(double)value);
		    this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		    return this->assign_string(this->tmp_str_buf->cstr(),
				         row_index, elem_index, repetition_idx);
		}
		double dvalue = value;
		dvalue -= this->tzero_rec;
		dvalue /= this->tscal_rec;
		if ( !isfinite(dvalue) ) goto set_null_str;
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dvalue);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				  row_index, elem_index, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    if ( MIN_DOUBLE_ROUND_LLONG <= dvalue && dvalue <= MAX_DOUBLE_ROUND_LLONG ) {
			this->tmp_str_buf->printf(this->fmt_str.cstr(),
						  round_d2ll(dvalue));
			return this->assign_string(this->tmp_str_buf->cstr(),
				      row_index, elem_index, repetition_idx);
		    }
		    else goto set_null_str;
		}
	    }
	}
	else {
	set_null_str:
	    if ( this->tnull_is_set_rec == true ) {
		return this->assign_string(this->tany.cstr(TNULL_IDX),
			       row_index, elem_index, repetition_idx);
	    }
	    else return this->assign_string(NULL, row_index,
					    elem_index, repetition_idx);
	}
	return this->assign_float(value, row_index,
				  elem_index, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_double(value, row_index,
				   elem_index, repetition_idx);
    }
    else if ( !isfinite(value) &&
	      ( this->type_rec == FITS::LONG_T ||
		this->type_rec == FITS::SHORT_T ||
		this->type_rec == FITS::BYTE_T ||
		this->type_rec == FITS::LONGLONG_T ) ) {
	return this->assign_longlong(this->tnull_w_rec, row_index,
				     elem_index, repetition_idx);
    }
    else {
	double dvalue = value;
	dvalue -= this->tzero_rec;
	dvalue /= this->tscal_rec;
	return this->assign_double(dvalue, row_index,
				   elem_index, repetition_idx);
    }
}

/* long long arg */

/**
 * @brief  (long long)ǥͤ (٥)
 *
 *   value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  ΥдؿǤ NULLͤͿϤǤޤNULLͤͿϡ
 *  fits_table_col::assign(double, long) ѤƤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( long long value, long row_index )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		tstring buf(64);
		buf.printf("%lld",value);
		this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		return this->assign_string(this->tmp_str_buf->cstr(),
					   row_index);
	    }
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    if ( isfinite(dval) ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    if ( this->tscal_rec != 1.0 || this->tzero_rec != 0.0 ) {
			if ( MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
			    value = round_d2ll(dval);
			}
			else goto set_null_str;
		    }
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),value);
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
	    }
	    else {
	    set_null_str:
		if ( this->tnull_is_set_rec == true ) {
		    return this->assign_string(this->tany.cstr(TNULL_IDX),
					       row_index);
		}
		else return this->assign_string(NULL, row_index);
	    }
	}
	return this->assign_longlong(value, row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_longlong(value, row_index);
    }
    else {
	if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return this->assign_longlong(value, row_index);
	}
	else {
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    return this->assign_double(dval, row_index);
	}
    }
}

/**
 * @brief  (long long)ǥͤ (٥)
 *
 *   value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  ΥдؿǤ NULLͤͿϤǤޤNULLͤͿϡ
 *  fits_table_col::assign(double, long, const char *, long) ѤƤ
 *  
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( long long value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		tstring buf(64);
		buf.printf("%lld",value);
		this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		return this->assign_string(this->tmp_str_buf->cstr(),
				      row_index, elem_name, repetition_idx);
	    }
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    if ( isfinite(dval) ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				   row_index, elem_name, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    if ( this->tscal_rec != 1.0 || this->tzero_rec != 0.0 ) {
			if ( MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
			    value = round_d2ll(dval);
			}
			else goto set_null_str;
		    }
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),value);
		    return this->assign_string(this->tmp_str_buf->cstr(),
				   row_index, elem_name, repetition_idx);
		}
	    }
	    else {
	    set_null_str:
		if ( this->tnull_is_set_rec == true ) {
		    return this->assign_string(this->tany.cstr(TNULL_IDX),
				    row_index, elem_name, repetition_idx);
		}
		else return this->assign_string(NULL, row_index,
						elem_name, repetition_idx);
	    }
	}
	return this->assign_longlong(value, row_index,
				     elem_name, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_longlong(value, row_index,
				     elem_name, repetition_idx);
    }
    else {
	if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return this->assign_longlong(value, row_index,
					 elem_name, repetition_idx);
	}
	else {
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    return this->assign_double(dval, row_index,
				       elem_name, repetition_idx);
	}
    }
}

/**
 * @brief  (long long)ǥͤ (٥)
 *
 *   value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  ΥдؿǤ NULLͤͿϤǤޤNULLͤͿϡ
 *  fits_table_col::assign(double, long, long, long) ѤƤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( long long value, long row_index,
					long elem_index, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		tstring buf(64);
		buf.printf("%lld",value);
		this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		return this->assign_string(this->tmp_str_buf->cstr(),
				     row_index, elem_index, repetition_idx);
	    }
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    if ( isfinite(dval) ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				  row_index, elem_index, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    if ( this->tscal_rec != 1.0 || this->tzero_rec != 0.0 ) {
			if ( MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
			    value = round_d2ll(dval);
			}
			else goto set_null_str;
		    }
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),value);
		    return this->assign_string(this->tmp_str_buf->cstr(),
				  row_index, elem_index, repetition_idx);
		}
	    }
	    else {
	    set_null_str:
		if ( this->tnull_is_set_rec == true ) {
		    return this->assign_string(this->tany.cstr(TNULL_IDX),
				   row_index, elem_index, repetition_idx);
		}
		else return this->assign_string(NULL, row_index,
						elem_index, repetition_idx);
	    }
	}
	return this->assign_longlong(value, row_index,
				     elem_index, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_longlong(value, row_index,
				     elem_index, repetition_idx);
    }
    else {
	if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return this->assign_longlong(value, row_index,
					 elem_index, repetition_idx);
	}
	else {
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    return this->assign_double(dval, row_index,
				       elem_index, repetition_idx);
	}
    }
}

/* long arg */

/**
 * @brief  (long)ǥͤ (٥)
 *
 *   value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  ΥдؿǤ NULLͤͿϤǤޤNULLͤͿϡ
 *  fits_table_col::assign(double, long) ѤƤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( long value, long row_index )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		tstring buf(64);
		buf.printf("%ld",value);
		this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		return this->assign_string(this->tmp_str_buf->cstr(),
					   row_index);
	    }
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    if ( isfinite(dval) ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    long long llv = value;
		    if ( this->tscal_rec != 1.0 || this->tzero_rec != 0.0 ) {
			if ( MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
			    llv = round_d2ll(dval);
			}
			else goto set_null_str;
		    }
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),llv);
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
	    }
	    else {
	    set_null_str:
		if ( this->tnull_is_set_rec == true ) {
		    return this->assign_string(this->tany.cstr(TNULL_IDX),
					       row_index);
		}
		else return this->assign_string(NULL, row_index);
	    }
	}
	return this->assign_long(value, row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_long(value, row_index);
    }
    else {
	if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return this->assign_long(value, row_index);
	}
	else {
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    return this->assign_double(dval, row_index);
	}
    }
}

/**
 * @brief  (long)ǥͤ (٥)
 *
 *   value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  ΥдؿǤ NULLͤͿϤǤޤNULLͤͿϡ
 *  fits_table_col::assign(double, long, const char *, long) ѤƤ
 *  
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( long value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		tstring buf(64);
		buf.printf("%ld",value);
		this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		return this->assign_string(this->tmp_str_buf->cstr(),
				      row_index, elem_name, repetition_idx);
	    }
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    if ( isfinite(dval) ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				   row_index, elem_name, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    long long llv = value;
		    if ( this->tscal_rec != 1.0 || this->tzero_rec != 0.0 ) {
			if ( MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
			    llv = round_d2ll(dval);
			}
			else goto set_null_str;
		    }
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),llv);
		    return this->assign_string(this->tmp_str_buf->cstr(),
				   row_index, elem_name, repetition_idx);
		}
	    }
	    else {
	    set_null_str:
		if ( this->tnull_is_set_rec == true ) {
		    return this->assign_string(this->tany.cstr(TNULL_IDX),
				    row_index, elem_name, repetition_idx);
		}
		else return this->assign_string(NULL, row_index,
						elem_name, repetition_idx);
	    }
	}
	return this->assign_long(value, row_index,
				 elem_name, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_long(value, row_index,
				 elem_name, repetition_idx);
    }
    else {
	if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return this->assign_long(value, row_index,
				     elem_name, repetition_idx);
	}
	else {
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    return this->assign_double(dval, row_index,
				       elem_name, repetition_idx);
	}
    }
}

/**
 * @brief  (long)ǥͤ (٥)
 *
 *   value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  ΥдؿǤ NULLͤͿϤǤޤNULLͤͿϡ
 *  fits_table_col::assign(double, long, long, long) ѤƤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( long value, long row_index,
					long elem_index, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		tstring buf(64);
		buf.printf("%ld",value);
		this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		return this->assign_string(this->tmp_str_buf->cstr(),
				     row_index, elem_index, repetition_idx);
	    }
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    if ( isfinite(dval) ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				  row_index, elem_index, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    long long llv = value;
		    if ( this->tscal_rec != 1.0 || this->tzero_rec != 0.0 ) {
			if ( MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
			    llv = round_d2ll(dval);
			}
			else goto set_null_str;
		    }
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),llv);
		    return this->assign_string(this->tmp_str_buf->cstr(),
				  row_index, elem_index, repetition_idx);
		}
	    }
	    else {
	    set_null_str:
		if ( this->tnull_is_set_rec == true ) {
		    return this->assign_string(this->tany.cstr(TNULL_IDX),
				   row_index, elem_index, repetition_idx);
		}
		else return this->assign_string(NULL, row_index,
						elem_index, repetition_idx);
	    }
	}
	return this->assign_long(value, row_index,
				 elem_index, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_long(value, row_index,
				 elem_index, repetition_idx);
    }
    else {
	if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return this->assign_long(value, row_index,
				     elem_index, repetition_idx);
	}
	else {
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    return this->assign_double(dval, row_index,
				       elem_index, repetition_idx);
	}
    }
}

/* int arg */

/**
 * @brief  (int)ǥͤ (٥)
 *
 *   value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  ΥдؿǤ NULLͤͿϤǤޤNULLͤͿϡ
 *  fits_table_col::assign(double, long) ѤƤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( int value, long row_index )
{
    return this->assign((long)value, row_index);
}

/**
 * @brief  (int)ǥͤ (٥)
 *
 *   value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  ΥдؿǤ NULLͤͿϤǤޤNULLͤͿϡ
 *  fits_table_col::assign(double, long, const char *, long) ѤƤ
 *  
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( int value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    return this->assign((long)value, row_index, 
			elem_name, repetition_idx);
}

/**
 * @brief  (int)ǥͤ (٥)
 *
 *   value 顤إå TZEROn  TSCALn ȿǤ¿ͤᡤŬڤ
 *  Ѵ(Υξϴݤ)Ԥʤäƥޤ<br>
 *  ΥдؿǤ NULLͤͿϤǤޤNULLͤͿϡ
 *  fits_table_col::assign(double, long, long, long) ѤƤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign( int value, long row_index,
					long elem_index, long repetition_idx )
{
    return this->assign((long)value, row_index, 
			elem_index, repetition_idx);
}

#if 0
/* short arg */

fits_table_col &fits_table_col::assign( short value, long row_index )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		tstring buf(64);
		buf.printf("%hd",value);
		this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		return this->assign_string(this->tmp_str_buf->cstr(),
					   row_index);
	    }
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    if ( isfinite(dval) ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    long long llv = value;
		    if ( this->tscal_rec != 1.0 || this->tzero_rec != 0.0 ) {
			if ( MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
			    llv = round_d2ll(dval);
			}
			else goto set_null_str;
		    }
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),llv);
		    return this->assign_string(this->tmp_str_buf->cstr(),
					       row_index);
		}
	    }
	    else {
	    set_null_str:
		if ( this->tnull_is_set_rec == true ) {
		    return this->assign_string(this->tany.cstr(TNULL_IDX),
					       row_index);
		}
		else return this->assign_string(NULL, row_index);
	    }
	}
	return this->assign_short(value, row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_short(value, row_index);
    }
    else {
	if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return this->assign_short(value, row_index);
	}
	else {
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    return this->assign_double(dval, row_index);
	}
    }
}

fits_table_col &fits_table_col::assign( short value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		tstring buf(64);
		buf.printf("%hd",value);
		this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		return this->assign_string(this->tmp_str_buf->cstr(),
				      row_index, elem_name, repetition_idx);
	    }
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    if ( isfinite(dval) ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				   row_index, elem_name, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    long long llv = value;
		    if ( this->tscal_rec != 1.0 || this->tzero_rec != 0.0 ) {
			if ( MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
			    llv = round_d2ll(dval);
			}
			else goto set_null_str;
		    }
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),llv);
		    return this->assign_string(this->tmp_str_buf->cstr(),
				   row_index, elem_name, repetition_idx);
		}
	    }
	    else {
	    set_null_str:
		if ( this->tnull_is_set_rec == true ) {
		    return this->assign_string(this->tany.cstr(TNULL_IDX),
				    row_index, elem_name, repetition_idx);
		}
		else return this->assign_string(NULL, row_index,
						elem_name, repetition_idx);
	    }
	}
	return this->assign_short(value, row_index,
				  elem_name, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_short(value, row_index,
				  elem_name, repetition_idx);
    }
    else {
	if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return this->assign_short(value, row_index,
				      elem_name, repetition_idx);
	}
	else {
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    return this->assign_double(dval, row_index,
				       elem_name, repetition_idx);
	}
    }
}

fits_table_col &fits_table_col::assign( short value, long row_index,
					long elem_index, long repetition_idx )
{
    if ( this->type_rec == FITS::STRING_T ) {
	if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		tstring buf(64);
		buf.printf("%hd",value);
		this->tmp_str_buf->printf(this->fmt_str.cstr(), buf.cstr());
		return this->assign_string(this->tmp_str_buf->cstr(),
				     row_index, elem_index, repetition_idx);
	    }
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    if ( isfinite(dval) ) {
		if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = this->tmp_str_buf->strchr('E');
			if ( 0 <= ep ) this->tmp_str_buf->at(ep) = 'D';
		    }
		    return this->assign_string(this->tmp_str_buf->cstr(),
				  row_index, elem_index, repetition_idx);
		}
		else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		    long long llv = value;
		    if ( this->tscal_rec != 1.0 || this->tzero_rec != 0.0 ) {
			if ( MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
			    llv = round_d2ll(dval);
			}
			else goto set_null_str;
		    }
		    this->tmp_str_buf->printf(this->fmt_str.cstr(),llv);
		    return this->assign_string(this->tmp_str_buf->cstr(),
				  row_index, elem_index, repetition_idx);
		}
	    }
	    else {
	    set_null_str:
		if ( this->tnull_is_set_rec == true ) {
		    return this->assign_string(this->tany.cstr(TNULL_IDX),
				   row_index, elem_index, repetition_idx);
		}
		else return this->assign_string(NULL, row_index,
						elem_index, repetition_idx);
	    }
	}
	return this->assign_short(value, row_index,
				  elem_index, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ||
	      this->type_rec == FITS::BIT_T ) {
	return this->assign_short(value, row_index,
				  elem_index, repetition_idx);
    }
    else {
	if ( this->tscal_rec == 1.0 && this->tzero_rec == 0.0 ) {
	    return this->assign_short(value, row_index,
				      elem_index, repetition_idx);
	}
	else {
	    double dval = value;
	    dval -= this->tzero_rec;
	    dval /= this->tscal_rec;
	    return this->assign_double(dval, row_index,
				       elem_index, repetition_idx);
	}
    }
}

#endif	/* 0 */

/* const char * arg */

/**
 * @brief  ʸͤǥͤ (٥)
 *
 *  ʸͤŬڤѴޤ<br>
 *  value  NULLʸξϡNULLͤͿ줿ΤȤޤλХ
 *  ʥơ֥䥢ơ֥ξ TNULLn ͤ
 *  ͤ򥻥ޤNULLʸνͤ "NULL" ǡͤ
 *  fits_table_col::assign_null_svalue() ѹǤޤ<br>
 *  ¿Υξ硤ʸ value ¿ͤѴ˥إå
 *   TZEROn  TSCALn ͤѴ¿ͤ򥻥ޤ(ब
 *  ξϡǤᤤޤ)
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
fits_table_col &fits_table_col::assign( const char *value, long row_index )
{
    tstring sval;
    if ( value == NULL ) sval.assign(this->null_svalue());
    else {
	if ( sval.assign(value).strtrim(' ').strcmp(this->null_svalue()) != 0 )
	    sval.assign(value);
    }
    if ( this->type_rec == FITS::STRING_T ) {
	if ( sval.strcmp(this->null_svalue()) == 0 ) {
	    if ( this->tnull_is_set_rec == true ) {
		sval.assign(this->tany.cstr(TNULL_IDX));
	    }
	    else {
		sval = NULL;
	    }
	}
	else if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		sval.assignf(this->fmt_str.cstr(),sval.cstr());
	    }
	    else if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		REMOVE_SPC(sval);
		double dval = sval.atof();
		dval -= this->tzero_rec;
		dval /= this->tscal_rec;
		if ( isfinite(dval) ) {
		    sval.assignf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = sval.strchr('E');
			if ( 0 <= ep ) sval.at(ep) = 'D';
		    }
		}
		else {
		    if ( this->tnull_is_set_rec == true ) {
			sval.assign(this->tany.cstr(TNULL_IDX));
		    }
		    else {
			sval = NULL;
		    }
		}
	    }
	    else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		REMOVE_SPC(sval);
		double dval = sval.atof();
		dval -= this->tzero_rec;
		dval /= this->tscal_rec;
		if ( isfinite(dval) && MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
		    sval.assignf(this->fmt_str.cstr(),round_d2ll(dval));
		}
		else {
		    if ( this->tnull_is_set_rec == true ) {
			sval.assign(this->tany.cstr(TNULL_IDX));
		    }
		    else {
			sval = NULL;
		    }
		}
	    }
	}
	return this->assign_string(sval.cstr(), row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	const char *v;
	v = sval.strtrim(' ').cstr();
	if ( sval.strcmp(this->null_svalue()) == 0 ) v = NULL;
	return this->assign_string(v, row_index);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	REMOVE_SPC(sval);
	double dval = sval.atof();
	long long v;
	if ( isfinite(dval) && MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG )
	    v = round_d2ll(dval);
	else v = this->tnull_w_rec;
	return this->assign_bit(v, row_index);
    }
    else if ( (sval.strcmp(this->null_svalue()) == 0)
	      && (this->type_rec == FITS::LONG_T ||
		  this->type_rec == FITS::SHORT_T ||
		  this->type_rec == FITS::BYTE_T ||
		  this->type_rec == FITS::LONGLONG_T) ) {
	return this->assign_longlong(this->tnull_w_rec, row_index);
    }
    else if ( sval.strcmp(this->null_svalue()) == 0 ) {
	return this->assign_double(NAN, row_index);
    }
    else {
	REMOVE_SPC(sval);
	double v = sval.atof();
	v -= this->tzero_rec;
	v /= this->tscal_rec;
	return this->assign_double(v, row_index);
    }
}

/**
 * @brief  ʸͤǥͤ (٥)
 *
 *  ʸͤŬڤѴޤ<br>
 *  value  NULLʸξϡNULLͤͿ줿ΤȤޤλХ
 *  ʥơ֥䥢ơ֥ξ TNULLn ͤ
 *  ͤ򥻥ޤNULLʸνͤ "NULL" ǡͤ
 *  fits_table_col::assign_null_svalue() ѹǤޤ<br>
 *  ¿Υξ硤ʸ value ¿ͤѴ˥إå
 *   TZEROn  TSCALn ͤѴ¿ͤ򥻥ޤ(ब
 *  ξϡǤᤤޤ)
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
fits_table_col &fits_table_col::assign( const char *value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    tstring sval;
    if ( value == NULL ) sval.assign(this->null_svalue());
    else {
	if ( sval.assign(value).strtrim(' ').strcmp(this->null_svalue()) != 0 )
	    sval.assign(value);
    }
    if ( this->type_rec == FITS::STRING_T ) {
	if ( sval.strcmp(this->null_svalue()) == 0 ) {
	    if ( this->tnull_is_set_rec == true ) {
		sval.assign(this->tany.cstr(TNULL_IDX));
	    }
	    else {
		sval = NULL;
	    }
	}
	else if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		sval.assignf(this->fmt_str.cstr(),sval.cstr());
	    }
	    else if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		REMOVE_SPC(sval);
		double dval = sval.atof();
		dval -= this->tzero_rec;
		dval /= this->tscal_rec;
		if ( isfinite(dval) ) {
		    sval.assignf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = sval.strchr('E');
			if ( 0 <= ep ) sval.at(ep) = 'D';
		    }
		}
		else {
		    if ( this->tnull_is_set_rec == true ) {
			sval.assign(this->tany.cstr(TNULL_IDX));
		    }
		    else {
			sval = NULL;
		    }
		}
	    }
	    else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		REMOVE_SPC(sval);
		double dval = sval.atof();
		dval -= this->tzero_rec;
		dval /= this->tscal_rec;
		if ( isfinite(dval) && MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
		    sval.assignf(this->fmt_str.cstr(),round_d2ll(dval));
		}
		else {
		    if ( this->tnull_is_set_rec == true ) {
			sval.assign(this->tany.cstr(TNULL_IDX));
		    }
		    else {
			sval = NULL;
		    }
		}
	    }
	}
	return this->assign_string(sval.cstr(), row_index,
				   elem_name, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	const char *v;
	v = sval.strtrim(' ').cstr();
	if ( sval.strcmp(this->null_svalue()) == 0 ) v = NULL;
	return this->assign_string(v, row_index, elem_name, repetition_idx);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	REMOVE_SPC(sval);
	double dval = sval.atof();
	long long v;
	if ( isfinite(dval) && MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG )
	    v = round_d2ll(dval);
	else v = this->tnull_w_rec;
	return this->assign_bit(v, row_index,
				elem_name, repetition_idx);
    }
    else if ( (sval.strcmp(this->null_svalue()) == 0)
	      && (this->type_rec == FITS::LONG_T ||
		  this->type_rec == FITS::SHORT_T ||
		  this->type_rec == FITS::BYTE_T ||
		  this->type_rec == FITS::LONGLONG_T) ) {
	return this->assign_longlong(this->tnull_w_rec, row_index,
				     elem_name, repetition_idx);
    }
    else if ( sval.strcmp(this->null_svalue()) == 0 ) {
	return this->assign_double(NAN, row_index, elem_name, repetition_idx);
    }
    else {
	REMOVE_SPC(sval);
	double v = sval.atof();
	v -= this->tzero_rec;
	v /= this->tscal_rec;
	return this->assign_double(v, row_index, elem_name, repetition_idx);
    }
}

/**
 * @brief  ʸͤǥͤ (٥)
 *
 *  ʸͤŬڤѴޤ<br>
 *  value  NULLʸξϡNULLͤͿ줿ΤȤޤλХ
 *  ʥơ֥䥢ơ֥ξ TNULLn ͤ
 *  ͤ򥻥ޤNULLʸνͤ "NULL" ǡͤ
 *  fits_table_col::assign_null_svalue() ѹǤޤ<br>
 *  ¿Υξ硤ʸ value ¿ͤѴ˥إå
 *   TZEROn  TSCALn ͤѴ¿ͤ򥻥ޤ(ब
 *  ξϡǤᤤޤ)
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @note  Хʥơ֥ʸ󷿤ΥǤ TNULLn ꤬ǽǤ
 *         SFITSIO ȼĥǤ
 * @attention  ĹϥݡȤƤޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
fits_table_col &fits_table_col::assign( const char *value, long row_index,
					long elem_index, long repetition_idx )
{
    tstring sval;
    if ( value == NULL ) sval.assign(this->null_svalue());
    else {
	if ( sval.assign(value).strtrim(' ').strcmp(this->null_svalue()) != 0 )
	    sval.assign(value);
    }
    if ( this->type_rec == FITS::STRING_T ) {
	if ( sval.strcmp(this->null_svalue()) == 0 ) {
	    if ( this->tnull_is_set_rec == true ) {
		sval.assign(this->tany.cstr(TNULL_IDX));
	    }
	    else {
		sval = NULL;
	    }
	}
	else if ( 0 < this->fmt_str.length() ) {
	    if ( this->tany.at(TDISP_IDX).strpbrk("A") == 0 ) {
		sval.assignf(this->fmt_str.cstr(),sval.cstr());
	    }
	    else if ( this->tany.at(TDISP_IDX).strpbrk("FEGD") == 0 ) {
		REMOVE_SPC(sval);
		double dval = sval.atof();
		dval -= this->tzero_rec;
		dval /= this->tscal_rec;
		if ( isfinite(dval) ) {
		    sval.assignf(this->fmt_str.cstr(),dval);
		    if ( this->tany.cchr(TDISP_IDX,0) == 'D' ) {
			ssize_t ep = sval.strchr('E');
			if ( 0 <= ep ) sval.at(ep) = 'D';
		    }
		}
		else {
		    if ( this->tnull_is_set_rec == true ) {
			sval.assign(this->tany.cstr(TNULL_IDX));
		    }
		    else {
			sval = NULL;
		    }
		}
	    }
	    else if ( this->tany.at(TDISP_IDX).strpbrk("I") == 0 ) {
		REMOVE_SPC(sval);
		double dval = sval.atof();
		dval -= this->tzero_rec;
		dval /= this->tscal_rec;
		if ( isfinite(dval) && MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG ) {
		    sval.assignf(this->fmt_str.cstr(),round_d2ll(dval));
		}
		else {
		    if ( this->tnull_is_set_rec == true ) {
			sval.assign(this->tany.cstr(TNULL_IDX));
		    }
		    else {
			sval = NULL;
		    }
		}
	    }
	}
	return this->assign_string(sval.cstr(), row_index,
				   elem_index, repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	const char *v;
	v = sval.strtrim(' ').cstr();
	if ( sval.strcmp(this->null_svalue()) == 0 ) v = NULL;
	return this->assign_string(v, row_index, elem_index, repetition_idx);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	REMOVE_SPC(sval);
	double dval = sval.atof();
	long long v;
	if ( isfinite(dval) && MIN_DOUBLE_ROUND_LLONG <= dval && dval <= MAX_DOUBLE_ROUND_LLONG )
	    v = round_d2ll(dval);
	else v = this->tnull_w_rec;
	return this->assign_bit(v, row_index,
				elem_index, repetition_idx);
    }
    else if ( (sval.strcmp(this->null_svalue()) == 0)
	      && (this->type_rec == FITS::LONG_T ||
		  this->type_rec == FITS::SHORT_T ||
		  this->type_rec == FITS::BYTE_T ||
		  this->type_rec == FITS::LONGLONG_T) ) {
	return this->assign_longlong(this->tnull_w_rec, row_index,
				     elem_index, repetition_idx);
    }
    else if ( sval.strcmp(this->null_svalue()) == 0 ) {
	return this->assign_double(NAN, row_index, elem_index, repetition_idx);
    }
    else {
	REMOVE_SPC(sval);
	double v = sval.atof();
	v -= this->tzero_rec;
	v /= this->tscal_rec;
	return this->assign_double(v, row_index, elem_index, repetition_idx);
    }
}

/* low-level member functions */

/* logical */
/**
 * @brief  ͤǥͤ򤽤Τޤ (٥)
 *
 *  ͤǥͤ򤽤Τޤޤ<br>
 *  η FITS::LOGICAL_T (TFORMn λ 'L' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  value ˤϡͤ򼨤 'T'ͤ򼨤 'F'NULL 򼨤¾
 *  (㤨 0) Ϳޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_logical( int value, long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }

    if ( this->type_rec == FITS::LOGICAL_T ) {
	fits::logical_t v;
	fits::logical_t *d_ptr = (fits::logical_t *)this->data_ptr_rec;
	if ( value == 'T' ) v = 'T';
	else if ( value == 'F' ) v = 'F';
	else v = '\0';
	d_ptr[this->elem_size_rec * row_index] = v;
	return *this;
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = this->tnull_w_rec;
	return this->assign_long(v, row_index);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = this->tnull_w_rec;
	return this->assign_short(v, row_index);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = this->tnull_w_rec;
	return this->assign_longlong(v, row_index);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = this->tnull_w_rec;
	return this->assign_byte(v, row_index);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = NAN;
	return this->assign_double(v, row_index);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = NAN;
	return this->assign_float(v, row_index);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit((value == 'T') ? 1 : 0, row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	const char *v;
	if ( value == 'T' ) v = "T";
	else if ( value == 'F' ) v = "F";
	else v = NULL;
	return this->assign_string(v,row_index);
    }
    else return *this;	/* invalid */
}

/**
 * @brief  ͤǥͤ򤽤Τޤ (٥)
 *
 *  ͤǥͤ򤽤Τޤޤ<br>
 *  η FITS::LOGICAL_T (TFORMn λ 'L' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  value ˤϡͤ򼨤 'T'ͤ򼨤 'F'NULL 򼨤¾
 *  (㤨 0) Ϳޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_logical( int value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit((value == 'T') ? 1 : 0,
				row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->assign_logical( value, 
				     row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ͤǥͤ򤽤Τޤ (٥)
 *
 *  ͤǥͤ򤽤Τޤޤ<br>
 *  η FITS::LOGICAL_T (TFORMn λ 'L' ޤ) ξ˺®
 *  Ǥޤ<br>
 *  value ˤϡͤ򼨤 'T'ͤ򼨤 'F'NULL 򼨤¾
 *  (㤨 0) Ϳޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_logical( int value, long row_index,
					 long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    if ( elem_index < 0 || repetition_idx < 0 ) return *this;	/*invalid */

    if ( this->type_rec == FITS::LOGICAL_T ) {
	fits::logical_t v;
	fits::logical_t *d_ptr = (fits::logical_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	if ( value == 'T' ) v = 'T';
	else if ( value == 'F' ) v = 'F';
	else v = '\0';
	d_ptr[this->elem_size_rec * row_index + e_idx] = v;
	return *this;
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = this->tnull_w_rec;
	return this->assign_long(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = this->tnull_w_rec;
	return this->assign_longlong(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = NAN;
	return this->assign_double(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v;
	if ( value == 'T' ) v = 1;
	else if ( value == 'F' ) v = 0;
	else v = NAN;
	return this->assign_float(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit((value == 'T') ? 1 : 0,
				row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	const char *v;
	if ( value == 'T' ) v = "T";
	else if ( value == 'F' ) v = "F";
	else v = NULL;
	return this->assign_string(v,row_index,elem_index,repetition_idx);
    }
    else return *this;	/* invalid */
}


/* short */
/**
 * @brief  (short)ǥͤ򤽤Τޤ (٥)
 *
 *  (short)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::SHORT_T (TFORMn λ 'I' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_short( short value, long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }

    if ( this->type_rec == FITS::SHORT_T ) {
	fits::short_t *d_ptr = (fits::short_t *)this->data_ptr_rec;
	d_ptr[this->elem_size_rec * row_index] = value;
	return *this;
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	return this->assign_long(value,row_index);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->assign_longlong(value,row_index);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_double(v,row_index);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_float(v,row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F', row_index);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%hd",value);
	return this->assign_string(buf.cstr(), row_index);
    }
    else return *this;	/* invalid */
}

/**
 * @brief  (short)ǥͤ򤽤Τޤ (٥)
 *
 *  (short)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::SHORT_T (TFORMn λ 'I' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_short( short value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->assign_short( value,
				   row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  (short)ǥͤ򤽤Τޤ (٥)
 *
 *  (short)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::SHORT_T (TFORMn λ 'I' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_short( short value, long row_index,
					 long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    if ( elem_index < 0 || repetition_idx < 0 ) return *this;	/* invalid */

    if ( this->type_rec == FITS::SHORT_T ) {
	fits::short_t *d_ptr = (fits::short_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	d_ptr[this->elem_size_rec * row_index + e_idx] = value;
	return *this;
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	return this->assign_long(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->assign_longlong(value,
				     row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_double(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_float(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F',
				    row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%hd",value);
	return this->assign_string(buf.cstr(), row_index, elem_index,
				   repetition_idx);
    }
    else return *this;	/* invalid */
}


/* long */
/**
 * @brief  (long)ǥͤ򤽤Τޤ (٥)
 *
 *  (long)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::LONG_T (TFORMn λ 'J' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 * @note  long 32-bit OS Ǥ 32-bit64-bit OS Ǥ 64-bit Ļ
 *        ۤȤɤǤ
 */
fits_table_col &fits_table_col::assign_long( long value, long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }

    if ( this->type_rec == FITS::LONG_T ) {
	fits::long_t *d_ptr = (fits::long_t *)this->data_ptr_rec;
	d_ptr += this->elem_size_rec * row_index;
	if ( MIN_INT32 <= value && value <= MAX_INT32 )
	    *d_ptr = value;
	else
	    *d_ptr = this->tnull_w_rec;
	return *this;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->assign_longlong(value,row_index);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( MIN_SHORT <= value && value <= MAX_SHORT ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_double(v,row_index);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_float(v,row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F',
				    row_index);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%ld",value);
	return this->assign_string(buf.cstr(), row_index);
    }
    else return *this;	/* invalid */
}

/**
 * @brief  (long)ǥͤ򤽤Τޤ (٥)
 *
 *  (long)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::LONG_T (TFORMn λ 'J' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 * @note  long 32-bit OS Ǥ 32-bit64-bit OS Ǥ 64-bit Ļ
 *        ۤȤɤǤ
 */
fits_table_col &fits_table_col::assign_long( long value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->assign_long( value,
				  row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  (long)ǥͤ򤽤Τޤ (٥)
 *
 *  (long)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::LONG_T (TFORMn λ 'J' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 * @note  long 32-bit OS Ǥ 32-bit64-bit OS Ǥ 64-bit Ļ
 *        ۤȤɤǤ
 */
fits_table_col &fits_table_col::assign_long( long value, long row_index,
					 long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    if ( elem_index < 0 || repetition_idx < 0 ) return *this;	/* invalid */

    if ( this->type_rec == FITS::LONG_T ) {
	fits::long_t *d_ptr = (fits::long_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	d_ptr += this->elem_size_rec * row_index + e_idx;
	if ( MIN_INT32 <= value && value <= MAX_INT32 )
	    *d_ptr = value;
	else
	    *d_ptr = this->tnull_w_rec;
	return *this;
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->assign_longlong(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( MIN_SHORT <= value && value <= MAX_SHORT ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_double(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_float(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F',
				    row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%ld",value);
	return this->assign_string(buf.cstr(), row_index, elem_index,
				   repetition_idx);
    }
    else return *this;	/* invalid */
}


/* long long */
/**
 * @brief  (long long)ǥͤ򤽤Τޤ (٥)
 *
 *  (long long)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::LONGLONG_T (TFORMn λ 'K' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_longlong( long long value, 
						 long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }

    if ( this->type_rec == FITS::LONGLONG_T ) {
	fits::longlong_t *d_ptr = (fits::longlong_t *)this->data_ptr_rec;
	d_ptr += this->elem_size_rec * row_index;
	if ( MIN_INT64 <= value && value <= MAX_INT64 )
	    *d_ptr = value;
	else
	    *d_ptr = this->tnull_w_rec;
	return *this;
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( MIN_LONG <= value && value <= MAX_LONG ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_long(v,row_index);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( MIN_SHORT <= value && value <= MAX_SHORT ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_double(v,row_index);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_float(v,row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F',
				    row_index);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%lld",value);
	return this->assign_string(buf.cstr(), row_index);
    }
    else return *this;	/* invalid */
}

/**
 * @brief  (long long)ǥͤ򤽤Τޤ (٥)
 *
 *  (long long)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::LONGLONG_T (TFORMn λ 'K' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_longlong( long long value, 
				   long row_index,
				   const char *elem_name, long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->assign_longlong( value, 
				      row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  (long long)ǥͤ򤽤Τޤ (٥)
 *
 *  (long long)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::LONGLONG_T (TFORMn λ 'K' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_longlong( long long value, 
				         long row_index,
					 long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    if ( elem_index < 0 || repetition_idx < 0 ) return *this;	/* invalid */

    if ( this->type_rec == FITS::LONGLONG_T ) {
	fits::longlong_t *d_ptr = (fits::longlong_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	d_ptr += this->elem_size_rec * row_index + e_idx;
	if ( MIN_INT64 <= value && value <= MAX_INT64 )
	    *d_ptr = value;
	else
	    *d_ptr = this->tnull_w_rec;
	return *this;
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( MIN_LONG <= value && value <= MAX_LONG ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_long(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( MIN_SHORT <= value && value <= MAX_SHORT ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_double(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_float(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F',
				    row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%lld",value);
	return this->assign_string(buf.cstr(), row_index, elem_index,
				   repetition_idx);
    }
    else return *this;	/* invalid */
}


/* byte */
/**
 * @brief  (byte)ǥͤ򤽤Τޤ (٥)
 *
 *  (unsigned char)ǥͤ򤽤ΤޤޤTZEROn  TSCALn 
 *  ͤˤѴϹԤʤޤ<br>
 *  η FITS::BYTE_T (TFORMn λ 'B' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_byte( unsigned char value, 
					     long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }

    if ( this->type_rec == FITS::BYTE_T ) {
	fits::byte_t *d_ptr = (fits::byte_t *)this->data_ptr_rec;
	d_ptr += this->elem_size_rec * row_index;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR )
	    *d_ptr = value;
	else
	    *d_ptr = this->tnull_w_rec;
	return *this;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	return this->assign_short(value,row_index);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	return this->assign_long(value,row_index);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->assign_longlong(value,row_index);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_double(v,row_index);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	float v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_float(v,row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F',
				    row_index);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%hhu",value);
	return this->assign_string(buf.cstr(), row_index);
    }
    else return *this;	/* invalid */
}

/**
 * @brief  (byte)ǥͤ򤽤Τޤ (٥)
 *
 *  (unsigned char)ǥͤ򤽤ΤޤޤTZEROn  TSCALn 
 *  ͤˤѴϹԤʤޤ<br>
 *  η FITS::BYTE_T (TFORMn λ 'B' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_byte( unsigned char value, 
				   long row_index,
				   const char *elem_name, long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->assign_byte( value, 
				  row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  (byte)ǥͤ򤽤Τޤ (٥)
 *
 *  (unsigned char)ǥͤ򤽤ΤޤޤTZEROn  TSCALn 
 *  ͤˤѴϹԤʤޤ<br>
 *  η FITS::BYTE_T (TFORMn λ 'B' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_byte( unsigned char value, 
					 long row_index,
					 long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    if ( elem_index < 0 || repetition_idx < 0 ) return *this;	/* invalid */

    if ( this->type_rec == FITS::BYTE_T ) {
	fits::byte_t *d_ptr = (fits::byte_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	d_ptr += this->elem_size_rec * row_index + e_idx;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR )
	    *d_ptr = value;
	else
	    *d_ptr = this->tnull_w_rec;
	return *this;
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	return this->assign_short(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	return this->assign_long(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->assign_longlong(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	double v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_double(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	double v;
	if ( this->tnull_is_set_rec == true && this->tnull_r_rec == value ) 
	    v = NAN;
	else v = value;
	return this->assign_float(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F',
				    row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	return this->assign_bit(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%hhu",value);
	return this->assign_string(buf.cstr(), row_index, elem_index,
				   repetition_idx);
    }
    else return *this;	/* invalid */
}


/* float */
/**
 * @brief  (float)ǥͤ򤽤Τޤ (٥)
 *
 *  ¿(float)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::FLOAT_T (TFORMn λ 'E' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_float( float value, long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }

    if ( this->type_rec == FITS::FLOAT_T ) {
	fits::float_t *d_ptr = (fits::float_t *)this->data_ptr_rec;
	d_ptr[this->elem_size_rec * row_index] = value;
	return *this;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	return this->assign_double((double)value,row_index);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_LONG <= value && value <= MAX_FLOAT_ROUND_LONG )
	    v = round_f2l(value);
	else v = this->tnull_w_rec;
	return this->assign_long(v,row_index);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_LLONG <= value && value <= MAX_FLOAT_ROUND_LLONG )
	    v = round_f2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_longlong(v,row_index);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_SHORT <= value && value <= MAX_FLOAT_ROUND_SHORT )
	    v = round_f2l(value);
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_UCHAR <= value && value <= MAX_FLOAT_ROUND_UCHAR )
	    v = round_f2l(value);
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v;
	if ( !isfinite(value) ) v = '\0';
	else v = (round(value) != 0.0) ? 'T' : 'F';
	return this->assign_logical(v, row_index);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_LLONG <= value && value <= MAX_FLOAT_ROUND_LLONG )
	    v = round_f2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_bit(v,row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	if ( !isfinite(value) ) {
	    return this->assign_string(NULL, row_index);
	}
	buf.printf("%.7G",(double)value);
	return this->assign_string(buf.cstr(), row_index);
    }
    else return *this;	/* invalid */
}

/**
 * @brief  (float)ǥͤ򤽤Τޤ (٥)
 *
 *  ¿(float)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::FLOAT_T (TFORMn λ 'E' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_float( float value, long row_index, 
					      const char *elem_name, 
					      long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_LLONG <= value && value <= MAX_FLOAT_ROUND_LLONG )
	    v = round_f2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_bit(v,row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->assign_float( value, 
				   row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  (float)ǥͤ򤽤Τޤ (٥)
 *
 *  ¿(float)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::FLOAT_T (TFORMn λ 'E' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_float( float value, long row_index,
					 long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    if ( elem_index < 0 || repetition_idx < 0 ) return *this;	/* invalid */

    if ( this->type_rec == FITS::FLOAT_T ) {
	fits::float_t *d_ptr = (fits::float_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	d_ptr[this->elem_size_rec * row_index + e_idx] = value;
	return *this;
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	return this->assign_double((double)value,
				   row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_LONG <= value && value <= MAX_FLOAT_ROUND_LONG )
	    v = round_f2l(value);
	else v = this->tnull_w_rec;
	return this->assign_long(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_LLONG <= value && value <= MAX_FLOAT_ROUND_LLONG )
	    v = round_f2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_longlong(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_SHORT <= value && value <= MAX_FLOAT_ROUND_SHORT )
	    v = round_f2l(value);
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_UCHAR <= value && value <= MAX_FLOAT_ROUND_UCHAR )
	    v = round_f2l(value);
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v;
	if ( !isfinite(value) ) v = '\0';
	else v = (round(value) != 0.0) ? 'T' : 'F';
	return this->assign_logical(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( isfinite(value) && MIN_FLOAT_ROUND_LLONG <= value && value <= MAX_FLOAT_ROUND_LLONG )
	    v = round_f2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_bit(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	if ( !isfinite(value) ) {
	    return this->assign_string(NULL, row_index,
				       elem_index, repetition_idx);
	}
	buf.printf("%.7G",(double)value);
	return this->assign_string(buf.cstr(), row_index, elem_index,
				   repetition_idx);
    }
    else return *this;	/* invalid */
}


/* double */
/**
 * @brief  (double)ǥͤ򤽤Τޤ (٥)
 *
 *  ¿(double)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::DOUBLE_T (TFORMn λ 'D' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_double( double value, long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }

    if ( this->type_rec == FITS::DOUBLE_T ) {
	fits::double_t *d_ptr = (fits::double_t *)this->data_ptr_rec;
	d_ptr[this->elem_size_rec * row_index] = value;
	return *this;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	return this->assign_float((float)value,row_index);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_LONG <= value && value <= MAX_DOUBLE_ROUND_LONG )
	    v = round_d2l(value);
	else v = this->tnull_w_rec;
	return this->assign_long(v,row_index);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_LLONG <= value && value <= MAX_DOUBLE_ROUND_LLONG )
	    v = round_d2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_longlong(v,row_index);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_SHORT <= value && value <= MAX_DOUBLE_ROUND_SHORT )
	    v = round_d2l(value);
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_UCHAR <= value && value <= MAX_DOUBLE_ROUND_UCHAR )
	    v = round_d2l(value);
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v;
	if ( !isfinite(value) ) v = '\0';
	else v = (round(value) != 0.0) ? 'T' : 'F';
	return this->assign_logical(v,row_index);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_LLONG <= value && value <= MAX_DOUBLE_ROUND_LLONG )
	    v = round_d2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_bit(v,row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	if ( !isfinite(value) ) {
	    return this->assign_string(NULL, row_index);
	}
	buf.printf("%.15G",value);
	return this->assign_string(buf.cstr(), row_index);
    }
    else return *this;	/* invalid */
}

/**
 * @brief  (double)ǥͤ򤽤Τޤ (٥)
 *
 *  ¿(double)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::DOUBLE_T (TFORMn λ 'D' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_double( double value, long row_index,
				   const char *elem_name, long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_LLONG <= value && value <= MAX_DOUBLE_ROUND_LLONG )
	    v = round_d2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_bit(v,row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->assign_double( value, 
				    row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  (double)ǥͤ򤽤Τޤ (٥)
 *
 *  ¿(double)ǥͤ򤽤ΤޤޤTZEROn  TSCALn ͤˤ
 *  ѴϹԤʤޤ<br>
 *  η FITS::DOUBLE_T (TFORMn λ 'D' ޤ) ξ˺®
 *  Ǥޤ
 *
 * @param   value 
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_double( double value, long row_index,
					 long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    if ( elem_index < 0 || repetition_idx < 0 ) return *this;	/* invalid */

    if ( this->type_rec == FITS::DOUBLE_T ) {
	fits::double_t *d_ptr = (fits::double_t *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	d_ptr[this->elem_size_rec * row_index + e_idx] = value;
	return *this;
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	return this->assign_float((float)value,
				  row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_LONG <= value && value <= MAX_DOUBLE_ROUND_LONG )
	    v = round_d2l(value);
	else v = this->tnull_w_rec;
	return this->assign_long(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	long long v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_LLONG <= value && value <= MAX_DOUBLE_ROUND_LLONG )
	    v = round_d2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_longlong(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_SHORT <= value && value <= MAX_DOUBLE_ROUND_SHORT )
	    v = round_d2l(value);
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_UCHAR <= value && value <= MAX_DOUBLE_ROUND_UCHAR )
	    v = round_d2l(value);
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	int v;
	if ( !isfinite(value) ) v = '\0';
	else v = (round(value) != 0.0) ? 'T' : 'F';
	return this->assign_logical(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BIT_T ) {
	long long v;
	if ( isfinite(value) && MIN_DOUBLE_ROUND_LLONG <= value && value <= MAX_DOUBLE_ROUND_LLONG )
	    v = round_d2ll(value);
	else v = this->tnull_w_rec;
	return this->assign_bit(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	if ( !isfinite(value) ) {
	    return this->assign_string(NULL, row_index,
				       elem_index, repetition_idx);
	}
	buf.printf("%.15G",value);
	return this->assign_string(buf.cstr(), row_index, elem_index,
				   repetition_idx);
    }
    else return *this;	/* invalid */
}


/* string */
/**
 * @brief  ʸͤǥͤ򤽤Τޤ (٥)
 *
 *  ʸͤǥͤ򤽤Τޤޤ<br>
 *  Хʥơ֥Υη FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *  ξ硤ޤϥơ֥ξ˺®ǥǤޤ<br>
 *  η FITS::ASCII_T ξ硤value  NULL Ϳȡʸ
 *  ٥NULL͡פ򥻥åȤޤ(TNULLn ȤӤϹԤʤޤ)FITS::ASCII_T
 *  ʳξϡvalue  NULL Ϳ FITS ˽ NULLͤ
 *  åȤޤ
 *
 * @param   value ͡
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 * @note  ܺ٤ư: ʸ󷿤Υξ硤value  NULL ξϥ뤬 '\0'
 *        ޤvalue  "" ξϥ뤬ʸޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
fits_table_col &fits_table_col::assign_string( const char *value, 
					       long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    //if ( value == NULL ) return *this;

    if ( this->type_rec == FITS::STRING_T ) {
	long i;
	char *d_ptr = (char *)this->data_ptr_rec;
	d_ptr += this->bytes_rec * this->elem_size_rec * row_index;
	if ( value == NULL ) {
	    for ( i=0 ; i < this->bytes_rec ; i++ ) d_ptr[i] = '\0';
	}
	else {
	    //for ( ; *value == ' ' ; value++ );
	    for ( i=0 ; i < this->bytes_rec && value[i] != '\0' ; i++ )
		d_ptr[i] = value[i];
	    for ( ; i < this->bytes_rec ; i++ )
		d_ptr[i] = ' ';
	}
	return *this;
    }
    else {
	tstring sval;
	sval.assign(value);
	REMOVE_SPC(sval);
	if ( this->type_rec == FITS::LOGICAL_T ) {
	    int v;
	    double dval;
	    size_t epos;
	    dval = sval.strtod(&epos);
	    if ( epos == 0 ) {
		if ( sval.cchr(0) == 'T' || sval.cchr(0) == 't' ) v = 'T';
		else if ( sval.cchr(0) == 'F' || sval.cchr(0) == 'f' ) v = 'F';
		else v = '\0';
	    }
	    else {
		if ( !isfinite(dval) ) v = '\0';
		else v = (round(dval) != 0.0) ? 'T' : 'F';
	    }
	    return this->assign_logical(v,row_index);
	}
	else if ( this->type_rec == FITS::BIT_T ) {
	    double dv = sval.atof();
	    long long v;
	    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG )
		v = round_d2ll(dv);
	    else v = this->tnull_w_rec;
	    return this->assign_bit(v,row_index);
	}
	else if ( value == NULL ) {
	    return this->assign_double(NAN,row_index);
	}
	else {
	    return this->assign_double(sval.atof(),row_index);
	}
    }
}

/**
 * @brief  ʸͤǥͤ򤽤Τޤ (٥)
 *
 *  ʸͤǥͤ򤽤Τޤޤ<br>
 *  Хʥơ֥Υη FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *  ξ硤ޤϥơ֥ξ˺®ǥǤޤ<br>
 *  η FITS::ASCII_T ξ硤value  NULL Ϳȡʸ
 *  ٥NULL͡פ򥻥åȤޤ(TNULLn ȤӤϹԤʤޤ)FITS::ASCII_T
 *  ʳξϡvalue  NULL Ϳ FITS ˽ NULLͤ
 *  åȤޤ
 *
 * @param   value ͡
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 * @note  ܺ٤ư: ʸ󷿤Υξ硤value  NULL ξϥ뤬 '\0'
 *        ޤvalue  "" ξϥ뤬ʸޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
fits_table_col &fits_table_col::assign_string( const char *value, 
				   long row_index,
				   const char *elem_name, long repetition_idx )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    if ( this->type_rec == FITS::BIT_T ) {
	tstring sval;
	sval.assign(value);
	REMOVE_SPC(sval);
	double dv = sval.atof();
	long long v;
	if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG )
	    v = round_d2ll(dv);
	else v = this->tnull_w_rec;
	return this->assign_bit(v,row_index,elem_index,repetition_idx,0);
    }
    else {
	return this->assign_string( value, 
				    row_index, elem_index, repetition_idx );
    }
}

/**
 * @brief  ʸͤǥͤ򤽤Τޤ (٥)
 *
 *  ʸͤǥͤ򤽤Τޤޤ<br>
 *  Хʥơ֥Υη FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *  ξ硤ޤϥơ֥ξ˺®ǥǤޤ<br>
 *  η FITS::ASCII_T ξ硤value  NULL Ϳȡʸ
 *  ٥NULL͡פ򥻥åȤޤ(TNULLn ȤӤϹԤʤޤ)FITS::ASCII_T
 *  ʳξϡvalue  NULL Ϳ FITS ˽ NULLͤ
 *  åȤޤ
 *
 * @param   value ͡
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 * @note  ܺ٤ư: ʸ󷿤Υξ硤value  NULL ξϥ뤬 '\0'
 *        ޤvalue  "" ξϥ뤬ʸޤ
 * @note  SFITSIO Ǥϡ෿ FITS::ASCII_T (TFORMn λ 'A' ޤ) 
 *        ξ硤TNULLn ꤬ʤСƬʸ '\0' Υ NULLͤȤ
 *        ޤʸ٥NULL͡פȸƤǤޤ
 */
fits_table_col &fits_table_col::assign_string( const char *value, 
					 long row_index,
					 long elem_index, long repetition_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    if ( elem_index < 0 || repetition_idx < 0 ) return *this;	/* invalid */
    //if ( value == NULL ) return *this;

    if ( this->type_rec == FITS::STRING_T ) {
	long i;
	char *d_ptr = (char *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx
	    + elem_index;
	if ( this->elem_size_rec <= e_idx ) 
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	d_ptr += 
	    this->bytes_rec * (this->elem_size_rec * row_index + e_idx);
	if ( value == NULL ) {
	    for ( i=0 ; i < this->bytes_rec ; i++ ) d_ptr[i] = '\0';
	}
	else {
	    //for ( ; *value == ' ' ; value++ );
	    for ( i=0 ; i < this->bytes_rec && value[i] != '\0' ; i++ )
		d_ptr[i] = value[i];
	    for ( ; i < this->bytes_rec ; i++ )
		d_ptr[i] = ' ';
	}
	return *this;
    }
    else {
	tstring sval;
	sval.assign(value);
	REMOVE_SPC(sval);
	if ( this->type_rec == FITS::LOGICAL_T ) {
	    int v;
	    double dval;
	    size_t epos;
	    dval = sval.strtod(&epos);
	    if ( epos == 0 ) {
		if ( sval.cchr(0) == 'T' || sval.cchr(0) == 't' ) v = 'T';
		else if ( sval.cchr(0) == 'F' || sval.cchr(0) == 'f' ) v = 'F';
		else v = '\0';
	    }
	    else {
		if ( !isfinite(dval) ) v = '\0';
		else v = (round(dval) != 0.0) ? 'T' : 'F';
	    }
	    return this->assign_logical(v,row_index,elem_index,repetition_idx);
	}
	else if ( this->type_rec == FITS::BIT_T ) {
	    double dv = sval.atof();
	    long long v;
	    if ( isfinite(dv) && MIN_DOUBLE_ROUND_LLONG <= dv && dv <= MAX_DOUBLE_ROUND_LLONG )
		v = round_d2ll(dv);
	    else v = this->tnull_w_rec;
	    return this->assign_bit(v,row_index,elem_index,repetition_idx);
	}
	else if ( value == NULL ) {
	    return this->assign_double(NAN,row_index,elem_index,repetition_idx);
	}
	else {
	    return this->assign_double(sval.atof(),
				       row_index,elem_index,repetition_idx);
	}
    }
}


/* bit */
/**
 * @brief  (bit)ǥͤ򤽤Τޤ (٥)
 *
 *  (bit)ǥͤ򤽤Τޤޤ<br>
 *  η FITS::BIT_T (TFORMn λ 'X' ޤ) ξ˺®
 *  ǤΥХǤκüΥӥåȤͤ򥻥åȤޤ
 *
 * @param   value ͡
 * @param   row_index ԥǥå
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 */
fits_table_col &fits_table_col::assign_bit( long long value, long row_index )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }

    if ( this->type_rec == FITS::BIT_T ) {
	/* MSB˽񤯻 */
	unsigned char *d_ptr = (unsigned char *)this->data_ptr_rec;
	if ( value != 0 ) 
	    d_ptr[this->full_bytes_rec * row_index] |= 0x080;
	else 
	    d_ptr[this->full_bytes_rec * row_index] &= ~(0x080);
	return *this;
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F',
				    row_index);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( MIN_SHORT <= value && value <= MAX_SHORT ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( MIN_LONG <= value && value <= MAX_LONG ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_long(v,row_index);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->assign_longlong(value,row_index);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	return this->assign_float(value,row_index);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	return this->assign_double(value,row_index);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%lld",value);
	return this->assign_string(buf.cstr(), row_index);
    }
    else return *this;	/* invalid */
}

/**
 * @brief  (bit)ǥͤ򤽤Τޤ (٥)
 *
 *  (bit)ǥͤ򤽤Τޤޤ<br>
 *  η FITS::BIT_T (TFORMn λ 'X' ޤ) ξ˻Ѥ
 *  дؿǡǤդΥӥåǤ鱦ǤդΥӥåȿͤ򥻥å
 *  Ǥޤ
 *
 * @param   value ͡
 * @param   row_index ԥǥå
 * @param   elem_name ̾ (TELEMn ¸ߤ̾)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @param   nbit ӥå (ά TELEMn ͤΥӥåȥեɻŬ)
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 * @note  ӥåȱ黻Ԥʤᡤ¹ԥȤ礭⤷ޤ
 */
fits_table_col &fits_table_col::assign_bit( long long value, long row_index,
			 const char *elem_name, long repetition_idx, int nbit )
{
    long elem_index = this->elem_index_rec.index(elem_name,0);
    return this->assign_bit( value, 
			     row_index, elem_index, repetition_idx, nbit );
}

/**
 * @brief  (bit)ǥͤ򤽤Τޤ (٥)
 *
 *  (bit)ǥͤ򤽤Τޤޤ<br>
 *  η FITS::BIT_T (TFORMn λ 'X' ޤ) ξ˻Ѥ
 *  дؿǡǤդΥӥåǤ鱦ǤդΥӥåȿͤ򥻥å
 *  Ǥޤ
 *
 * @param   value ͡
 * @param   row_index ԥǥå
 * @param   elem_index ǥǥå (TDIMn 1ܤΥǥå)
 * @param   repetition_idx  TDIMn 2ܤΥǥå
 * @param   nbit ӥå (ά 1 Ŭ)
 * @return  Ȥλ
 * @attention  ĹϥݡȤƤޤ
 * @note  ӥåȱ黻Ԥʤᡤ¹ԥȤ礭⤷ޤ
 */
fits_table_col &fits_table_col::assign_bit( long long value, long row_index,
			       long elem_index, long repetition_idx, int nbit )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ) {
	return *this;	/* invalid */
    }
    if ( elem_index < 0 || repetition_idx < 0 ) return *this;	/* invalid */

    if ( this->type_rec == FITS::BIT_T ) {
	unsigned char *d_ptr = (unsigned char *)this->data_ptr_rec;
	long e_idx = this->dcol_size_rec * repetition_idx 
	    + elem_index;
	int n_bit;
	long e_mod, e_div, e_align;
	int i,j;
	if ( nbit < 1 ) 
	    n_bit = this->bit_size_telem.i(elem_index % this->dcol_size_rec);
	else
	    n_bit = nbit;
	if ( this->elem_size_rec <= e_idx ) {
	    e_idx = this->elem_size_rec - this->dcol_size_rec 
		+ elem_index % this->dcol_size_rec;
	}
	d_ptr += this->full_bytes_rec * row_index;
	e_align = (n_bit-1) % 8;
	e_mod = e_idx % 8;
	e_div = e_idx / 8;
	/* */
	j = (n_bit - 1) / 8;
	//sli__eprintf("debug: e_align=%ld e_mod=%ld e_div=%ld j=%d n_bit=%d\n",
	//		       e_align,e_mod,e_div,j,n_bit);
	/* */
	if ( e_mod + e_align < 8 ) {		/* case A : إեȤ */
	    int n_shift = 7 - (e_mod + e_align);	/* եȤbit */
	    unsigned char mask;
	    int sft;
	    i = n_bit;					/* ̤bit */
	    mask = (unsigned char)0xff >> (e_mod + n_shift);
	    mask <<= n_shift;
	    d_ptr[e_div] &= ~mask;	/* clear */
	    sft = 8 * j - n_shift;
	    //sli__eprintf("debug0: n_shift=%d mask=[%hhx] sft=%d\n",n_shift,mask,sft);
	    //sli__eprintf("debug0a: i=%d\n",i);
	    if ( 0 <= sft )
		d_ptr[e_div] |= ((value >> sft) & mask);
	    else 
		d_ptr[e_div] |= ((value << -sft) & mask);
	    i -= 1 + e_align;
	    //sli__eprintf("debug0b: i=%d\n",i);
	    while ( 0 < i && 0 < j ) {
		j--;
		mask = (unsigned char)0xff >> (8 - n_shift);
		d_ptr[e_div] &= ~mask;	/* clear */
		//sli__eprintf("debug0L1: mask=[%hhx] sft=%d\n",mask,sft);
		if ( 0 <= sft )
		    d_ptr[e_div] |= ((value >> sft) & mask);
		else
		    d_ptr[e_div] |= ((value << -sft) & mask);
		i -= n_shift;
		e_div ++;
		mask = (unsigned char)0xff << n_shift;
		d_ptr[e_div] &= ~mask;	/* clear */
		sft = 8 * j - n_shift;
		//sli__eprintf("debug0L2:  mask=[%hhx] sft=%d\n",mask,sft);
		if ( 0 <= sft )
		    d_ptr[e_div] |= ((value >> sft) & mask);
		else
		    d_ptr[e_div] |= ((value << -sft) & mask);
		i -= 8 - n_shift;
	    }
	}
	else {					/* case B : إեȤ */
	    int n_shift = e_mod + e_align - 7;		/* եȤbit */
	    unsigned char mask;
	    int sft;
	    i = n_bit;
	    mask = (unsigned char)0xff >> e_mod;
	    //sli__eprintf("debug1: mask = [%hhx]\n",mask);
	    d_ptr[e_div] &= ~mask;	/* clear */
	    sft = 8 * j + n_shift;
	    if ( 0 <= sft ) 
		d_ptr[e_div] |= ((value >> sft) & mask);
	    else 
		d_ptr[e_div] |= ((value << -sft) & mask);
	    i -= 8 - e_mod;
	    e_div ++;
	    mask = (unsigned char)0xff << (8 - n_shift);
	    d_ptr[e_div] &= ~mask;	/* clear */
	    sft = 8 * j - (8 - n_shift);
	    if ( 0 <= sft )
		d_ptr[e_div] |= ((value >> sft) & mask);
	    else
		d_ptr[e_div] |= ((value << -sft) & mask);
	    i -= n_shift;
	    while ( 0 < i && 0 < j ) {
		j--;
		mask = (unsigned char)0xff >> n_shift;
		d_ptr[e_div] &= ~mask;	/* clear */
		sft = 8 * j + n_shift;
		if ( 0 <= sft )
		    d_ptr[e_div] |= ((value >> sft) & mask);
		else
		    d_ptr[e_div] |= ((value << -sft) & mask);
		i -= 8 - n_shift;
		e_div ++;
		mask = (unsigned char)0xff << (8 - n_shift);
		d_ptr[e_div] &= ~mask;	/* clear */
		sft = 8 * j - (8 - n_shift);
		if ( 0 <= sft )
		    d_ptr[e_div] |= ((value >> sft) & mask);
		else
		    d_ptr[e_div] |= ((value << -sft) & mask);
		i -= n_shift;
	    }
	}
	return *this;
    }
    else if ( this->type_rec == FITS::LOGICAL_T ) {
	return this->assign_logical((value != 0) ? 'T' : 'F',
				    row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::BYTE_T ) {
	unsigned char v;
	if ( MIN_UCHAR <= value && value <= MAX_UCHAR ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_byte(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::SHORT_T ) {
	short v;
	if ( MIN_SHORT <= value && value <= MAX_SHORT ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_short(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONG_T ) {
	long v;
	if ( MIN_LONG <= value && value <= MAX_LONG ) v = value;
	else v = this->tnull_w_rec;
	return this->assign_long(v,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::LONGLONG_T ) {
	return this->assign_longlong(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::FLOAT_T ) {
	return this->assign_float(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::DOUBLE_T ) {
	return this->assign_double(value,row_index,elem_index,repetition_idx);
    }
    else if ( this->type_rec == FITS::STRING_T ) {
	tstring buf(64);
	buf.printf("%lld",value);
	return this->assign_string(buf.cstr(), row_index, elem_index,
				   repetition_idx);
    }
    else return *this;	/* invalid */
}

/**
 * @brief  Ĺ󵭽һҤ򥻥å (٥)
 *
 *  elem_idx ϡĹ󵭽һҤʣ¸ߤ˻ꤷޤ
 *  㤨 TFORM  '4PE(999)' ξ elem_idx  0  3 ͤȤ
 *  Ǥޤ
 *
 * @param   length 󵭽һ: ǿ
 * @param   offset 󵭽һ: Хȥեå
 * @param   row_index ԥǥå
 * @param   elem_idx 󵭽һҤֹ (ά0)
 * @return  Ȥλ
 */
fits_table_col &fits_table_col::assign_arrdesc( long length, long offset,
						long row_index, long elem_idx )
{
    if ( row_index < 0 || this->row_size_rec <= row_index ||
	 elem_idx < 0 || this->elem_size_rec <= elem_idx ) {
	return *this;	/* invalid */
    }

    if ( this->type_rec == FITS::LONGARRDESC_T ) {
	fits::longarrdesc_t *d_ptr = (fits::longarrdesc_t *)this->data_ptr_rec;
	d_ptr[this->elem_size_rec * row_index + elem_idx].length = length;
	d_ptr[this->elem_size_rec * row_index + elem_idx].offset = offset;
	if ( this->vl_max_length_rec < length ) {
	    this->vl_max_length_rec = length;
	}
    }
    else if ( this->type_rec == FITS::LLONGARRDESC_T ) {
	fits::llongarrdesc_t *d_ptr = (fits::llongarrdesc_t *)this->data_ptr_rec;
	d_ptr[this->elem_size_rec * row_index + elem_idx].length = length;
	d_ptr[this->elem_size_rec * row_index + elem_idx].offset = offset;
	if ( this->vl_max_length_rec < length ) {
	    this->vl_max_length_rec = length;
	}
    }

    return *this;
}


/**
 * @brief  rowѥˡѹ ()
 *
 *  rowѥݤˡɤΤ褦ˡǹԤʤꤷޤ
 *   3 Ĥ򤷤ޤ<br>
 *    "min", "auto"  ... Ǿ¤ݡꥵɬrealloc()ƤФ<br>
 *    "pow"  ... 2nǳ<br>
 *    NULL¾ ... ߤˡݻ
 *
 * @param   strategy ˡ
 * @return  Ȥλ
 *
 */
fits_table_col &fits_table_col::set_alloc_strategy( const char *strategy )
{
    this->data_rec.set_alloc_strategy(strategy);
    return *this;
}


/* */


/**
 * @brief  shallow copy Ĥ˻ (̤)
 * @note   ֥Ȥ return ľǻѤ롥
 */
/* : Ȥ֤Ȥreturn ʸǤȥ󥹥 */
/*       2󥳥ԡȤʻˤʤΤ void ˤʤäƤ */
void fits_table_col::set_scopy_flag()
{
    this->shallow_copy_ok = true;
    return;
}

/* protected */

/**
 * @brief  μΤΥдؿ
 *
 * @note   Υдؿ protected Ǥ
 */
fits_table_col &fits_table_col::set_protection( bool protect )
{
    this->protected_rec = protect;
    return *this;
}

/**
 * @brief  Ȥ fits_table ֥ȤϿ
 *
 * @note   Υдؿ protected Ǥ
 */
fits_table_col &fits_table_col::register_manager( fits_table *ptr )
{
    this->manager = ptr;
    return *this;
}

/* private */

/**
 * @brief   tnull ͤ򥻥å or ꥻå (NULLǥꥻå)
 *
 * @note   Υдؿ private Ǥ
 */
void fits_table_col::set_tnull( const long long *new_tnull_ptr )
{
    int the_type = this->type_rec;

    if ( new_tnull_ptr == NULL ) {
      this->tnull_r_rec = INDEF_LLONG;
      this->tnull_w_rec = get_indef_for_type(the_type);
      this->tnull_longlong_rec = INDEF_LLONG;
      this->tnull_long_rec = INDEF_LONG;
      this->tnull_short_rec = INDEF_SHORT;
      this->tnull_byte_rec = INDEF_UCHAR;
    }
    else {
      this->tnull_r_rec = *new_tnull_ptr;
      this->tnull_w_rec = get_tnull_for_type(the_type, *new_tnull_ptr);
      if ( MIN_LLONG <= *new_tnull_ptr && *new_tnull_ptr <= MAX_LLONG )
	  this->tnull_longlong_rec = *new_tnull_ptr;
      else this->tnull_longlong_rec = INDEF_LLONG;
      if ( MIN_LONG <= *new_tnull_ptr && *new_tnull_ptr <= MAX_LONG ) 
	  this->tnull_long_rec = *new_tnull_ptr;
      else this->tnull_long_rec = INDEF_LONG;
      if ( MIN_SHORT <= *new_tnull_ptr && *new_tnull_ptr <= MAX_SHORT ) 
	  this->tnull_short_rec = *new_tnull_ptr;
      else this->tnull_short_rec = INDEF_SHORT;
      if ( MIN_UCHAR <= *new_tnull_ptr && *new_tnull_ptr <= MAX_UCHAR ) 
	  this->tnull_byte_rec = *new_tnull_ptr;
      else this->tnull_byte_rec = INDEF_UCHAR;
    }

    return;
}

/**
 * @brief  this->def_all_recthis->def_rec 򹹿
 *
 * @note   Υдؿ private Ǥ
 */
void fits_table_col::update_def_rec_ptr()
{
    /* all version */
    this->def_all_rec.ttype = this->tany.cstr(TTYPE_IDX);
    this->def_all_rec.ttype_comment = this->tany.cstr(TTYPE_COMMENT_IDX);
    this->def_all_rec.tunit = this->tany.cstr(TUNIT_IDX);
    this->def_all_rec.tunit_comment = this->tany.cstr(TUNIT_COMMENT_IDX);
    this->def_all_rec.tdisp = this->tany.cstr(TDISP_IDX);
    this->def_all_rec.tdisp_comment = this->tany.cstr(TDISP_COMMENT_IDX);
    this->def_all_rec.tform = this->tany.cstr(TFORM_IDX);
    this->def_all_rec.tform_comment = this->tany.cstr(TFORM_COMMENT_IDX);
    this->def_all_rec.tdim  = this->tany.cstr(TDIM_IDX);
    this->def_all_rec.tdim_comment  = this->tany.cstr(TDIM_COMMENT_IDX);
    this->def_all_rec.tnull = this->tany.cstr(TNULL_IDX);
    this->def_all_rec.tnull_comment = this->tany.cstr(TNULL_COMMENT_IDX);
    this->def_all_rec.tzero = this->tany.cstr(TZERO_IDX);
    this->def_all_rec.tzero_comment = this->tany.cstr(TZERO_COMMENT_IDX);
    this->def_all_rec.tscal = this->tany.cstr(TSCAL_IDX);
    this->def_all_rec.tscal_comment = this->tany.cstr(TSCAL_COMMENT_IDX);

    //this->def_all_rec.talas = this->talas.cstrarray();
    this->def_all_rec.talas = this->tany.cstr(TALAS_IDX);
    this->def_all_rec.talas_comment = this->tany.cstr(TALAS_COMMENT_IDX);
    //this->def_all_rec.telem = this->telem_def.cstrarray();
    this->def_all_rec.telem = this->tany.cstr(TELEM_IDX);
    this->def_all_rec.telem_comment = this->tany.cstr(TELEM_COMMENT_IDX);

    /* normal version */
    this->def_rec.ttype = this->tany.cstr(TTYPE_IDX);
    this->def_rec.ttype_comment = this->tany.cstr(TTYPE_COMMENT_IDX);
    this->def_rec.tunit = this->tany.cstr(TUNIT_IDX);
    this->def_rec.tunit_comment = this->tany.cstr(TUNIT_COMMENT_IDX);
    this->def_rec.tdisp = this->tany.cstr(TDISP_IDX);
    this->def_rec.tform = this->tany.cstr(TFORM_IDX);
    this->def_rec.tdim  = this->tany.cstr(TDIM_IDX);
    this->def_rec.tnull = this->tany.cstr(TNULL_IDX);
    this->def_rec.tzero = this->tany.cstr(TZERO_IDX);
    this->def_rec.tscal = this->tany.cstr(TSCAL_IDX);

    //this->def_rec.talas = this->talas.cstrarray();
    //this->def_rec.telem = this->telem_def.cstrarray();
    this->def_rec.talas = this->tany.cstr(TALAS_IDX);
    this->def_rec.telem = this->tany.cstr(TELEM_IDX);

    return;
}

/**
 * @brief  ǥեͤ˥򥻥å
 *
 * @note   Υдؿ private Ǥ
 */
void fits_table_col::update_zero_of_data_rec()
{
    if ( this->type_rec == FITS::LONGARRDESC_T ||
	 this->type_rec == FITS::LLONGARRDESC_T  ) {
	this->update_zero_of_data_rec( this->heap_type_rec, true );
    }
    else {
	this->update_zero_of_data_rec( this->type_rec, false );
    }
}

/* set default value of internal memory buffer */
/**
 * @brief  ǥեͤ˥򥻥å
 *
 * @note   Υдؿ private Ǥ
 */
void fits_table_col::update_zero_of_data_rec( int new_type, bool is_heap )
{
    double zval = 0;
    mdarray *tgt_buf_rec;

    zval -= this->tzero_rec;
    zval /= this->tscal_rec;

    if ( is_heap == true ) {
	fits::llongarrdesc_t v = {0,0};	      /* largest heap pointer type */
	tgt_buf_rec = &(this->heap_rec);
	this->data_rec.assign_default((const void *)&v);
    }
    else {
	fits::doublecomplex_t v = 0 + 0i;     /* largest data type: dcomplex */
	tgt_buf_rec = &(this->data_rec);
	this->heap_rec.assign_default((const void *)&v);
    }

    if ( new_type == FITS::DOUBLE_T ) {
	fits::double_t v = zval;
	tgt_buf_rec->assign_default((const void *)&v);
    }
    else if ( new_type == FITS::FLOAT_T ) {
	fits::float_t v = zval;
	tgt_buf_rec->assign_default((const void *)&v);
    }
    else if ( new_type == FITS::SHORT_T ) {
	if ( MIN_DOUBLE_ROUND_INT16 <= zval && zval <= MAX_DOUBLE_ROUND_INT16 ) {
	    fits::short_t v = round_d2i32(zval);
	    tgt_buf_rec->assign_default((const void *)&v);
	}
	else tgt_buf_rec->assign_default((const void *)&INDEF_INT16);
    }
    else if ( new_type == FITS::LONG_T ) {
	if ( MIN_DOUBLE_ROUND_INT32 <= zval && zval <= MAX_DOUBLE_ROUND_INT32 ) {
	    fits::long_t v = round_d2i32(zval);
	    tgt_buf_rec->assign_default((const void *)&v);
	}
	else tgt_buf_rec->assign_default((const void *)&INDEF_INT32);
    }
    else if ( new_type == FITS::BYTE_T ) {
	if ( MIN_DOUBLE_ROUND_UCHAR <= zval && zval <= MAX_DOUBLE_ROUND_UCHAR ) {
	    fits::byte_t v = round_d2i32(zval);
	    tgt_buf_rec->assign_default((const void *)&v);
	}
	else tgt_buf_rec->assign_default((const void *)&INDEF_UCHAR);
    }
    else if ( new_type == FITS::LONGLONG_T ) {
	if ( MIN_DOUBLE_ROUND_INT64 <= zval && zval <= MAX_DOUBLE_ROUND_INT64 ) {
	    fits::longlong_t v = round_d2i64(zval);
	    tgt_buf_rec->assign_default((const void *)&v);
	}
	else tgt_buf_rec->assign_default((const void *)&INDEF_INT64);
    }
    else if ( new_type == FITS::DOUBLECOMPLEX_T ) {
	fits::doublecomplex_t v = zval + 0.0i;
	tgt_buf_rec->assign_default((const void *)&v);
    }
    else if ( new_type == FITS::COMPLEX_T ) {
	fits::complex_t v = zval + 0.0i;
	tgt_buf_rec->assign_default((const void *)&v);
    }
    else if ( new_type == FITS::LOGICAL_T ) {
	fits::logical_t v = 'F';
	tgt_buf_rec->assign_default((const void *)&v);
    }
    else if ( new_type == FITS::STRING_T ) {
	/* fits::byte_t v = ' '; */
	tstring dummy;
	dummy.resize(this->data_rec.bytes()).clean(' ');
	tgt_buf_rec->assign_default((const void *)dummy.cstr());
    }
    else {	/* BIT_T */
	tgt_buf_rec->assign_default((const void *)NULL);
    }
}

/*
 * BEGIN : convert_col_type()
 *
 * : fits_image ˤ convert_image_type() ¸ߤ롥ƤƱ
 *       ʤΤǡɤѹƱ˹Ԥʤ
 */

struct fits_table_col_cnv_prms {
    /* 饤Ȥ(sizeofĹΤ!) */
    double new_min;
    double new_max;
    double old_zero;
    double new_zero;
    double old_scale;
    double new_scale;
    long long old_null;
    long long new_null;
    bool old_null_is_set;
};

namespace cnv
{
#define MAKE_FUNC(fncname,org_sz_type,org_type,new_sz_type,new_type,fnc) \
static void fncname(const void *org_val_ptr, void *new_val_ptr, size_t n, int f_b, void *u_ptr) \
{ \
    const struct fits_table_col_cnv_prms *p = (const struct fits_table_col_cnv_prms *)u_ptr; \
    const org_type *org_t_ptr = (const org_type *)org_val_ptr; \
    new_type *new_t_ptr = (new_type *)new_val_ptr; \
    const double p_new_min = p->new_min; \
    const double p_new_max = p->new_max; \
    const double p_old_zero = p->old_zero; \
    const double p_new_zero = p->new_zero; \
    const double p_old_scale = p->old_scale; \
    const double p_new_scale = p->new_scale; \
    double val; \
    size_t i, n_v; \
    ssize_t d; \
    if ( f_b < 0 ) { \
	new_t_ptr += n; \
	new_t_ptr --; \
	org_t_ptr += n; \
	org_t_ptr --; \
	d = -1; \
	n_v = n; \
    } \
    else if ( 0 < f_b ) { \
	d = 1; \
	n_v = n; \
    } \
    else { \
	d = 0; \
	n_v = ((0 < n) ? 1 : 0); \
    } \
    \
    if ( org_sz_type < 0 /* org is float type */ ) { \
      if ( new_sz_type < 0 /* new is float type */ ) { \
	for ( i=0 ; i < n_v ; i++ ) { \
	  if ( !isfinite(*org_t_ptr) ) { \
	    *new_t_ptr = (new_type)(NAN); \
	  } \
	  else { \
	    val = *org_t_ptr; \
	    val *= p_old_scale; \
	    val += p_old_zero; \
	    val -= p_new_zero; \
	    val /= p_new_scale; \
	    *new_t_ptr = (new_type)val; \
	  } \
	  new_t_ptr += d; \
	  org_t_ptr += d; \
	} \
	for ( ; i < n ; i++ ) new_t_ptr[i] = new_t_ptr[0]; \
      } \
      else { /* new is not float type */ \
	for ( i=0 ; i < n_v ; i++ ) { \
	  if ( !isfinite(*org_t_ptr) ) { \
	    *new_t_ptr = p->new_null; \
	  } \
	  else { \
	    val = *org_t_ptr; \
	    val *= p_old_scale; \
	    val += p_old_zero; \
	    val -= p_new_zero; \
	    val /= p_new_scale; \
	    if ( isfinite(val) && p_new_min <= val && val <= p_new_max ) \
	      *new_t_ptr = (new_type)fnc(val); \
	    else *new_t_ptr = p->new_null; \
	  } \
	  new_t_ptr += d; \
	  org_t_ptr += d; \
	} \
	for ( ; i < n ; i++ ) new_t_ptr[i] = new_t_ptr[0]; \
      } \
    } \
    else { /* org is not float type */ \
      if ( new_sz_type < 0 /* new is float type */ ) { \
	for ( i=0 ; i < n_v ; i++ ) { \
	  if ( p->old_null_is_set == true && p->old_null == *org_t_ptr ) { \
	    *new_t_ptr = (new_type)(NAN); \
	  } \
	  else { \
	    val = *org_t_ptr; \
	    val *= p_old_scale; \
	    val += p_old_zero; \
	    val -= p_new_zero; \
	    val /= p_new_scale; \
	    *new_t_ptr = (new_type)val; \
	  } \
	  new_t_ptr += d; \
	  org_t_ptr += d; \
	} \
	for ( ; i < n ; i++ ) new_t_ptr[i] = new_t_ptr[0]; \
      } \
      else { /* new is not float type */ \
	for ( i=0 ; i < n_v ; i++ ) { \
	  if ( p->old_null_is_set == true && p->old_null == *org_t_ptr ) { \
	    *new_t_ptr = p->new_null; \
	  } \
	  else { \
	    val = *org_t_ptr; \
	    val *= p_old_scale; \
	    val += p_old_zero; \
	    val -= p_new_zero; \
	    val /= p_new_scale; \
	    if ( isfinite(val) && p_new_min <= val && val <= p_new_max ) \
	      *new_t_ptr = (new_type)fnc(val); \
	    else *new_t_ptr = p->new_null; \
	  } \
	  new_t_ptr += d; \
	  org_t_ptr += d; \
	} \
	for ( ; i < n ; i++ ) new_t_ptr[i] = new_t_ptr[0]; \
      } \
    } \
}
SLI__MDARRAY__DO_OPERATION_2TYPES(MAKE_FUNC,round_f2i32,round_f2i32,round_f2i32,round_f2i64,round_d2i32,round_d2i32,round_d2i32,round_d2i64,,,,,);
#undef MAKE_FUNC
}

/**
 * @brief  ԥͤѴ mdarray ɸѴɤɤȽ ()
 *
 * @note    private ʴؿǤ
 */
static bool simple_conv_is_ok( const fits_table_col &src, 
			       int dst_type, long dst_bytes, 
			       double dst_tzero, double dst_tscal,
			       bool dst_tnull_is_set, long long dst_tnull )
{
    bool do_simple_conv = false;
    bool src_is_float_type = false;
    bool dst_is_float_type = false;

    if ( src.type() == FITS::FLOAT_T || src.type() == FITS::DOUBLE_T ) {
	src_is_float_type = true;
    }
    if ( dst_type == FITS::FLOAT_T || dst_type == FITS::DOUBLE_T ) {
	dst_is_float_type = true;
    }

    /* ξȤưξ */
    if ( src_is_float_type == true && dst_is_float_type == true &&
	 src.tzero()  == dst_tzero && 
	 src.tscal() == dst_tscal ) {
	do_simple_conv = true;
    }
    /* ξȤξ */
    else if ( src_is_float_type == false && 
	      dst_is_float_type == false &&
	      src.bytes() <= dst_bytes &&
	      src.tzero()        == dst_tzero &&
	      src.tscal()       == dst_tscal &&
	      ((src.tnull_is_set() == false && dst_tnull_is_set == false) ||
	       (src.tnull_is_set() == true && dst_tnull_is_set == true &&
		src.tnull() == dst_tnull)) ) {
	do_simple_conv = true;
    }
    /*  float Ѵ */
    else if ( src_is_float_type == false && 
	      dst_is_float_type == true &&
	      src.tzero()        == dst_tzero &&
	      src.tscal()       == dst_tscal &&
	      src.tnull_is_set() == false ) {
	do_simple_conv = true;
    }

    return do_simple_conv;
}

/**
 * @brief  ǡѴѹ (٥)
 *
 * @note   Υдؿ private Ǥ
 */
fits_table_col &fits_table_col::convert_col_type( int new_type, 
						const double *new_zero_ptr, 
						const double *new_scale_ptr,
						const long long *new_null_ptr )
{
    struct fits_table_col_cnv_prms prms;
    void (*func_cnv_ptr)(const void *,void *,size_t,int,void *);
    ssize_t org_size_type = this->data_rec.size_type();
    ssize_t new_size_type = org_size_type;
    ssize_t pos;
    int sym = -1;
    bool rounding_bak;

    if ( this->full_bytes_rec == 0 ) return *this;	/* not defined */

    pos = this->tany.at(TFORM_IDX).strpbrk("IJKEDB");
    if ( pos < 0 ) {
	err_report1(__FUNCTION__,"WARNING",
		 "cannot convert this TFORM : %s.",this->tany.cstr(TFORM_IDX));
	return *this;
    }

    prms.old_null = 0;
    prms.old_null_is_set = false;
    prms.new_zero = 0.0;
    prms.new_scale = 1.0;
    prms.new_null = 0;
    prms.new_min = 0.0;
    prms.new_max = 0.0;

    prms.old_zero = this->tzero();
    if ( new_zero_ptr != NULL ) prms.new_zero = *new_zero_ptr;

    prms.old_scale = this->tscal();
    if ( new_scale_ptr != NULL ) prms.new_scale = *new_scale_ptr;
    if ( prms.new_scale == 0 ) {
	err_report(__FUNCTION__,"WARNING",
		   "ZERO TSCAL cannot be set, 1.0 is used.");
	prms.new_scale = 1.0;
    }

    if ( this->type_rec == FITS::DOUBLE_T ||
	 this->type_rec == FITS::FLOAT_T ) {
    }
    else {
	prms.old_null = this->tnull();
	prms.old_null_is_set = this->tnull_is_set();
    }
    if ( new_type == FITS::DOUBLE_T ||
	 new_type == FITS::FLOAT_T ) {
	new_null_ptr = NULL;
    }
    else {
	if ( new_null_ptr != NULL ) {
	    prms.new_null = get_tnull_for_type(new_type, *(new_null_ptr));
	}
	else {
	    prms.new_null = get_indef_for_type(new_type);
	}
	prms.new_min = get_min_for_type(new_type);
	prms.new_max = get_max_for_type(new_type);
    }

    switch ( new_type ) {
    case FITS::DOUBLE_T:
	new_size_type = -sizeof(fits::double_t);
	sym = 'D';
	break;
    case FITS::FLOAT_T:
	new_size_type = -sizeof(fits::float_t);
	sym = 'E';
	break;
    case FITS::LONGLONG_T:
	new_size_type = sizeof(fits::longlong_t);
	sym = 'K';
	break;
    case FITS::LONG_T:
	new_size_type = sizeof(fits::long_t);
	sym = 'J';
	break;
    case FITS::SHORT_T:
	new_size_type = sizeof(fits::short_t);
	sym = 'I';
	break;
    case FITS::BYTE_T:
	new_size_type = sizeof(fits::byte_t);
	sym = 'B';
	break;
    default:
	err_report1(__FUNCTION__,"WARNING",
		    "unsuppoeted type %d; unchanged",new_type);
	new_type = this->type_rec;
	new_size_type = this->data_rec.size_type();
	break;
    }

    rounding_bak = this->data_rec.rounding();
    this->data_rec.set_rounding(true);

    try {

	if ( simple_conv_is_ok(*this, 
			   new_type, abs(new_size_type),
			   prms.new_zero, prms.new_scale,
			   (new_null_ptr != NULL), prms.new_null) == true ) {

	    this->data_rec.convert( new_size_type );

	}
	else {

	    func_cnv_ptr = NULL;

    /* ѴΤδؿ */
#define SEL_FUNC(fncname,org_sz_type,org_type,new_sz_type,new_type,fnc) \
    if ( org_size_type == org_sz_type && new_size_type == new_sz_type ) { \
	func_cnv_ptr = &cnv::fncname; \
    }
    SLI__MDARRAY__DO_OPERATION_2TYPES(SEL_FUNC,,,,,,,,,,,,,else);
#undef SEL_FUNC

	    if ( func_cnv_ptr == NULL ) {
		err_throw(__FUNCTION__,"FATAL",
			  "detected unexpected NULL pointer");
	    }

	    this->data_rec.convert_via_udf( new_size_type, 
					    func_cnv_ptr, (void *)&prms );

	}

    }
    catch (...) {
	this->data_rec.set_rounding(rounding_bak);
	err_throw(__FUNCTION__,"FATAL","caught exception");
    }

    this->data_rec.set_rounding(rounding_bak);

    this->type_rec = new_type;
    this->bytes_rec = this->data_rec.bytes();

    this->full_bytes_rec = this->data_rec.col_length() * this->bytes_rec;
    if ( 0 < sym ) this->tany.at(TFORM_IDX).at(pos) = sym;

    if ( new_zero_ptr != NULL ) {
	this->assign_tzero(prms.new_zero);
    }
    else {
	this->erase_tzero();
    }
    if ( new_scale_ptr != NULL ) {
	this->assign_tscal(prms.new_scale);
    }
    else {
	this->erase_tscal();
    }
    if ( new_null_ptr != NULL ) {
	this->assign_tnull(*new_null_ptr);
    }
    else {
	this->erase_tnull();
    }
    
    return *this;
}

/** 
 * @brief  shallow copy ǽ֤ (̤)
 * 
 * @return  shallow copyǽʤ鿿<br>
 *          ʳλϵ
 * @note  Υдؿ private Ǥ
 */
bool fits_table_col::request_shallow_copy( fits_table_col *from_obj ) const
{
    return false;
}

/** 
 * @brief  shallow copy 򥭥󥻥뤹 (̤)
 * 
 * @note  Υдؿ private Ǥ
 */
void fits_table_col::cancel_shallow_copy( fits_table_col *from_obj ) const
{
    return;
}

/**
 * @brief  ȤˤĤơshallow copy طΥ꡼󥢥åפԤʤ (̤)
 * 
 * @note  Υдؿ private Ǥ
 */
void fits_table_col::cleanup_shallow_copy( bool do_deep_copy_for_this ) const
{
    return;
}


/**
 * @brief  inlineдؿthrow˻
 *
 * @note   Υдؿ private Ǥ
 */
void *fits_table_col::err_throw_void_p( const char *fnc, 
				    const char *lv, const char *mes)
{
    err_throw(fnc,lv,mes);
    return (void *)NULL;
}

/**
 * @brief  inlineдؿthrow˻
 *
 * @note   Υдؿ private Ǥ
 */
const void *fits_table_col::err_throw_const_void_p( const char *fnc, 
					 const char *lv, const char *mes) const
{
    err_throw(fnc,lv,mes);
    return (const void *)NULL;
}


}	/* namespace sli */

#include "private/c_memset.cc"

